/* packet-bmac.c
 * Routines for PROTONAME dissection
 * Copyright 2000, Tyler McHenry <smchenry@isi.edu>
 *
 * $Id: packet-bmac.c 4754 2006-01-31 23:58:47Z smchenry $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>

#include <epan/packet.h>
#include "prefs.h"

#include "packet-moteradio.h"
#include "packet-bmac.h"
#include "packet-data.h"
#include "packet-bmactest.h"

static dissector_handle_t data_handle;
static dissector_handle_t bmactest_handle;

/* Values for the "type" field. */

/* Forward declaration we need below */
void proto_reg_handoff_bmac(void);

/* Initialize the protocol and registered fields */
static int proto_bmac = -1;
static int hf_bmac_dest = -1;
static int hf_bmac_type = -1;
static int hf_bmac_group = -1;
static int hf_bmac_datalen = -1;

/* Initialize the subtree pointers */
static gint ett_bmac = -1;

/* Preferences */
guint bmac_max_payload = BMAC_DEFAULT_DATA_LEN;

static const guint16 mote_broadcast_address = MOTERADIO_BROADCAST;

/* Code to actually dissect the packets */
void
dissect_bmac(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
  guint8 type, data_length;
  guint16 dest;
  tvbuff_t *next_tvb;
  proto_item *ti;
  proto_tree *bmac_tree;

  if (check_col(pinfo->cinfo, COL_PROTOCOL)) 
    col_set_str(pinfo->cinfo, COL_PROTOCOL, "B-MAC");
    
  type = tvb_get_guint8(tvb, 2);

  switch(type) 
    {
    default:
      if (check_col(pinfo->cinfo, COL_INFO)) 
	col_set_str(pinfo->cinfo, COL_INFO, "Application Data");
    }

  dest = tvb_get_letohs(tvb, 0);

  if (dest == MOTERADIO_BROADCAST) {
    SET_ADDRESS(&pinfo->dst, AT_MOTE, 2, &mote_broadcast_address);
    SET_ADDRESS(&pinfo->dl_dst, AT_MOTE, 2, &mote_broadcast_address);
  } else {
    SET_ADDRESS(&pinfo->dst, AT_MOTE, 2, tvb_get_ptr(tvb, 0, 2));
    SET_ADDRESS(&pinfo->dl_dst, AT_MOTE, 2, tvb_get_ptr(tvb, 0, 2));
  }

  if (tree) {

    ti = proto_tree_add_item(tree, proto_bmac, tvb, 0, -1, FALSE);

    bmac_tree = proto_item_add_subtree(ti, ett_bmac);

    proto_tree_add_item(bmac_tree,
			hf_bmac_dest, tvb, 0, 2, TRUE);
    proto_tree_add_item(bmac_tree,
			hf_bmac_type, tvb, 2, 1, FALSE);
    proto_tree_add_item(bmac_tree,
			hf_bmac_group, tvb, 3, 1, FALSE);
    proto_tree_add_item(bmac_tree,
			hf_bmac_datalen, tvb, 4, 1, FALSE);

  }

  data_length = tvb_get_guint8(tvb, 4);
  
  if (data_length > bmac_max_payload) {
    data_length = bmac_max_payload;
    if (check_col(pinfo->cinfo, COL_INFO)) 
      col_append_str(pinfo->cinfo, COL_INFO, "[Bad payload size]");
  }
  
  next_tvb = tvb_new_subset(tvb, 5, data_length, data_length);
    
  if (data_length == BMACTEST_PACKET_LEN) {
    
    if (bmactest_handle == NULL) 
      bmactest_handle = find_dissector("bmactest");
    
    call_dissector(bmactest_handle, next_tvb, pinfo, tree);
    
  } else {
    
    if (data_handle == NULL) 
      data_handle = find_dissector("data");
    
    call_dissector(data_handle, next_tvb, pinfo, tree);
  }

}


/* Register the protocol with Ethereal */

void
proto_register_bmac(void)
{                
  module_t *bmac_module;
 
  static const value_string typevalues[] = {
    { 0, NULL },
  };

/* Setup list of header fields  See Section 1.6.1 for details*/
	static hf_register_info hf[] = {
	  { &hf_bmac_dest,
	    { "Destination Address",           "bmac.dest",
	      FT_UINT16, BASE_DEC, NULL, 0,          
	      "Destination is the intended recipient of this packet", HFILL }
	  },
	  { &hf_bmac_type,
	    { "Message Type",           "bmac.type",
	      FT_UINT8, BASE_DEC, typevalues, 0,          
	      "Type is the major type of the B-MAC (AM) packet", HFILL }
	  },
	  { &hf_bmac_group,
	    { "Group",           "bmac.group",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "(Unknown Meaning!!)", HFILL }
	  },
	  { &hf_bmac_datalen,
	    { "Data Length",           "bmac.datalen",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "Number of bytes of meaningful application data in this packet", HFILL }
	  },
	};

/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_bmac,
	};

/* Register the protocol name and description */
	proto_bmac = proto_register_protocol("Berkeley MAC",
	    "B-MAC", "bmac");

/* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_bmac, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));
        
/* Preferences Module */
        bmac_module = prefs_register_protocol(proto_bmac, NULL);

	prefs_register_uint_preference(bmac_module, "max_payload", "Maximum payload size (in bytes)",
				       "Set the size of the payload buffer in B-MAC packets (should be equivalent to TOSH_DATA_LEN in mote code)",
				       10, &bmac_max_payload);

}


void
proto_reg_handoff_bmac(void)
{
  data_handle = find_dissector("data");
  bmactest_handle = find_dissector("bmactest");

  register_dissector("bmac", dissect_bmac, proto_bmac);
}


