/* packet-linkstate.c
 * Routines for Linkstate Data dissection
 * Copyright 2005, Tyler McHenry <smchenry@isi.edu>
 *
 * $Id: packet-linkstate.c 4313 2005-08-07 11:04:14Z smchenry $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <math.h>
#include <glib.h>

#include <epan/packet.h>
#include <epan/address.h>
#include <epan/conversation.h>
#include <epan/tap.h>
#include "prefs.h"

#include "packet-linkstate.h"

/* Forward declaration we need below */

/* External functions */
void proto_reg_handoff_linkstate(void);
void linkstate_init(void);

/* Initialize the protocol and registered fields */
int proto_linkstate = -1;
static int hf_linkstate_state = -1;
static int hf_linkstate_node_id = -1;
static int hf_linkstate_sync_seq_num = -1;
static int hf_linkstate_num_rx_sync = -1;
static int hf_linkstate_num_lost_sync = -1;
static int hf_linkstate_bcast_seq_num = -1;
static int hf_linkstate_num_rx_bcast = -1;
static int hf_linkstate_num_lost_bcast = -1;
static int hf_linkstate_ucast_seq_num = -1;
static int hf_linkstate_num_lost_umsgs = -1;
static int hf_linkstate_num_tx_frags = -1;
static int hf_linkstate_num_rx_frags = -1;

/* Preferences */

/* Initialize the subtree pointers */
static gint ett_linkstate = -1;

/* Code to actually dissect the packets */

static void
dissect_linkstate(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
  proto_item *ti;
  proto_tree *linkstate_tree;
  gboolean nonzero = FALSE;
  int i;

  if (check_col(pinfo->cinfo, COL_INFO)) 
    col_set_str(pinfo->cinfo, COL_INFO, "Linkstate Data");

  for (i = 0; i < LINKSTATE_PACKET_LEN; i++) {
    if (tvb_get_guint8(tvb, i) != 0) {
      nonzero = TRUE;
      break;
    }
  }

  if (!nonzero) return;

  if (tree) {

    ti = proto_tree_add_item(tree, proto_linkstate, tvb, 0, -1, FALSE);

    linkstate_tree = proto_item_add_subtree(ti, ett_linkstate);

    proto_tree_add_item(linkstate_tree,
			hf_linkstate_state, tvb, 0, 1, TRUE);
    proto_tree_add_item(linkstate_tree,
			hf_linkstate_node_id, tvb, 1, 2, TRUE);
    proto_tree_add_item(linkstate_tree,
			hf_linkstate_sync_seq_num, tvb, 3, 1, TRUE);
    proto_tree_add_item(linkstate_tree,
			hf_linkstate_num_rx_sync, tvb, 4, 1, TRUE);
    proto_tree_add_item(linkstate_tree,
			hf_linkstate_num_lost_sync, tvb, 5, 1, TRUE);
    proto_tree_add_item(linkstate_tree,
			hf_linkstate_bcast_seq_num, tvb, 6, 1, TRUE);
    proto_tree_add_item(linkstate_tree,
			hf_linkstate_num_rx_bcast, tvb, 7, 1, TRUE);
    proto_tree_add_item(linkstate_tree,
			hf_linkstate_num_lost_bcast, tvb, 8, 1, TRUE);
    proto_tree_add_item(linkstate_tree,
			hf_linkstate_ucast_seq_num, tvb, 9, 1, TRUE);
    proto_tree_add_item(linkstate_tree,
			hf_linkstate_num_lost_umsgs, tvb, 10, 1, TRUE);
    proto_tree_add_item(linkstate_tree,
			hf_linkstate_num_tx_frags, tvb, 11, 1, TRUE);
    proto_tree_add_item(linkstate_tree,
			hf_linkstate_num_rx_frags, tvb, 12, 1, TRUE);

  }
}

void
proto_register_linkstate(void)
{                 

  static const value_string statevalues[] = {
    { 0, "Active" },
    { -1, "Inactive" },
    { 0, NULL },
  };

/* Setup list of header fields  See Section 1.6.1 for details*/
	static hf_register_info hf[] = {
	  { &hf_linkstate_state,
	    { "Neighbor State",           "linkstate.state",
	      FT_INT8, BASE_DEC, statevalues, 0,          
	      "State of this entry", HFILL }
	  },
	  { &hf_linkstate_node_id,
	    { "Node",           "linkstate.node_id",
	      FT_UINT16, BASE_DEC, NULL, 0,          
	      "ID of Node that this entry refers to ", HFILL }
	  },
	  { &hf_linkstate_sync_seq_num,
	    { "SYNC Sequence Number",           "linkstate.sync_seq_num",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "Latest Sync Sequence Number from this node", HFILL }
	  },
	  { &hf_linkstate_num_rx_sync,
	    { "SYNC Packets Received",           "linkstate.num_rx_sync",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "Total SYNCs received from this node", HFILL }
	  },
	  { &hf_linkstate_num_lost_sync,
	    { "SYNC Packets Lost",           "linkstate.num_lost_sync",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "SYNCs missing that should've been heard from this node", HFILL }
	  },
	  { &hf_linkstate_bcast_seq_num,
	    { "Broadcast Sequence Number",           "linkstate.bcast_seq_num",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "Latest broadcast sequence number from this node", HFILL }
	  },
	  { &hf_linkstate_num_rx_bcast,
	    { "Broadcast Packets Received",           "linkstate.num_rx_bcast",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "Total broadcasts received from this node", HFILL }
	  },
	  { &hf_linkstate_num_lost_bcast,
	    { "Broadcast Packets Lost",           "linkstate.num_lost_bcast",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "Broadcasts not heard that should've been", HFILL }
	  },
	  { &hf_linkstate_ucast_seq_num,
	    { "Unicast Sequence Number",           "linkstate.ucast_seq_num",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "Latest unicast sequence number heard", HFILL }
	  },
	  { &hf_linkstate_num_lost_umsgs,
	    { "Unicast Packets Lost",           "linkstate.num_lost_umsgs",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "Number of unicasts not heard that should've been", HFILL }
	  },
	  { &hf_linkstate_num_tx_frags,
	    { "Unicast Fragments Sent from Node",           "linkstate.num_tx_frags",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "Number of unicast fragments that this node reports have been sent to us", HFILL }
	  },
	  { &hf_linkstate_num_rx_frags,
	    { "Unicast Fragments Received from Node",           "linkstate.num_rx_frags",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "Number of unicast fragments that we actually got", HFILL }
	  },
	};

/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_linkstate,
	};

  register_init_routine(&linkstate_init);

/* Register the protocol name and description */
	proto_linkstate = proto_register_protocol("Linkstate Data",
	    "Linkstate", "linkstate");

/* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_linkstate, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));
        
}

void
linkstate_init(void)
{
  /* n/t */
}

void
proto_reg_handoff_linkstate(void)
{

  register_dissector("linkstate", dissect_linkstate, proto_linkstate);
  register_init_routine(&linkstate_init);

}

