/* packet-moteradio.c
 * Routines for Mote Radio dissection
 * Copyright 2000, Tyler McHenry <smchenry@isi.edu>
 *
 * $Id: packet-moteradio.c 5192 2006-04-21 00:31:13Z smchenry $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>

#include <epan/packet.h>
#include "prefs.h"

#include "packet-moteradio.h"
#include "packet-smac.h"
#include "packet-scpmac.h"
#include "packet-bmac.h"
#include "packet-rssi.h"

typedef enum MACType {
    unknown,
    unimportant,
    bmac,
    smac,
    scpmac,
} MACType;

typedef enum CRCType {
    radio,
    serial,
} CRCType;

enum { MD_AUTO,
       MD_INVALID,
       MD_BMAC,
       MD_SMAC,
       MD_SCPMAC };

const enum_val_t mac_default[] = {
  {"auto", "No Assumption (best guess)", MD_AUTO},
  {"smac", "B-MAC", MD_BMAC},
  {"bmac", "S-MAC", MD_SMAC},
  {"scpmac", "SCP-MAC", MD_SCPMAC},
  {NULL, NULL, -1}
};

static GList* allocated_memory = NULL;

/* Forward declaration we need below */
static void proto_init_moteradio(void);
void proto_reg_handoff_moteradio(void);
extern void dissect_smac(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree);
MACType determine_mac(tvbuff_t *tvb);
gint16 check_crc(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, MACType mac, CRCType crc);
gint16 moteradio_update_crc(guint8 data, gint16 crc);
void add_crc_notes(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, MACType mac,
                   CRCType crc, gint16 computed_crc);
void read_meta_flags(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, mote_priv_data *priv);

static dissector_handle_t smac_handle = -1;
static dissector_handle_t bmac_handle = -1;
static dissector_handle_t scpmac_handle = -1;
static dissector_handle_t rssi_handle = -1;

/* Preferences */
static gint moteradio_mac_default = 0;
gboolean moteradio_meta_flags = FALSE;
extern guint bmac_max_payload;

/* Initialize the protocol and registered fields */
int proto_moteradio = -1;
static int proto_badcrc = -1;
static int hf_moteradio_length = -1;
static int hf_moteradio_rx_timestamp = -1;
static int hf_moteradio_radio_crc = -1;
static int hf_moteradio_collision_candidate = -1;
static int hf_moteradio_rssi_data = -1;
static int hf_moteradio_moteside_crc_ok = -1;
static int hf_moteradio_moteside_crc_bad = -1;
static int hf_moteradio_undefined_flag_10 = -1;
static int hf_moteradio_undefined_flag_20 = -1;
static int hf_moteradio_undefined_flag_40 = -1;
static int hf_moteradio_snooper_debug = -1;
static int hf_moteradio_serial_crc = -1;

/* Initialize the subtree pointers */
static gint ett_moteradio = -1;

/* Code to actually dissect the packets */
void
dissect_moteradio(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
  tvbuff_t *next_tvb;
  proto_item *ti;
  proto_tree *moteradio_tree;
  guint8 radio_crc_start, serial_crc_start, reported_length, available_length;
  guint8 length;
  gint16 reported_crc, computed_crc;
  mote_priv_data *priv;
  MACType mac = unknown;

  /* Allocate structure for persistant data for this packet if it doesn't exist */
  
  if ((priv = (mote_priv_data *)p_get_proto_data(pinfo->fd, proto_moteradio)) == NULL) {

    priv = (mote_priv_data *)malloc(sizeof(mote_priv_data));
    allocated_memory = g_list_append(allocated_memory, priv);

    priv->initialized = FALSE;
    p_add_proto_data(pinfo->fd, proto_moteradio, (void*)priv);
  }

  /* Pre-compute commonly used indices */
  
  length = tvb_get_guint8(tvb, MOTERADIO_META_HEADER_LEN);
  radio_crc_start = tvb_length(tvb) - MOTERADIO_RADIO_CRC_LEN - MOTERADIO_TRAILER_LEN;
  serial_crc_start = tvb_length(tvb) - MOTERADIO_SERIAL_CRC_LEN;

  /* Set base packet information */
  
  if (check_col(pinfo->cinfo, COL_PROTOCOL)) 
    col_set_str(pinfo->cinfo, COL_PROTOCOL, "Mote Radio");
    
  if (check_col(pinfo->cinfo, COL_INFO)) 
    col_set_str(pinfo->cinfo, COL_INFO, "Mote Radio Packet");

  if (check_col(pinfo->cinfo, COL_PACKET_LENGTH)) 
    col_add_fstr(pinfo->cinfo, COL_PACKET_LENGTH, "%d", length);

  /* Add field dissection */
  
  if (tree) {

    ti = proto_tree_add_item(tree, proto_moteradio, tvb, 0, -1, FALSE);

    moteradio_tree = proto_item_add_subtree(ti, ett_moteradio);

    proto_tree_add_item(moteradio_tree,
			hf_moteradio_rx_timestamp, tvb, 0,
                        MOTERADIO_META_HEADER_LEN, TRUE);
    proto_tree_add_item(moteradio_tree,
			hf_moteradio_length, tvb, MOTERADIO_META_HEADER_LEN,
                        MOTERADIO_HEADER_LEN, FALSE);
    proto_tree_add_item(moteradio_tree,
                        hf_moteradio_radio_crc, tvb, radio_crc_start,
                        MOTERADIO_RADIO_CRC_LEN, TRUE);

    if (moteradio_meta_flags) {
        proto_tree_add_item(moteradio_tree,
                            hf_moteradio_collision_candidate, tvb,
                            radio_crc_start + MOTERADIO_RADIO_CRC_LEN,
                            MOTERADIO_META_FLAGS_LEN, TRUE);
        proto_tree_add_item(moteradio_tree,
                            hf_moteradio_rssi_data, tvb,
                            radio_crc_start + MOTERADIO_RADIO_CRC_LEN,
                            MOTERADIO_META_FLAGS_LEN, TRUE);
        proto_tree_add_item(moteradio_tree,
                            hf_moteradio_moteside_crc_ok, tvb,
                            radio_crc_start + MOTERADIO_RADIO_CRC_LEN,
                            MOTERADIO_META_FLAGS_LEN, TRUE);
        proto_tree_add_item(moteradio_tree,
                            hf_moteradio_moteside_crc_bad, tvb,
                            radio_crc_start + MOTERADIO_RADIO_CRC_LEN,
                            MOTERADIO_META_FLAGS_LEN, TRUE);
        proto_tree_add_item(moteradio_tree,
                            hf_moteradio_undefined_flag_10, tvb,
                            radio_crc_start + MOTERADIO_RADIO_CRC_LEN,
                            MOTERADIO_META_FLAGS_LEN, TRUE);
        proto_tree_add_item(moteradio_tree,
                            hf_moteradio_undefined_flag_20, tvb,
                            radio_crc_start + MOTERADIO_RADIO_CRC_LEN,
                            MOTERADIO_META_FLAGS_LEN, TRUE);
        proto_tree_add_item(moteradio_tree,
                            hf_moteradio_undefined_flag_40, tvb,
                            radio_crc_start + MOTERADIO_RADIO_CRC_LEN,
                            MOTERADIO_META_FLAGS_LEN, TRUE);
        proto_tree_add_item(moteradio_tree,
                            hf_moteradio_snooper_debug, tvb,
                            radio_crc_start + MOTERADIO_RADIO_CRC_LEN,
                            MOTERADIO_META_FLAGS_LEN, TRUE);
        proto_tree_add_item(moteradio_tree,
                            hf_moteradio_serial_crc, tvb,
                            serial_crc_start,
                            MOTERADIO_SERIAL_CRC_LEN, TRUE);

    }
  }

  /* Fill in the persistent data if it has not already been filled (it is
     possible that we have seen this packet before and re-computing the
     values may in fact cause errors) */
  
  if (priv->initialized == FALSE) {
      
      priv->rx_timestamp = tvb_get_letoh64(tvb, 0);

      read_meta_flags(tvb, pinfo, tree, priv);

    /* Unnecessary to check radio CRC if we see that the moteside check passed.
       Conversely, if the serial CRC is bad it is pointless to check the radio CRC
       since we can't trust it anyway. However if the mote-side check failed and
       the serial CRC is okay, we should re-compute the radio CRC so that we can
       display the computed CRC to the user. */
    
      if (!moteradio_meta_flags || ((priv->meta_flags & MOTERADIO_FLAG_MOTESIDE_CRC_OK) == 0)) {

          reported_crc = tvb_get_letohs(tvb, radio_crc_start);
          computed_crc = check_crc(tvb, pinfo, tree, mac, radio);
          
          priv->crc_error_radio = ((reported_crc != computed_crc) && (computed_crc != 1));
          priv->computed_crc_radio = computed_crc;

      } else {

          priv->crc_error_radio = FALSE;
          priv->computed_crc_radio = tvb_get_letohs(tvb, radio_crc_start);
      }
      
      priv->initialized = TRUE;
  }

  /* Extract the next level protocol's data from the moteradio packet */
  
  reported_length = length - MOTERADIO_HEADER_LEN - MOTERADIO_RADIO_CRC_LEN; 
  available_length = tvb_length(tvb) - MOTERADIO_HEADER_LEN -
      MOTERADIO_RADIO_CRC_LEN - MOTERADIO_TRAILER_LEN;

  next_tvb = tvb_new_subset(tvb, MOTERADIO_META_HEADER_LEN + MOTERADIO_HEADER_LEN, 
			    MIN(available_length, reported_length),
			    reported_length);


  /* Call the appropriate dissector for the next protocol */
  
  if ((priv->meta_flags & MOTERADIO_FLAG_RSSI_DATA) == 0) {
      
      mac = determine_mac(tvb);
      
      if (mac == smac) {
          if (smac_handle <= 0) 
              smac_handle = find_dissector("smac");    
          call_dissector(smac_handle, next_tvb, pinfo, tree);
      } else if (mac == bmac) {
          if (bmac_handle <= 0) 
              bmac_handle = find_dissector("bmac");    
          call_dissector(bmac_handle, next_tvb, pinfo, tree);
      } else if (mac == scpmac) {
          if (scpmac_handle <= 0)
              scpmac_handle = find_dissector("scpmac");
          call_dissector(scpmac_handle, next_tvb, pinfo, tree);
      }
  } else {
      if (rssi_handle <= 0)
          rssi_handle = find_dissector("rssi");

      priv->collision = TRUE;
      
      call_dissector(rssi_handle, next_tvb, pinfo, tree);

      /* Pass to S-MAC *only if* we are specifically snooping
         S-MAC. (otherwise we can't assume or detect which
         MAC the collision was on) */
      
      if (moteradio_mac_default == MD_SMAC)
          tap_queue_packet(smac_tap, pinfo, NULL);
  }


  /* After the higher-level protocols are done with it, get the persistant data back
     and add any notes that might be necessary. This is done AFTER higher level
     dissectors so that the higher-level dissectors can overwrite the default
     packet description without clobbering the notes in the description field. */
  
  priv = (mote_priv_data *)p_get_proto_data(pinfo->fd, proto_moteradio);
  
  if (priv->initialized == TRUE) {

      if (priv->crc_error_serial == TRUE) {
          add_crc_notes(tvb, pinfo, tree, mac, serial, priv->computed_crc_serial);
      } else if (priv->crc_error_radio == TRUE) {
          add_crc_notes(tvb, pinfo, tree, mac, radio, priv->computed_crc_radio);
      }

      if (!moteradio_meta_flags || !(priv->crc_error_serial)) {
      
          if (priv->debug == TRUE && check_col(pinfo->cinfo, COL_INFO))
              col_append_str(pinfo->cinfo, COL_INFO, " [Debug]");
          
          if (priv->collision == TRUE && check_col(pinfo->cinfo, COL_INFO))
              col_append_str(pinfo->cinfo, COL_INFO, " [Collision]");
      }
  }
  
}


/* Register the protocol with Ethereal */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_moteradio(void)
{                 
  module_t *moteradio_module;

  static const value_string flagvalues [] = {
      { 0, "NO" },
      { 1, "YES" },
  };
  
/* Setup list of header fields  See Section 1.6.1 for details*/
	static hf_register_info hf[] = {
	  { &hf_moteradio_length,
	    { "Packet Length",           "moteradio.length",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "The full length of this packet (including CRC and length bytes)", HFILL }
	  },
	  { &hf_moteradio_rx_timestamp,
	    { "Time Received by Kernel",           "moteradio.rx_timestamp",
	      FT_UINT64, BASE_DEC, NULL, 0,          
	      "The exact time (in microseconds) that the start symbol for this packet was received by the moteradio driver", HFILL }
	  },
	  { &hf_moteradio_radio_crc,
	    { "CRC-16 Checksum (Radio)",           "moteradio.radio_crc",
	      FT_UINT16, BASE_HEX, NULL, 0,          
	      "A CRC-16 checksum for the portion of this packet that was transmitted over the radio", HFILL }
	  },
          { &hf_moteradio_collision_candidate, 
            { "Collision Candidate Flag", "moteradio.collision_candidate",
              FT_UINT8, BASE_DEC, flagvalues, MOTERADIO_FLAG_POSSIBLE_COLLISION,
              "Collision Candidate Flag indicates that the snooper thinks that this packet may have collided with another.", HFILL }
          },
          { &hf_moteradio_rssi_data, 
            { "RSSI Data Flag", "moteradio.rssi_data",
              FT_UINT8, BASE_DEC, flagvalues, MOTERADIO_FLAG_RSSI_DATA,
              "RSSI Data indicates that this packet is an RSSI reding generated by the mote (may indicate a total collision)", HFILL }
          },
          { &hf_moteradio_moteside_crc_ok, 
            { "Moteside CRC OK Flag", "moteradio.moteside_crc_ok",
              FT_UINT8, BASE_DEC, flagvalues, MOTERADIO_FLAG_MOTESIDE_CRC_OK,
              "Moteside CRC OK Flag indicates that the snooper checked the CRC and found it to be correct.", HFILL }
          },
          { &hf_moteradio_moteside_crc_bad, 
            { "Moteside CRC Bad Flag", "moteradio.moteside_crc_bad",
              FT_UINT8, BASE_DEC, flagvalues, MOTERADIO_FLAG_MOTESIDE_CRC_BAD,
              "Moteside CRC Bad Flag indicates that the snooper checked the CRC and found it to be bad.", HFILL }
          },
          { &hf_moteradio_undefined_flag_10, 
            { "Undefined Flag (0x10)", "moteradio.undefined_flag_10",
              FT_UINT8, BASE_DEC, flagvalues, MOTERADIO_FLAG_UNDEFINED_10,
              "Flag with hex value 0x10 is currently undefined.", HFILL }
          },
          { &hf_moteradio_undefined_flag_20, 
            { "Undefined Flag (0x20)", "moteradio.undefined_flag_20",
              FT_UINT8, BASE_DEC, flagvalues, MOTERADIO_FLAG_UNDEFINED_20,
              "Flag with hex value 0x20 is currently undefined.", HFILL }
          },
          { &hf_moteradio_undefined_flag_40, 
            { "Undefined Flag (0x40)", "moteradio.undefined_flag_40",
              FT_UINT8, BASE_DEC, flagvalues, MOTERADIO_FLAG_UNDEFINED_40,
              "Flag with hex value 0x40 is currently undefined.", HFILL }
          },
          { &hf_moteradio_snooper_debug, 
            { "Snooper Debug Flag", "moteradio.snooper_debug",
              FT_UINT8, BASE_DEC, flagvalues, MOTERADIO_FLAG_DEBUG,
              "Snooper Debug Flag indicates that this packet is debugging info for the snooper program and should be ignored.", HFILL }
          },
          { &hf_moteradio_serial_crc,
	    { "CRC-16 Checksum (Serial)",           "moteradio.serial_crc",
	      FT_UINT16, BASE_HEX, NULL, 0,          
	      "A CRC-16 checksum for the portion of this packet that was transmitted over the serial line", HFILL }
	  },
	};

/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_moteradio,
	};

/* Register the protocol name and description */
	proto_moteradio = proto_register_protocol("Mica-2 Mote Radio",
	    "Mote Radio", "moteradio");
	proto_badcrc = proto_register_protocol("Bad CRC",
	    "Bad CRC", "badcrc");

	proto_set_cant_toggle(proto_badcrc);

/* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_moteradio, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

	register_init_routine(&proto_init_moteradio);
        
/* Register preferences module (See Section 2.6 for more on preferences) */       
        moteradio_module = prefs_register_protocol(proto_moteradio, proto_reg_handoff_moteradio);
     
	prefs_register_enum_preference(moteradio_module,
				       "mac_default",
				       "Assume all ambiguous packets are",
				       "If Ethereal cannot tell for certain that a packet is a particular MAC type, you may instruct it to choose one, or let it guess.",
				       &moteradio_mac_default,
				       mac_default, FALSE);

	prefs_register_bool_preference(moteradio_module,
                                       "meta_flags",
                                       "Assume snooper uses meta flags",
                                       "Whether the dissector should expect a meta flag byte after the CRC in all packets. Use only if your snooper actually does this.",
                                       &moteradio_meta_flags);

        
}


/* If this dissector uses sub-dissector registration add a registration routine.
   This exact format is required because a script is used to find these routines 
   and create the code that calls these routines.
   
   This function is also called by preferences whenever "Apply" is pressed 
   (see prefs_register_protocol above) so it should accommodate being called 
   more than once. */

void
proto_reg_handoff_moteradio(void)
{
  if (smac_handle == -1)
    register_dissector("moteradio", dissect_moteradio, proto_moteradio); 

  smac_handle = find_dissector("smac");
  bmac_handle = find_dissector("bmac");
  scpmac_handle = find_dissector("scpmac");
  rssi_handle = find_dissector("rssi");

  /*
      moteradio_handle = create_dissector_handle(dissect_moteradio, proto_moteradio);
      dissector_add("wtap_encap", WTAP_ENCAP_RAW_IP, moteradio_handle);
  */
}

static void
proto_init_moteradio(void)
{
  GList* mem_list = g_list_first(allocated_memory);
  mote_priv_data* priv;

  while(mem_list) {
    priv = (mote_priv_data*)(mem_list->data);
    free(priv);
    
    mem_list = g_list_next(mem_list);
  }

  g_list_free(allocated_memory);

  allocated_memory = NULL;
}

MACType determine_mac(tvbuff_t *tvb) 
{
  gint8 bmac_prob = 0;
  gint8 scpmac_prob = 0;

  if (tvb_length(tvb) == BMAC_HEADER_LEN + bmac_max_payload +
      MOTERADIO_META_HEADER_LEN + MOTERADIO_HEADER_LEN + 
      MOTERADIO_RADIO_CRC_LEN + MOTERADIO_TRAILER_LEN) {

    /* Potentially B-MAC */

    if (tvb_get_guint8(tvb, MOTERADIO_META_HEADER_LEN + 5) > bmac_max_payload) {

      bmac_prob = -1; /* definitely NOT B-MAC */

    } else if (((tvb_get_guint8(tvb, 1) & 0xF0) >> 4) != 0) {

      bmac_prob += MOTERADIO_BMAC_THRESHOLD; /* definitely B-MAC */

    } else if (moteradio_mac_default == MD_AUTO) { /* Maybe B-MAC, see how likely */
    
      if (tvb_get_guint8(tvb, MOTERADIO_META_HEADER_LEN + 4) == 0x7d) {
	bmac_prob++;
      }

      if (tvb_get_letohs(tvb, tvb_length(tvb) - MOTERADIO_RADIO_CRC_LEN - MOTERADIO_TRAILER_LEN) == 0 ||
	  tvb_get_letohs(tvb, tvb_length(tvb) - MOTERADIO_RADIO_CRC_LEN - MOTERADIO_TRAILER_LEN) == 1) {
	bmac_prob++;
      }

      if (tvb_get_guint8(tvb, MOTERADIO_META_HEADER_LEN + 2) != 0) {
	bmac_prob++;
      }
    }
  } 

  
  if (bmac_prob >= MOTERADIO_BMAC_THRESHOLD)
    return bmac;
  
  
  /* Don't know enough at this point to detect SCP-MAC. User needs
     to set it as default for ambiguous packets */

  if (moteradio_mac_default == MD_BMAC) return bmac;
  if (moteradio_mac_default == MD_SMAC) return smac;
  if (moteradio_mac_default == MD_SCPMAC) return scpmac;
   
  return smac;
}

gint16 check_crc(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, MACType mac, CRCType crc)
{
  gint16 computed_crc = 0;
  gint16 computed_serial_crc = 0;
  guint8 length;
  int i;

  if (crc == radio) {
  
      if (mac == smac || mac == unknown) {
          
          length = tvb_get_guint8(tvb, MOTERADIO_META_HEADER_LEN);
          computed_crc = 0; /* HURRR this is important! */
          
          for (i = MOTERADIO_META_HEADER_LEN; i < tvb_length(tvb) - MOTERADIO_RADIO_CRC_LEN - MOTERADIO_TRAILER_LEN; i++) 
              computed_crc = moteradio_update_crc(tvb_get_guint8(tvb, i), computed_crc);
          
          return computed_crc;
          
      } else if (mac == bmac) {
          return 1;
      }

  } else if (crc == serial) {

      if (!moteradio_meta_flags)
          return 1;

      /* IMPORTANT: The serial CRC *includes* the 0x7E start symbol which is BEFORE the meta-headers
         added by the kernel (which are obviously not in any CRC). We must add 0x7E to the beginning
         of the CRC manually */
      
      computed_serial_crc = 0;
      computed_serial_crc = moteradio_update_crc(0x7E, computed_serial_crc);
      for (i = MOTERADIO_META_HEADER_LEN; i < tvb_length(tvb) - MOTERADIO_SERIAL_CRC_LEN; i++) 
          computed_serial_crc = moteradio_update_crc(tvb_get_guint8(tvb, i), computed_serial_crc);
      
      return computed_serial_crc;
      
  } else {
      return 1;
  }

  return 1;
}

gint16 moteradio_update_crc(guint8 data, gint16 crc)
{
  char i;
  guint16 tmp;
  tmp = (guint16)(data);
  crc = crc ^ (tmp << 8);
  for (i = 0; i < 8; i++) {
    if (crc & 0x8000)
      crc = crc << 1 ^ 0x1021;  /* << is done before ^ */
    else
      crc = crc << 1;
    }
  return crc;
}

void add_crc_notes(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, MACType mac,
		   CRCType crc, gint16 computed_crc)
{
    gint16 reported_crc;

    if (crc == radio)
        reported_crc = tvb_get_letohs(tvb, tvb_length(tvb) - MOTERADIO_RADIO_CRC_LEN - MOTERADIO_TRAILER_LEN);
    else if (crc == serial)
        reported_crc = tvb_get_letohs(tvb, tvb_length(tvb) - MOTERADIO_SERIAL_CRC_LEN);
    else
        reported_crc = 1;

  if (mac == smac || moteradio_meta_flags) {
      
      if (check_col(pinfo->cinfo, COL_INFO)) {
          if (crc == radio && moteradio_meta_flags)
              col_append_str(pinfo->cinfo, COL_INFO, " [Bad Radio CRC]");
          else if (crc == serial)
              col_append_str(pinfo->cinfo, COL_INFO, " [Bad Serial CRC]");
          else 
              col_append_str(pinfo->cinfo, COL_INFO, " [Bad CRC]");
      }
          
      proto_tree_add_protocol_format(tree, proto_badcrc,
                                     tvb, 0, 0, "[Bad %s CRC: 0x%X reported, 0x%X computed]",
                                     (crc == serial) ? "Serial" : "Radio",
                                     (guint16)reported_crc, (guint16)computed_crc);
          
  } else if (mac == bmac) {
      if (check_col(pinfo->cinfo, COL_INFO)) 
	col_append_str(pinfo->cinfo, COL_INFO, " [Bad CRC]");
      
      proto_tree_add_protocol_format(tree, proto_badcrc,
				   tvb, 0, 0, "[Bad CRC: (mote-side check)]");
  }
}


void read_meta_flags(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, mote_priv_data *priv)
{
    gint16 radio_crc_start, serial_crc_start;
    gint16 reported_crc, computed_crc;
    
    radio_crc_start = tvb_length(tvb) - MOTERADIO_RADIO_CRC_LEN -
        MOTERADIO_TRAILER_LEN;

    serial_crc_start = tvb_length(tvb) - MOTERADIO_SERIAL_CRC_LEN;
    
    if (moteradio_meta_flags) {
        reported_crc = tvb_get_letohs(tvb, serial_crc_start);
        computed_crc = check_crc(tvb, pinfo, tree, unimportant, serial);
        
        priv->crc_error_serial = (reported_crc != computed_crc) ? TRUE : FALSE;
        priv->computed_crc_serial = computed_crc;


        /* If the serial CRC is bad, we cannot trust the flag values. In that case,
           proceed as if we had no flags at all. */
        
        if (!(priv->crc_error_serial)) {
            priv->meta_flags = tvb_get_guint8(tvb, radio_crc_start + MOTERADIO_RADIO_CRC_LEN);
            
            priv->collision = ((priv->meta_flags & MOTERADIO_FLAG_POSSIBLE_COLLISION) != 0 &&
                               priv->crc_error_radio);

            priv->debug = ((priv->meta_flags & MOTERADIO_FLAG_DEBUG) != 0);
            
        } else {
            priv->meta_flags = 0;
            priv->collision = FALSE;
            priv->debug = FALSE;
        }
        
    } else {
        priv->meta_flags = 0;
        priv->crc_error_serial = FALSE;
        priv->computed_crc_serial = 1;
        priv->collision = FALSE;
        priv->debug = FALSE;
    }
}
