/* packet-scpmac.c
 * Routines for SCP-MAC dissection
 * Copyright 2005, Tyler McHenry <smchenry@isi.edu>
 *
 * $Id: packet-scpmac.c 5853 2006-10-22 17:39:47Z smchenry $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* IMPORTANT NOTE:
 * 
 * Because CSMA headers are rather short and simple and because SCP-MAC
 * hijacks one or more CSMA fields for SCP-MAC specific uses, I'm 
 * considering CSMA/SCP-MAC one protocol and writing them in the same
 * dissector.
 *
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <math.h>
#include <glib.h>

#include <epan/packet.h>
#include <epan/address.h>
#include <epan/conversation.h>
#include <epan/tap.h>
#include "prefs.h"

#include "packet-scpmac.h"
#include "packet-moteradio.h"
#include "packet-data.h"

static dissector_handle_t data_handle = NULL;
static const guint16 mote_broadcast_address = MOTERADIO_BROADCAST;
static const guint16 mote_csma_ctrl_pkt_address = SCPMAC_CSMA_CTRL_ADDR;

/* Forward declaration we need below */

static void scpmac_set_column_values(tvbuff_t *tvb, packet_info *pinfo, 
				     guint8 *csma_type, guint8 *scp_type,
				     guint16 *source);

/* External functions */
void proto_reg_handoff_scpmac(void);
void scpmac_init(void);

/* Initialize the protocol and registered fields */
/* Fields that are commented out in SCP-MAC source are added in this
   dissector but also commented out */

int proto_scpmac = -1;
static int hf_scpmac_csma_type = -1;
static int hf_scpmac_scp_type = -1;
static int hf_scpmac_source = -1;
static int hf_scpmac_dest = -1;
static int hf_scpmac_seq_num = -1; 
static int hf_scpmac_frag_num = -1;
static int hf_scpmac_long_seq_num = -1;
static int hf_scpmac_poll_time = -1; /* Piggybacked on dest for BCAST */
static int hf_scpmac_num_rx_sync = -1;
static int hf_scpmac_time_diff = -1;
static int hf_scpmac_tx_time = -1;
static int hf_scpmac_backoff_slots = -1;
static int hf_scpmac_tone_len = -1;
static int hf_scpmac_tone_start_time = -1;
static int hf_scpmac_tone_done_time = -1;

/* Preferences */

/* Initialize the subtree pointers */
static gint ett_scpmac = -1;

/* Code to actually dissect the packets */

void
dissect_scpmac(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
  tvbuff_t *next_tvb;
  proto_item *ti;
  proto_tree *scpmac_tree;
  guint16 length = tvb_length(tvb);
  gint extraoffset = 0;
  guint i;
  guint data_length;
  guint8 csma_type, scp_type;
  guint16 source, dest;

  scpmac_set_column_values(tvb, pinfo, &csma_type, &scp_type, &source);

  if (tree) {
    ti = proto_tree_add_item(tree, proto_scpmac, tvb, 0, -1, FALSE);
    scpmac_tree = proto_item_add_subtree(ti, ett_scpmac);

    proto_tree_add_item(scpmac_tree,
			hf_scpmac_csma_type, tvb, 0, 1, TRUE);
    proto_tree_add_item(scpmac_tree,
			hf_scpmac_scp_type, tvb, 0, 1, TRUE);

    if (scp_type == TYPE_SCP_UCAST_PKT || csma_type == TYPE_CSMA_RTS_PKT) {
      proto_tree_add_item(scpmac_tree,
			hf_scpmac_source, tvb, 1, 2, TRUE);
    }

    if (scp_type != TYPE_SCP_BCAST_PKT &&  csma_type != TYPE_CSMA_BCAST_PKT) {
      if (csma_type == TYPE_CSMA_CTS_PKT || csma_type == TYPE_CSMA_ACK_PKT) {
      	proto_tree_add_item(scpmac_tree,
			    hf_scpmac_dest, tvb, 1, 2, TRUE);
      } else {
      	proto_tree_add_item(scpmac_tree,
			    hf_scpmac_dest, tvb, 3, 2, TRUE);
      }
    }
    
    if (scp_type == TYPE_SCP_BCAST_PKT) {
      proto_tree_add_item(scpmac_tree,
			  hf_scpmac_poll_time, tvb, 3, 2, TRUE);
    } else if (scp_type == TYPE_SCP_SYNC_PKT) {
      proto_tree_add_item(scpmac_tree,
			  hf_scpmac_poll_time, tvb, 5, 2, TRUE);
    }   
#ifdef TBD    
#ifdef SCPMAC_IMPLEMENT_SEQFRAG_NUM      
    if (csma_type == TYPE_CSMA_BCAST_PKT) {
      proto_tree_add_item(scpmac_tree,
			  hf_scpmac_long_seq_num, tvb, 6, 1, TRUE); 
    } else {
      proto_tree_add_item(scpmac_tree,
			  hf_scpmac_seq_num, tvb, 6, 1, TRUE);
      proto_tree_add_item(scpmac_tree,
			  hf_scpmac_frag_num, tvb, 6, 1, TRUE);
    }
    extraoffset += 1;
#endif

    if (scp_type == TYPE_SCP_SYNC_PKT) {   
      
      proto_tree_add_item(scpmac_tree,
			  hf_scpmac_poll_time, tvb, 6 + extraoffset, 2, TRUE);
      proto_tree_add_item(scpmac_tree,
			  hf_scpmac_num_rx_sync, tvb, 8 + extraoffset, 1, TRUE);
      proto_tree_add_item(scpmac_tree,
			  hf_scpmac_time_diff, tvb, 9 + extraoffset, 2, TRUE);
      
#ifdef SCPMAC_IMPLEMENT_EXTENDED_SYNCS
      proto_tree_add_item(scpmac_tree,
			  hf_scpmac_tx_time, tvb, 11 + extraoffset, 1, TRUE);
      proto_tree_add_item(scpmac_tree,
			  hf_scpmac_backoff_slots, tvb, 12 + extraoffset, 2, TRUE);
      proto_tree_add_item(scpmac_tree,
			  hf_scpmac_tone_len, tvb, 14 + extraoffset, 1, TRUE);
      proto_tree_add_item(scpmac_tree,
			  hf_scpmac_tone_start_time, tvb, 15 + extraoffset, 2, TRUE);
      proto_tree_add_item(scpmac_tree,
			  hf_scpmac_tone_done_time, tvb, 17 + extraoffset, 2, TRUE);
      extraoffset += 8;
#endif
    }
#endif //TBD

  }

  if (scp_type != TYPE_SCP_SYNC_PKT || 
	csma_type != TYPE_CSMA_RTS_PKT ||
	csma_type != TYPE_CSMA_CTS_PKT ||
	csma_type != TYPE_CSMA_ACK_PKT ) {
    data_handle = find_dissector("data");    
    
    data_length = tvb_length(tvb) - (SCPMAC_CSMA_HEADER_LEN);
    
    next_tvb = tvb_new_subset(tvb, (SCPMAC_CSMA_HEADER_LEN),
			      data_length, -1);
  
    if (data_handle == NULL) 
      data_handle = find_dissector("data");
      
    call_dissector(data_handle, next_tvb, pinfo, tree);
  } 

}

static void scpmac_set_column_values(tvbuff_t *tvb, packet_info *pinfo, 
				     guint8 *csma_type, guint8 *scp_type,
				     guint16 *source)
{
  guint16 dest;

  if (check_col(pinfo->cinfo, COL_PROTOCOL)) 
    col_set_str(pinfo->cinfo, COL_PROTOCOL, "SCP-MAC");

  *csma_type = (tvb_get_guint8(tvb, 0) & 0xF0) >> 4;
  *scp_type = (tvb_get_guint8(tvb, 0) & 0x0F);    

  switch(*csma_type) 
    {
    case TYPE_CSMA_RTS_PKT:
      if (check_col(pinfo->cinfo, COL_INFO)) 
	col_set_str(pinfo->cinfo, COL_INFO, "Request-to-Send");
      break;
    case TYPE_CSMA_CTS_PKT:
      if (check_col(pinfo->cinfo, COL_INFO)) 
	col_set_str(pinfo->cinfo, COL_INFO, "Clear-to-Send");
      break;
    case TYPE_CSMA_ACK_PKT:
      if (check_col(pinfo->cinfo, COL_INFO)) 
	col_set_str(pinfo->cinfo, COL_INFO, "Acknowledgement");
      break;
    case TYPE_CSMA_UCAST_PKT:
    case TYPE_CSMA_BCAST_PKT:
      switch (*scp_type) 
	{
	case TYPE_SCP_UCAST_PKT:
	  if (check_col(pinfo->cinfo, COL_INFO)) 
	    col_set_str(pinfo->cinfo, COL_INFO, "Application Data [Unicast]");
	  break;
	case TYPE_SCP_BCAST_PKT:
	  if (check_col(pinfo->cinfo, COL_INFO)) 
	    col_set_str(pinfo->cinfo, COL_INFO, "Application Data [Broadcast]");
	  break;
	case TYPE_SCP_SYNC_PKT:
	  if (check_col(pinfo->cinfo, COL_INFO)) 
	    col_set_str(pinfo->cinfo, COL_INFO, "Polling Synchronization Data");
	  break;
	case TYPE_SCP_SYNC_REQ_PKT:
	  if (check_col(pinfo->cinfo, COL_INFO)) 
	    col_set_str(pinfo->cinfo, COL_INFO, "Requesting Synchronization Data");
	  break;
	default:
	  switch(*csma_type) {
	  case TYPE_CSMA_UCAST_PKT:
	    if (check_col(pinfo->cinfo, COL_INFO)) 
	      col_set_str(pinfo->cinfo, COL_INFO, "Unrecognized SCP-MAC Packet [Unicast]");
	    break;
	  case TYPE_CSMA_BCAST_PKT:
	    if (check_col(pinfo->cinfo, COL_INFO)) 
	      col_set_str(pinfo->cinfo, COL_INFO, "Unrecognized SCP-MAC Packet [Broadcast]");
	    break;
	  default:
	    if (check_col(pinfo->cinfo, COL_INFO)) 
	      col_set_str(pinfo->cinfo, COL_INFO, "Unrecognized SCP-MAC Packet");
	  }
	}
      break;
    default:
      if (check_col(pinfo->cinfo, COL_INFO)) 
	col_set_str(pinfo->cinfo, COL_INFO, "Unrecognized SCP-MAC Packet");
    }

  if (*csma_type == TYPE_CSMA_BCAST_PKT) {
    *source = tvb_get_letohs(tvb, 1);
    dest = 0xFFFF;

    SET_ADDRESS(&pinfo->src, AT_MOTE, 2, tvb_get_ptr(tvb, 1, 2));
    SET_ADDRESS(&pinfo->dl_src, AT_MOTE, 2, tvb_get_ptr(tvb, 1, 2));
    SET_ADDRESS(&pinfo->dst, AT_MOTE, 2, &mote_broadcast_address);
    SET_ADDRESS(&pinfo->dl_dst, AT_MOTE, 2, &mote_broadcast_address);

  } else if ((*csma_type == TYPE_CSMA_CTS_PKT)||(*csma_type == TYPE_CSMA_ACK_PKT)){
    *source = 0xFFFF;
    dest = tvb_get_letohs(tvb, 1);

    SET_ADDRESS(&pinfo->src, AT_MOTE, 2, &mote_csma_ctrl_pkt_address);
    SET_ADDRESS(&pinfo->dl_src, AT_MOTE, 2, &mote_csma_ctrl_pkt_address);
    SET_ADDRESS(&pinfo->dst, AT_MOTE, 2, tvb_get_ptr(tvb, 1, 2));
    SET_ADDRESS(&pinfo->dl_dst, AT_MOTE, 2, tvb_get_ptr(tvb, 1, 2));

  } else {
    *source = tvb_get_letohs(tvb, 1);
    dest = tvb_get_letohs(tvb, 3);

    SET_ADDRESS(&pinfo->src, AT_MOTE, 2, tvb_get_ptr(tvb, 1, 2));
    SET_ADDRESS(&pinfo->dl_src, AT_MOTE, 2, tvb_get_ptr(tvb, 1, 2));
    SET_ADDRESS(&pinfo->dst, AT_MOTE, 2, tvb_get_ptr(tvb, 3, 2));
    SET_ADDRESS(&pinfo->dl_dst, AT_MOTE, 2, tvb_get_ptr(tvb, 3, 2));
  }
  
  pinfo->ptype = proto_scpmac;
  pinfo->srcport = 0;
  pinfo->destport = 0;

  if (check_col(pinfo->cinfo, COL_DEF_DST)) 
    col_add_fstr(pinfo->cinfo, COL_DEF_DST, "%d", dest);

  if (dest == MOTERADIO_BROADCAST) {
    if (check_col(pinfo->cinfo, COL_RES_DST)) 
      col_set_str(pinfo->cinfo, COL_RES_DST, "Broadcast");
  } else if (dest == SCPMAC_CSMA_CTRL_ADDR) {
    if (check_col(pinfo->cinfo, COL_RES_DST)) 
      col_set_str(pinfo->cinfo, COL_RES_DST, "CSMA Control Pkt");
  } else {
    if (check_col(pinfo->cinfo, COL_RES_DST)) 
      col_add_fstr(pinfo->cinfo, COL_RES_DST, "Node %d", dest);
  }

  if (check_col(pinfo->cinfo, COL_DEF_SRC)) 
    col_add_fstr(pinfo->cinfo, COL_DEF_SRC, "%d", *source);

  if (check_col(pinfo->cinfo, COL_RES_SRC)) 
    col_add_fstr(pinfo->cinfo, COL_RES_SRC, "Node %d", *source);

}

/* Register the protocol with Ethereal */

void
proto_register_scpmac(void)
{
  static gint *ett[] = {
    &ett_scpmac,
  };

  static const value_string csma_typevalues[] = {
    { TYPE_CSMA_UCAST_PKT, "Unicast Packet" },
    { TYPE_CSMA_BCAST_PKT, "Broadcast Packet" },
    { TYPE_CSMA_RTS_PKT, "RTS Packet" },
    { TYPE_CSMA_CTS_PKT, "CTS Packet" },
    { TYPE_CSMA_ACK_PKT, "ACK Packet" },
    { 0, NULL },
  };

  static const value_string scp_typevalues[] = {
    { TYPE_SCP_UCAST_PKT, "Unicast Data Packet" },
    { TYPE_SCP_BCAST_PKT, "Broadcast Data Packet" },
    { TYPE_SCP_SYNC_PKT, "Sync Packet" },
    { TYPE_SCP_SYNC_REQ_PKT, "Sync Request Packet" },
    { 0, NULL },
  };

  static hf_register_info hf[] = {
    { &hf_scpmac_csma_type,
      { "CSMA Packet Type",           "scpmac.csma_type",
	FT_UINT8, BASE_DEC, csma_typevalues, 0xF0,          
	"", HFILL }
    },
    { &hf_scpmac_scp_type,
      { "SCP Packet Type",           "scpmac.scp_type",
	FT_UINT8, BASE_DEC, scp_typevalues, 0x0F,          
	"", HFILL }
    },
    { &hf_scpmac_source,
      { "Source Address",           "scpmac.source",
	FT_UINT16, BASE_DEC, NULL, 0,          
	"", HFILL }
    },
    { &hf_scpmac_dest,
      { "Destination Address",           "scpmac.dest",
	FT_UINT16, BASE_DEC, NULL, 0,          
	"", HFILL }
    },
    { &hf_scpmac_seq_num,
      { "Sequence Number",           "scpmac.seq_num",
	FT_UINT8, BASE_DEC, NULL, 0x1F,          
	"", HFILL }
    },
    { &hf_scpmac_frag_num,
      { "Fragment Number",           "scpmac.frag_num",
	FT_UINT8, BASE_DEC, NULL, 0xE0,          
	"", HFILL }
    },
    { &hf_scpmac_long_seq_num,
      { "Sequence Number",           "scpmac.long_seq_num",
	FT_UINT8, BASE_DEC, NULL, 0,          
	"", HFILL }
    },
    { &hf_scpmac_poll_time,
      { "Next Polling Time",           "scpmac.poll_time",
	FT_UINT16, BASE_DEC, NULL, 0,          
	"", HFILL }
    },
    { &hf_scpmac_num_rx_sync,
      { "Sync Packets Received",           "scpmac.num_rx_sync",
	FT_UINT8, BASE_DEC, NULL, 0,          
	"", HFILL }
    },
    { &hf_scpmac_time_diff,
      { "Time Difference",           "scpmac.time_diff",
	FT_INT16, BASE_DEC, NULL, 0,          
	"", HFILL }
    },
    { &hf_scpmac_tx_time,
      { "Transmission Time",           "scpmac.tx_time",
	FT_UINT8, BASE_DEC, NULL, 0,          
	"", HFILL }
    },
    { &hf_scpmac_backoff_slots,
      { "Backoff Slots",           "scpmac.backoff_slots",
	FT_UINT16, BASE_DEC, NULL, 0,          
	"", HFILL }
    },
    { &hf_scpmac_tone_len,
      { "Tone Length",           "scpmac.tone_len",
	FT_UINT8, BASE_DEC, NULL, 0,          
	"", HFILL }
    },
    { &hf_scpmac_tone_start_time,
      { "Tone Start Time",           "scpmac.tone_start_time",
	FT_UINT16, BASE_DEC, NULL, 0,          
	"", HFILL }
    },
    { &hf_scpmac_tone_done_time,
      { "Tone Done Time",           "scpmac.tone_done_time",
	FT_UINT16, BASE_DEC, NULL, 0,          
	"", HFILL }
    },
  };

  if (proto_scpmac != -1) return;

   assert(proto_scpmac == -1);

  register_init_routine(&scpmac_init);

  proto_scpmac = proto_register_protocol("Scheduled Channel Polling MAC",
					   "SCP-MAC", "scpmac");

   assert(proto_scpmac != -1);

  proto_register_field_array(proto_scpmac, hf, array_length(hf));
  proto_register_subtree_array(ett, array_length(ett));
}

void
scpmac_init(void)
{
  /* n/t */
}

void
proto_reg_handoff_scpmac(void)
{
  data_handle = find_dissector("data");

  register_dissector("scpmac", dissect_scpmac, proto_scpmac);

  register_init_routine(&scpmac_init);
}


