/* packet-smac.c
 * Routines for S-MAC dissection
 * Copyright 2005, Tyler McHenry <smchenry@isi.edu>
 *
 * $Id: packet-smac.c 5194 2006-04-21 00:33:11Z smchenry $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <math.h>
#include <glib.h>

#include <epan/packet.h>
#include <epan/address.h>
#include <epan/conversation.h>
#include <epan/tap.h>
#include "prefs.h"

#include "packet-moteradio.h"
#include "packet-smac.h"
#include "packet-smactest.h"
#include "packet-data.h"

static dissector_handle_t data_handle = NULL;
static dissector_handle_t smactest_handle = NULL;

/* External data structs */
GList* smac_sync_points = NULL;
GHashTable* smac_last_seq_nums = NULL;

/* Internal data structs */

static GList* smac_allocated_memory = NULL;
static guint16 smac_num_sync_packets = 0;
static const guint16 mote_broadcast_address = MOTERADIO_BROADCAST;

/* Forward declaration we need below */

/* External functions */
void proto_reg_handoff_smac(void);
void smac_init(void);

/* Internal functions */
static void smac_set_column_values(tvbuff_t *tvb, packet_info *pinfo, 
				   guint8 *type, guint16 *source);
static void smac_build_tree(tvbuff_t *tvb, proto_tree *tree, guint16 type);
static void smac_call_subdissector(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree);
static void smac_compute_true_seq_num(tvbuff_t *tvb, smac_priv_data* s_priv,
				      guint16 source, guint64 time_now);

/* Initialize the protocol and registered fields */
int proto_smac = -1;
int smac_tap = -1;
static int hf_smac_type = -1;
static int hf_smac_subtype = -1;
static int hf_smac_dest = -1;
static int hf_smac_source = -1;
static int hf_smac_duration = -1;
static int hf_smac_frag_num = -1;
static int hf_smac_position = -1; /* optional (smac_fast_path_schedule)*/
static int hf_smac_timestamp = -1; /* optional (smac_tx_timestamp)*/
static int hf_smac_state = -1;
static int hf_smac_seq_num = -1;
static int hf_smac_syncnode= -1; /* optional (smac_global_schedule) */
static int hf_smac_age = -1; /* optional (smac_global_schedule) */
static int hf_smac_sleep_time = -1;
static int hf_smac_num_sched_neighbors = -1; /* optional (smac_snooper_debug) */
static int hf_smac_tx_data_reset = -1; /* optional (smac_snooper_debug) */
static int hf_smac_tx_sync_reset = -1; /* optional (smac_snooper_debug) */

extern int proto_moteradio;

/* Preferences */

const enum_val_t sync_model[] = {
  {"periods", "Period Count", SYNC_MODEL_PERIODS},
  {"timer", "Independant Timer", SYNC_MODEL_TIMER},
  {NULL, NULL, -1}
};

gboolean smac_fast_path_schedule = FALSE;
gboolean smac_tx_timestamp = FALSE;
gboolean smac_global_schedule = FALSE;
gboolean smac_snooper_debug = FALSE;
gboolean smac_exclude_badcrc = TRUE;
guint8 smac_duty_cycle = SMAC_DUTY_CYCLE_DEFAULT;
guint smac_sync_model = SYNC_MODEL_PERIODS;
guint16 smac_sync_period_length_periods = SMAC_SYNC_PERIOD_LENGTH_DEFAULT_PERIODS;
guint16 smac_sync_period_length_ms = SMAC_SYNC_PERIOD_LENGTH_DEFAULT_MS;

/* Initialize the subtree pointers */
static gint ett_smac = -1;

/* Code to actually dissect the packets */
void
dissect_smac(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
  const mote_priv_data *m_priv;
  smac_priv_data *s_priv;
  guint8 type;
  guint16 source;
  guint64 time_now;

  /* ****** Step 1: Get/Init private data */

  m_priv = (const mote_priv_data*)p_get_proto_data(pinfo->fd, proto_moteradio);

  time_now = m_priv->rx_timestamp;

  if ((s_priv = (smac_priv_data*)p_get_proto_data(pinfo->fd, proto_smac)) == NULL) {

    s_priv = (smac_priv_data*)malloc(sizeof(smac_priv_data));
    smac_allocated_memory = g_list_append(smac_allocated_memory, s_priv);

    s_priv->initialized = FALSE;
    p_add_proto_data(pinfo->fd, proto_smac, (void*)s_priv);
  }

  /* ****** Step 2: Populate Column Data */

  smac_set_column_values(tvb, pinfo, &type, &source);


  /* ****** Step 3: COnversation tracking */

  if (smac_exclude_badcrc == FALSE || (m_priv->crc_error_radio == FALSE && m_priv->crc_error_serial == FALSE)) {

    if (!find_conversation(pinfo->fd->num, &pinfo->src, &pinfo->dst,
			   pinfo->ptype, pinfo->srcport, pinfo->destport, 0)) {
     
      conversation_new(pinfo->fd->num, &pinfo->src, &pinfo->dst, pinfo->ptype,
		       pinfo->srcport, pinfo->destport, 0);
    }
  }

  /* ****** Step 4: Log private data and do any computations that should be executed
     only once per packet. */

  if (s_priv->initialized == FALSE) {
    
    s_priv->type = type;

    if (type == TYPE_SYNC_PKT) {
      s_priv->next_sleep = tvb_get_letohs(tvb, (smac_global_schedule == TRUE) ? 11 : 5);
      
      /* Compute true seq num has the effect of adding the packet to
	 the sync_points table. Don't do this for Bad CRC packets unless
	 the user is feeling particularly stupid */

      if (smac_exclude_badcrc == FALSE || (m_priv->crc_error_radio == FALSE && m_priv->crc_error_serial == FALSE)) 
	smac_compute_true_seq_num(tvb, s_priv, source, time_now);
    }
    
    s_priv->initialized = TRUE;
  }
    

  /* ****** Step 5: Build protocol tree */

  if (tree)
    smac_build_tree(tvb, tree, type);


  /* ****** Step 6: Call subdissector */

  if (type == TYPE_DATA_PKT)
    smac_call_subdissector(tvb, pinfo, tree);


  /* ****** Step 7: Queue packet for tapping */

  if (smac_exclude_badcrc == FALSE || (m_priv->crc_error_radio == FALSE && m_priv->crc_error_serial == FALSE))
    tap_queue_packet(smac_tap, pinfo, NULL);

}


static void smac_set_column_values(tvbuff_t *tvb, packet_info *pinfo, 
				   guint8 *type, guint16 *source)
{
  guint16 dest;

  if (check_col(pinfo->cinfo, COL_PROTOCOL)) 
    col_set_str(pinfo->cinfo, COL_PROTOCOL, "S-MAC");
    
  *type = (tvb_get_guint8(tvb, 0) & 0xF0) >> 4;

  switch(*type) 
    {
    case TYPE_DATA_PKT:
      if (check_col(pinfo->cinfo, COL_INFO)) 
	col_set_str(pinfo->cinfo, COL_INFO, "Application Data");
      break;
    case TYPE_RTS_PKT:
      if (check_col(pinfo->cinfo, COL_INFO)) 
	col_set_str(pinfo->cinfo, COL_INFO, "Request-to-Send");
      break;
    case TYPE_CTS_PKT:
      if (check_col(pinfo->cinfo, COL_INFO)) 
	col_set_str(pinfo->cinfo, COL_INFO, "Clear-to-Send");
      break;
    case TYPE_ACK_PKT:
      if (check_col(pinfo->cinfo, COL_INFO)) 
	col_set_str(pinfo->cinfo, COL_INFO, "Acknowledgement");
      break;
    case TYPE_SYNC_PKT:
      if (check_col(pinfo->cinfo, COL_INFO)) 
	col_set_str(pinfo->cinfo, COL_INFO, "Schedule Synchronization Data");
      break;
    default:
      if (check_col(pinfo->cinfo, COL_INFO)) 
	col_set_str(pinfo->cinfo, COL_INFO, "Unrecognized S-MAC Packet");
    }

  if (*type == TYPE_SYNC_PKT) {
    *source = tvb_get_letohs(tvb, 1);
    dest = 0xFFFF;

    SET_ADDRESS(&pinfo->src, AT_MOTE, 2, tvb_get_ptr(tvb, 1, 2));
    SET_ADDRESS(&pinfo->dl_src, AT_MOTE, 2, tvb_get_ptr(tvb, 1, 2));
    SET_ADDRESS(&pinfo->dst, AT_MOTE, 2, &mote_broadcast_address);
    SET_ADDRESS(&pinfo->dl_dst, AT_MOTE, 2, &mote_broadcast_address);

  } else {
    dest = tvb_get_letohs(tvb, 1);
    *source = tvb_get_letohs(tvb, 3);

    SET_ADDRESS(&pinfo->dst, AT_MOTE, 2, tvb_get_ptr(tvb, 1, 2));
    SET_ADDRESS(&pinfo->dl_dst, AT_MOTE, 2, tvb_get_ptr(tvb, 1, 2));
    SET_ADDRESS(&pinfo->src, AT_MOTE, 2, tvb_get_ptr(tvb, 3, 2));
    SET_ADDRESS(&pinfo->dl_src, AT_MOTE, 2, tvb_get_ptr(tvb, 3, 2));
  }
  
  pinfo->ptype = proto_smac;
  pinfo->srcport = 0;
  pinfo->destport = 0;

  if (check_col(pinfo->cinfo, COL_DEF_DST)) 
    col_add_fstr(pinfo->cinfo, COL_DEF_DST, "%d", dest);

  if (dest == MOTERADIO_BROADCAST) {
    if (check_col(pinfo->cinfo, COL_RES_DST)) 
      col_set_str(pinfo->cinfo, COL_RES_DST, "Broadcast");
  } else {
    if (check_col(pinfo->cinfo, COL_RES_DST)) 
      col_add_fstr(pinfo->cinfo, COL_RES_DST, "Node %d", dest);
  }

  if (check_col(pinfo->cinfo, COL_DEF_SRC)) 
    col_add_fstr(pinfo->cinfo, COL_DEF_SRC, "%d", *source);

  if (check_col(pinfo->cinfo, COL_RES_SRC)) 
    col_add_fstr(pinfo->cinfo, COL_RES_SRC, "Node %d", *source);

}
 
static void smac_compute_true_seq_num(tvbuff_t *tvb, smac_priv_data* s_priv,
				      guint16 source, guint64 time_now)
{
  smac_sync_point_t *sync_point = NULL;
  smac_extended_seq_num_t *last_ext_seq_num = NULL;
  smac_extended_seq_num_t *cur_ext_seq_num = NULL;
  guint8 cur_seq_num;

  smac_num_sync_packets++;
  
  sync_point = (smac_sync_point_t*)malloc(sizeof(smac_sync_point_t));
  sync_point->node = source;
  sync_point->discard = FALSE;

  cur_seq_num = tvb_get_guint8(tvb, 4);
  
  last_ext_seq_num = g_hash_table_lookup(smac_last_seq_nums, &(sync_point->node));
  
  if (last_ext_seq_num == NULL) {
    cur_ext_seq_num = (smac_extended_seq_num_t*)malloc(sizeof(smac_extended_seq_num_t));
    cur_ext_seq_num->seq_num = cur_seq_num;
    cur_ext_seq_num->wraparounds = 0;
    cur_ext_seq_num->total_syncs = 1;
    
    g_hash_table_insert(smac_last_seq_nums, &(sync_point->node), cur_ext_seq_num);
    
  } else {

    cur_ext_seq_num = last_ext_seq_num;
    cur_ext_seq_num->total_syncs++;
    
    if (cur_seq_num < last_ext_seq_num->seq_num)
      cur_ext_seq_num->wraparounds++;
    
    cur_ext_seq_num->seq_num = cur_seq_num;
  }
  
  sync_point->seq_num = cur_ext_seq_num->seq_num + 
    (cur_ext_seq_num->wraparounds * (MAX_SEQ_NUM+1));
  sync_point->sleep_time = time_now + MSEC_TO_USEC(s_priv->next_sleep);
  sync_point->recv_time = time_now;

  smac_sync_points = g_list_append(g_list_first(smac_sync_points), sync_point);
}

static void smac_build_tree(tvbuff_t *tvb, proto_tree *tree, guint16 type)
{
  proto_item *ti = NULL;
  proto_tree *smac_tree = NULL;
  int extraoffset = 0;

  ti = proto_tree_add_item(tree, proto_smac, tvb, 0, -1, FALSE);
  
  smac_tree = proto_item_add_subtree(ti, ett_smac);
  
  proto_tree_add_item(smac_tree,
		      hf_smac_type, tvb, 0, 1, FALSE);
  proto_tree_add_item(smac_tree,
		      hf_smac_subtype, tvb, 0, 1, FALSE);
  
  /* Determine packet type:
     TYPE_DATA_PKT: Normal headers with payload at end
     TYPE_[RTS|CTS|ACK]_PKT: Normal headers without payload
     TYPE_SYNC_PKT: Sync headers without payload
     
     Type is high-order nybble of first byte. Low order
     nybble is subtype and is app-specific (?)
  */
  
  /* Note that for some reason, the S-MAC addresses are little-endian
     i.e. NOT in network byte order! */
  
  if (type == TYPE_SYNC_PKT) {
    
    proto_tree_add_item(smac_tree,
			hf_smac_source, tvb, 1, 2, TRUE);
    proto_tree_add_item(smac_tree,
			hf_smac_state, tvb, 3, 1, FALSE);
    proto_tree_add_item(smac_tree,
			hf_smac_seq_num, tvb, 4, 1, FALSE);
    
    if (smac_global_schedule == TRUE) {
      proto_tree_add_item(smac_tree,
			  hf_smac_syncnode, tvb, 5, 2, TRUE);
      proto_tree_add_item(smac_tree,
			  hf_smac_age, tvb, 7, 4, TRUE);
      extraoffset += 6;
    }
    
    proto_tree_add_item(smac_tree,
			hf_smac_sleep_time, tvb, 5 + extraoffset, 2, TRUE);
    
    if (smac_snooper_debug == TRUE) {
      proto_tree_add_item(smac_tree,
			  hf_smac_num_sched_neighbors, tvb, 7 + extraoffset, 1, FALSE);
      proto_tree_add_item(smac_tree,
			  hf_smac_tx_data_reset, tvb, 8 + extraoffset, 1, FALSE);
      proto_tree_add_item(smac_tree,
			  hf_smac_tx_sync_reset, tvb, 9 + extraoffset, 1, FALSE);

      extraoffset += 3; 
    }
    
  } else {
    
    proto_tree_add_item(smac_tree,
			hf_smac_dest, tvb, 1, 2, TRUE);
    proto_tree_add_item(smac_tree,
			hf_smac_source, tvb, 3, 2, TRUE);
    proto_tree_add_item(smac_tree,
			hf_smac_duration, tvb, 5, 2, TRUE);
    
    if (type == TYPE_DATA_PKT) {
      
      proto_tree_add_item(smac_tree,
			  hf_smac_frag_num, tvb, 7, 1, FALSE);      
      
      if (smac_fast_path_schedule == TRUE) {
	proto_tree_add_item(smac_tree,
			    hf_smac_position, tvb, 8, 2, TRUE);
	extraoffset += 2;
      }
      
      if (smac_tx_timestamp == TRUE) {
	proto_tree_add_item(smac_tree,
			    hf_smac_timestamp, tvb, 10 + extraoffset, 4, TRUE);
	extraoffset += 4;
      }
    }
  }
}

static void smac_call_subdissector(tvbuff_t *tvb, packet_info *pinfo, 
				   proto_tree *tree)
{
  tvbuff_t *next_tvb = NULL;
  guint16 data_length;
  guint16 extraoffset;

  extraoffset = (smac_fast_path_schedule ? 2 : 0) + 
                (smac_tx_timestamp ? 4 : 0);

  /* Sub-dissectors for S-MAC apps would be called here. For
     now treat everything as raw data. */
  
  data_length = tvb_length(tvb) - (SMAC_BASIC_HEADER_LEN + extraoffset);
  
  next_tvb = tvb_new_subset(tvb, (SMAC_BASIC_HEADER_LEN + extraoffset),
			    data_length, -1);
  
  if (data_length == SMACTEST_PACKET_LEN) {
    if (smactest_handle == NULL) 
      smactest_handle = find_dissector("smactest");
      
    call_dissector(smactest_handle, next_tvb, pinfo, tree);

  } else {
    if (data_handle == NULL) 
      data_handle = find_dissector("data");
      
    call_dissector(data_handle, next_tvb, pinfo, tree);
  }
}

/* Register the protocol with Ethereal */

void
proto_register_smac(void)
{                 
  module_t *smac_module;

  static const value_string typevalues[] = {
      { TYPE_DATA_PKT, "Data Packet" },
      { TYPE_RTS_PKT, "RTS Packet" },
      { TYPE_CTS_PKT, "CTS Packet" },
      { TYPE_ACK_PKT, "ACK Packet" },
      { TYPE_SYNC_PKT, "Sync Packet" },
      { 0, NULL },
  };

/* Setup list of header fields  See Section 1.6.1 for details*/
	static hf_register_info hf[] = {
	  { &hf_smac_type,
	    { "Message Type",           "smac.type",
	      FT_UINT8, BASE_DEC, typevalues, 0xf0,          
	      "Type is the major type of the S-MAC packet", HFILL }
	  },
	  { &hf_smac_subtype,
	    { "Message Subtype",           "smac.subtype",
	      FT_UINT8, BASE_DEC, NULL, 0x0f,          
	      "Subtype is the minor type of the S-MAC packet", HFILL }
	  },
	  { &hf_smac_dest,
	    { "Destination Address",           "smac.dest",
	      FT_UINT16, BASE_DEC, NULL, 0,          
	      "Destination is the intended recipient of this packet", HFILL }
	  },
	  { &hf_smac_source,
	    { "Source Address",           "smac.source",
	      FT_UINT16, BASE_DEC, NULL, 0,          
	      "Source is the origin of this packet", HFILL }
	  },
	  { &hf_smac_duration,
	    { "Duration",           "smac.duration",
	      FT_UINT16, BASE_DEC, NULL, 0,          
	      "Duration gives an indication of how long the recipient will have to listen to receive the packet (?)", HFILL }
	  },
	  { &hf_smac_frag_num,
	    { "Fragment Number",           "smac.frag_num",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "(Unknown Meaning!!)", HFILL }
	  },
	  { &hf_smac_position,
	    { "FPS Position",           "smac.position",
	      FT_UINT16, BASE_DEC, NULL, 0,          
	      "(Optional) Position in Fast Path Schedule", HFILL }
	  },
	  { &hf_smac_timestamp,
	    { "Transmission Time",           "smac.timestamp",
	      FT_UINT32, BASE_DEC, NULL, 0,          
	      "(Optional) Time of transmission", HFILL }
	  },
	  { &hf_smac_state,
	    { "State",           "smac.state",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "State gives state information for SYNC messages", HFILL }
	  },
	  { &hf_smac_seq_num,
	    { "Sequence Number",           "smac.seq_num",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "Sequence number represents the position of this packet in a given series of SYNC messages", HFILL }
	  },
	  { &hf_smac_syncnode,
	    { "Global Schedule Initializer",           "smac.syncnode",
	      FT_UINT16, BASE_DEC, NULL, 0,          
	      "(Optional) Node ID of the node whose schedule we are following", HFILL }
	  },
	  { &hf_smac_age,
	    { "Global Schedule Age",           "smac.age",
	      FT_UINT32, BASE_DEC, NULL, 0,          
	      "(Optional) Age of the current global schedule", HFILL }
	  },
	  { &hf_smac_sleep_time,
	    { "Next Sleep Time",           "smac.sleep_time",
	      FT_UINT16, BASE_DEC, NULL, 0,          
	      "Next Sleep Time is the next time (relative to now) that the source node will enter its sleep cycle", HFILL }
	  },
	  { &hf_smac_num_sched_neighbors,
	    { "Number of Schedule Neighbors",           "smac.num_sched_neighbors",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "(Optional) Number of neighbors on the same schedule as this node", HFILL }
	  },
	  { &hf_smac_tx_data_reset,
	    { "Transmission Reset (Data)",           "smac.tx_data_reset",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "(Optional) (Unknown Meaning!!)", HFILL }
	  },
	  { &hf_smac_tx_sync_reset,
	    { "Transmission Reset (Sync)",           "smac.tx_sync_reset",
	      FT_UINT8, BASE_DEC, NULL, 0,          
	      "(Optional) (Unknown Meaning!!)", HFILL }
	  },
	};

/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_smac,
	};

  register_init_routine(&smac_init);

/* Register the protocol name and description */
	proto_smac = proto_register_protocol("Sensor MAC",
	    "S-MAC", "smac");

/* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_smac, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));
        
/* Register preferences module (See Section 2.6 for more on preferences) */       
        smac_module = prefs_register_protocol(proto_smac, NULL);

	prefs_register_bool_preference(smac_module,
				       "fast_path_schedule",
				       "Assume S-MAC Application Data packets have Fast Path Schedule position header",
				       "Whether the dissector should look for the Fast Path Schedule position header in S-MAC Application Data packets",
				       &smac_fast_path_schedule);

	prefs_register_bool_preference(smac_module,
				       "tx_timestamp",
				       "Assume S-MAC Application Data packets have a transmission timestamp",
				       "Whether the dissector should look for the tx timestamp header in S-MAC Application Data packets",
				       &smac_tx_timestamp);

	prefs_register_bool_preference(smac_module,
				       "global_schedule",
				       "Assume S-MAC Schedule Synchronization packets have global schedule headers",
				       "Whether the dissector should look for the global schedule headers in S-MAC Schedule Synchronization packets",
				       &smac_global_schedule);

	prefs_register_bool_preference(smac_module,
				       "snooper_debug",
				       "Assume S-MAC Schedule Synchronization packets have extra debugging headers",
				       "Whether the dissector should look for the snooper debugging headers in S-MAC Schedule Synchronization packets",
				       &smac_snooper_debug);

	prefs_register_bool_preference(smac_module,
				       "exclude_badcrc",
				       "Exclude packets with bad CRCs from S-MAC statistics, conversation, and schedule tracking",
				       "Whether to filter possibly corrupted packets from contributing to aggregate data. This eliminates false nodes (from corruped src and dest fields) at the expense of losing some accuracy for real data",
				       &smac_exclude_badcrc);

	prefs_register_uint_preference(smac_module,
				       "sync_duty_cycle",
				       "Duty Cycle (percentage)",
				       "Sets the relative precentage of time that nodes spend awake as opposed to asleep.",
				       10, &smac_duty_cycle);

	prefs_register_enum_preference(smac_module,
				       "sync_model",
				       "Synchronization frequency determined by",
				       "The S-MAC specification says that sync frequency depends on period count, but the reference implentation depends on an independent timer that corresponds to period count only under default parameters. Choose Period Count if you are using an unmodified reference implementation, or if you have a different implementation that follows the spec. If you are using the reference implementation but have changed the synchronization frequency parameter, you should choose Independant Timer.",
				       &smac_sync_model,
				       sync_model, FALSE);

	prefs_register_uint_preference(smac_module,
				       "sync_period_length_periods",
				       "\"Period Count\" Synchronization Frequency (periods)",
				       "Sets the points at which the schedule analyzer should expect to hear SYNC packets. Leave as default unless you are not using the reference implementation. If you have changed SYNC_PERIOD in the reference implementation, use the Independant Timer model.",
				       10, &smac_sync_period_length_periods);

	prefs_register_uint_preference(smac_module,
				       "sync_period_length_ms",
				       "\"Independant Timer\" Synchronization Frequency (milliseconds)",
				       "Sets the points at which the schedule analyzer should expect to hear SYNC packets. Leave as default unless you have also changed the value of SYNC_PERIOD in your S-MAC application.",
				       10, &smac_sync_period_length_ms);
     

	smac_tap = register_tap("smac");
}


static gboolean smac_hash_destroy_helper(gpointer key, gpointer value, 
					 gpointer userdata) 
{
  /* Free the value (which is a struct on the heap) but leave the key alone
     since it points into the sync_points GList which is deallocated
     seperately later */

  free((smac_extended_seq_num_t*)value);

  return TRUE;
}

void
smac_init(void)
{
  GList* mem_list = NULL;
  smac_sync_point_t *sync_point = NULL;

  for (mem_list = g_list_first(smac_allocated_memory); mem_list; 
       mem_list = g_list_next(mem_list))
    free(mem_list->data);
  
  g_list_free(smac_allocated_memory);

  if (smac_last_seq_nums != NULL) {
    g_hash_table_foreach_remove(smac_last_seq_nums, &smac_hash_destroy_helper, NULL);
    g_hash_table_destroy(smac_last_seq_nums);
  }

 for (mem_list = g_list_first(smac_sync_points); mem_list; 
      mem_list = g_list_next(mem_list)) {
    sync_point = mem_list->data;
    /*    printf("%u\t%u\t%u\t%llu\n", sync_point->node, sync_point->seq_num, 
	  sync_point->eff_seq_num, sync_point->sleep_time); */
    free(mem_list->data);   
  }

  g_list_free(smac_sync_points);

  smac_allocated_memory = NULL;
  smac_sync_points = NULL;
  smac_last_seq_nums = NULL;

  smac_last_seq_nums = g_hash_table_new(&g_int_hash, &g_int_equal);
}

void
proto_reg_handoff_smac(void)
{
  data_handle = find_dissector("data");
  smactest_handle = find_dissector("smactest");

  register_dissector("smac", dissect_smac, proto_smac);

  register_init_routine(&smac_init);
}





