#ifndef _PACKET_SMAC_H_
#define _PACKET_SMAC_H_

#define SMAC_BASIC_HEADER_LEN  8 

#define SMAC_MAX_APPROX_VALS 10
#define SMAC_MIN_SYNC_POINTS  5

#define SMAC_MIN_PERIOD_CORRELATION 0.9999999
#define SMAC_MAX_PERIOD_ITERATIONS  5

/* All time values are in microseconds (usecs) unless indicated otherwise */

#define SEC_TO_USEC(x) x * G_GINT64_CONSTANT(1000000)
#define MSEC_TO_USEC(x) x * G_GINT64_CONSTANT(1000)
#define SMU_TO_USEC(s, ms, us) SEC_TO_USEC(s) + MSEC_TO_USEC(ms) + us

#define NEAREST_GUINT32(f) (guint32)((f) + 0.5)
#define NEAREST_GUINT64(f) (guint64)((f) + 0.5)

#define USEC_TO_SEC(x) NEAREST_GUINT32(((double)x)/1000000.0)

/* Period value by definition of the S-MAC protocol *should*
   be exactly 1200000 (1.2s), but this has proven inaccurate.
   Values below have been determined empirically */

#define SMAC_PERIOD_LENGTH_DEFAULT          1328955

/* Fast Path Schedule and TX Timestamp only apply to data packets
   and thus do not seem to affect the period (like those options
   which apply to SYNC packets */

#define SMAC_PERIOD_SKEW_FAST_PATH_SCHEDULE       0 
#define SMAC_PERIOD_SKEW_TX_TIMESTAMP             0 
#define SMAC_PERIOD_SKEW_GLOBAL_SCHEDULE      31079
#define SMAC_PERIOD_SKEW_SNOOPER_DEBUG        10441


/* Other defaults */

#define SMAC_DUTY_CYCLE_DEFAULT                  10      /* 10% */
#define SMAC_SYNC_PERIOD_LENGTH_DEFAULT_PERIODS  10      /* 10 pds */
#define SMAC_SYNC_PERIOD_LENGTH_DEFAULT_MS       12000   /* 12s */


/* Composite or pre-converted values */

#define SMAC_SYNC_PERIOD_LENGTH  MSEC_TO_USEC((smac_sync_model == SYNC_MODEL_PERIODS) ? smac_sync_period_length_periods : smac_sync_period_length_ms)
#define SMAC_CONTENTION_TIME     NEAREST_GUINT64((smac_duty_cycle / 100.0) * smac_measured_period) 
#define SMAC_TIME_TO_INACTIVE    SMAC_SYNC_PERIOD_LENGTH * 10

/* Protocol-defined consntats that cannot be modified at runtime */

#define SMAC_SYNC_INTERVAL_MS    NEAREST_GUINT64((SMAC_CONTENTION_TIME / 100.0) * 10.0)
#define SMAC_RTS_CTS_BCAST_INTERVAL_MS SMAC_CONTENTION_TIME - SMAC_SYNC_INTERVAL_MS

/* Other constants that cannot be modified at runtime in any way */

#define SMAC_TIMING_TOLERANCE           65000  /* 65 ms tolerance */
#define SMAC_PHASE_DRIFT_TOLERANCE      1.005  /* 110% of min delta */
#define SMAC_SYNC_DELTA_SEQ_TOLERANCE   3     /* 3 missed seq nums */
#define SMAC_SYNC_ANOMALY_THRESHOLD     75    /* Percentage, see below */
#define SMAC_SYNC_ANOMALY_PTS_NEEDED    10    /* See below */
#define SMAC_SYNC_ANOMALY_PTS_RESET     20    /* See below */

/* Anomalies:
 *
 * When nodes adopt a new schedule, their sleep time deltas get way too
 * short. To distinguish this anomalous shortness from a truly reduced
 * sync period, first regard all sync deltas that are below
 * SMAC_SYNC_ANOMALY_THRESHOLD% of the "default" sleep delta (which varies
 * depending on which sync freq model is used). Then, if we accumulate 
 * more than SMAC_SYNC_ANOMALY_PTS_NEEDED of them without a gap of more than
 * SMAC_SYNC_ANOMALY_PTS_RESET points between any two, start counting the
 * anomalies as real data from that point onward.
 *
 */

/* Values for the "type" field. */

enum { TYPE_DATA_PKT, 
       TYPE_RTS_PKT, 
       TYPE_CTS_PKT, 
       TYPE_ACK_PKT, 
       TYPE_SYNC_PKT,
       TYPE_UNKNOWN };

enum { SYNC_MODEL_PERIODS,
       SYNC_MODEL_TIMER };

typedef struct _smac_hdr {
  guint16  src;
  guint16  dst;
} smac_hdr;


typedef struct _smac_sync_point {
  gint node; /* This should really be a guint16 but it messes up the hash keys
		for the sequence number wrapping fix if it's not a gint! */
  guint32 seq_num;
  guint32 eff_seq_num;
  guint64 sleep_time;
  guint64 recv_time;
  gboolean discard; /* Do not use in linreg for whatever reason */
} smac_sync_point_t;

typedef struct _smac_priv_data {
  gboolean initialized;
  guint8 type;
  guint8 next_sleep; /* Ignored if type != TYPE_SYNC_PKT */
} smac_priv_data;

typedef struct _smac_extended_seq_num {
  guint8  seq_num;
  guint16 wraparounds;
  guint32 total_syncs;
} smac_extended_seq_num_t;

#define MAX_SEQ_NUM 255

void 
dissect_smac(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree);

/* Global values that other modules including this header
   can access */

extern int proto_smac;
extern int smac_tap;

extern gboolean smac_fast_path_schedule;
extern gboolean smac_tx_timestamp;
extern gboolean smac_global_schedule;
extern gboolean smac_snooper_debug;
extern gboolean smac_exclude_badcrc;
extern guint8 smac_duty_cycle;
extern guint smac_sync_model;
extern guint16 smac_sync_period_length_periods;
extern guint16 smac_sync_period_length_ms;

extern GList* smac_sync_points;
extern GHashTable* smac_last_seq_nums;

#endif
