/* smac_schedules.c
 * S-MAC schedules addition for ethereal
 *
 * $Id: smac_schedules.c 5883 2006-10-30 19:36:22Z smchenry $
 *
 * Copyright 2005, University of Southern California
 * By Tyler McHenry <smchenry@isi.edu>
 *
 * based on voip_calls.c
 * Copyright 2004, Ericsson , Spain
 * By Francisco Alcoba <francisco.alcoba@ericsson.com>
 *
 * based on h323_calls.c
 * Copyright 2004, Iskratel, Ltd, Kranj
 * By Miha Jemec <m.jemec@iskratel.si>
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation,  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <glib.h>
#include <assert.h>

#include <epan/packet.h>
#include <epan/epan_dissect.h>
#include "util.h"
#include <epan/tap.h>
#include "register.h"

#include "smac_schedules.h"
#include "smac_visualization.h"
#include <epan/dissectors/packet-moteradio.h>
#include <epan/dissectors/packet-smac.h>

extern void protect_thread_critical_region(void);
extern void unprotect_thread_critical_region(void);

/* External functions */

const smac_schedules_tapinfo_t* smac_schedules_get_info(void);
void smac_schedules_reset(smac_schedules_tapinfo_t *tapinfo);
void smac_schedules_remove();
void smac_schedules_reanalyze_loose_packets();
void smac_schedules_check_inactive();
int smac_schedules_packet(void *t, packet_info *pinfo, 
			  epan_dissect_t *edt, void *priv);
gboolean smac_addr_is_broadcast(address* addr);

/* Internal functions */

static guint64 smac_schedules_relative_phase(guint64 time_now, guint64 period);
static guint64 smac_schedules_phase(guint64 time_now);
static smac_node_info_t* smac_find_existing_node(GList* master_list, address* node_addr);
static smac_schedule_info_t* smac_find_existing_schedule(packet_info* pinfo,
							 smac_priv_data* s_priv,
							 mote_priv_data* m_priv);
static smac_schedule_info_t* smac_find_schedule_for_contention_pkt(GList *schedules_list, 
								   guint64 time_now,
								   smac_priv_data *s_priv);
static gboolean smac_analyze_packet(packet_info *pinfo, mote_priv_data *m_priv, 
				    smac_priv_data *s_priv, gboolean reset);
static gboolean smac_analyze_collision(packet_info *pinfo, mote_priv_data *m_priv,
                                       smac_priv_data *s_priv);
static void smac_record_collision (smac_collisions* collisions, packet_info *pinfo,
                                   guint8 type);
static gboolean smac_test_collision_sync (packet_info *pinfo, mote_priv_data *m_priv,
                                          smac_schedule_info_t *schedule);
static gboolean smac_test_collision_rcb (packet_info *pinfo, mote_priv_data *m_priv,
                                         smac_schedule_info_t *schedule);
static gboolean smac_test_collision_data (packet_info *pinfo, mote_priv_data *m_priv,
                                          smac_schedule_info_t *schedule);
static void smac_get_node_from_hash(gpointer key, gpointer value, gpointer node_list);
static guint64 smac_compute_default_period();
static guint64 smac_compute_period();
static smac_sync_point_t* smac_find_greatest_sleep_time(const GList * const points_list, guint16 node, guint32 seq_num);
static smac_sync_point_t* smac_find_least_sleep_time(const GList * const points_list, guint16 node, guint32 seq_num);
static gboolean smac_is_anomaly(smac_sync_point_t *point_delta);
static guint64 smac_find_min_delta_sleep(GList* sync_points);
static unsigned int smac_compute_periods_per_sync(GList *node_list, GList* sync_points,
						  guint64 min_delta_sleep);
static guint32 smac_compute_num_schedules(GList *sync_points, guint64 min_delta_sleep, 
					  guint64 period);
static guint64 smac_moving_approx(guint64 cur_val, guint64 new_val, 
				  guint16 num_vals, guint16 max_vals);
static smac_sync_point_t* smac_compute_sync_delta(GList *sync_points,
						  smac_sync_point_t* sync_point);
static void smac_compute_all_effective_seq_nums(GList *sync_points, guint64 min_delta_sleep, guint64 period);
static void smac_compute_effective_seq_num(GList *sync_points, 
					   smac_sync_point_t* sync_point,
					   guint64 min_delta_sleep, guint64 period);
static smac_estimate_t smac_compute_estimate_period(GList* sync_points, guint16 node,
						    gboolean use_effectives);



/* Global data */

smac_schedules_tapinfo_t the_tapinfo_struct = EMPTY_TAPINFO_STRUCT;
static gboolean reset_cnt = FALSE;
guint64 smac_measured_period = SMAC_PERIOD_LENGTH_DEFAULT;
int smac_vis_tap = -1;

const smac_schedules_tapinfo_t* smac_schedules_get_info(void)
{
    return &the_tapinfo_struct;
}

void smac_schedules_reset(smac_schedules_tapinfo_t *tapinfo)
{
    GList* list;
    GList* list2;
    smac_schedule_info_t* schedule;
    smac_node_info_t* node;
                                                                                                                   
    /* free the data items first */

    for (list = g_list_first(tapinfo->schedules_list); list;
         list = g_list_next(list)) {
        schedule = list->data;
        g_list_free(schedule->nodes_list);

        for (list2 = g_list_first(schedule->broadcast_packets_list); list2;
             list2 = g_list_next(list2))
	    g_free(list2->data);
        g_list_free(schedule->broadcast_packets_list);

        g_free(schedule);
    }
    g_list_free(tapinfo->schedules_list);

    for (list = g_list_first(tapinfo->nodes_list); list;
         list = g_list_next(list)) {
        node = list->data;
        g_list_free(node->schedules_list);

        for (list2 = g_list_first(node->status_list); list2;
             list2 = g_list_next(list2))
	    g_free(list2->data);
        g_list_free(node->status_list);

        g_free(node);
    }

    g_list_free(tapinfo->nodes_list);

    for (list = g_list_first(tapinfo->loose_packets_list); list;
         list = g_list_next(list))
        g_free(list->data);

    g_list_free(tapinfo->loose_packets_list);
    
    for (list = g_list_first(tapinfo->collision_packets_list); list;
         list = g_list_next(list))
        g_free(list->data);
    
    g_list_free(tapinfo->collision_packets_list);
    
    tapinfo->num_packets = 0;
    tapinfo->num_sync_packets = 0;
    tapinfo->time_zero = 0;
    tapinfo->time_now = 0;
    tapinfo->schedules_list = NULL;
    tapinfo->nodes_list = NULL;
    tapinfo->loose_packets_list = NULL;
    tapinfo->collisions.num_total = 0;
    tapinfo->collisions.num_sync = 0;
    tapinfo->collisions.num_data = 0;
    tapinfo->collision_packets_list = NULL;
    tapinfo->dummy = 0;

    reset_cnt = TRUE;

    smac_measured_period = smac_compute_period();

    return;
}


void smac_schedules_remove()
{
    protect_thread_critical_region();
    remove_tap_listener(&(the_tapinfo_struct.dummy));
    unprotect_thread_critical_region();
}

gboolean smac_addr_is_broadcast(address* addr)
{
    address tmp_addr;
    const guint16 broadcast = MOTERADIO_BROADCAST;

    SET_ADDRESS(&tmp_addr, AT_MOTE, 2, (void*)(&broadcast));

    return ADDRESSES_EQUAL(addr, &tmp_addr) ? TRUE : FALSE;
}

static guint64 smac_schedules_relative_phase(guint64 time_now, guint64 period) 
{
    smac_schedules_tapinfo_t *tapinfo = &the_tapinfo_struct;
    return (((guint64)time_now) - ((guint64)tapinfo->time_zero)) % ((guint64)period);
}
static guint64 smac_schedules_phase(guint64 time_now) 
{
    return smac_schedules_relative_phase(time_now, smac_measured_period);
}

void smac_schedules_reanalyze_loose_packets() 
{
    smac_schedules_tapinfo_t *tapinfo = &the_tapinfo_struct;
    smac_loose_packet_t *loose_packet = NULL;
    GList *list = g_list_first(tapinfo->loose_packets_list);

    tapinfo->loose_packets_list = NULL;

    for (list = g_list_first(list); list; list = g_list_next(list)) {
        loose_packet = list->data;
        if (smac_analyze_packet(&(loose_packet->pinfo), loose_packet->m_priv,
                                loose_packet->s_priv, FALSE))
            free(loose_packet);
    }

    g_list_free(list);
}

void smac_schedules_check_inactive() 
{
    smac_schedules_tapinfo_t *tapinfo = &the_tapinfo_struct;
    GList* nodes_list = tapinfo->nodes_list;
    GList* status_list = NULL;
    smac_node_info_t* node;
    smac_schedule_nodestatus_t* status;

    for (nodes_list = g_list_first(nodes_list); nodes_list;
         nodes_list = g_list_next(nodes_list)) {

        node = (smac_node_info_t*)(nodes_list->data);

        for (status_list = g_list_first(node->status_list); status_list;
             status_list = g_list_next(status_list)) {

            status = (smac_schedule_nodestatus_t*)(status_list->data);

            if (status->active == TRUE) {
                if (status->last_sync < (tapinfo->time_now - SMAC_TIME_TO_INACTIVE)) {
                    status->active = FALSE;
                    status->schedule->active_nodes--;
                }
            }
        }
    }
}

static smac_node_info_t* smac_find_existing_node(GList* master_list, address* node_addr)
{
    GList *list = NULL;
    smac_node_info_t *tmp_listinfo = NULL;

    if (smac_addr_is_broadcast(node_addr)) return NULL;

    for (list = g_list_first(master_list); list; list = g_list_next(list)) {
        tmp_listinfo = (smac_node_info_t*)(list->data);

        if (ADDRESSES_EQUAL(node_addr, &(tmp_listinfo->addr)))
            return (smac_node_info_t*)(list->data);
    }

    return NULL;
}

static smac_schedule_info_t* smac_find_existing_schedule(packet_info* pinfo,
							 smac_priv_data* s_priv,
							 mote_priv_data* m_priv)
{
    smac_schedules_tapinfo_t *tapinfo = &the_tapinfo_struct;
    GList *schedules_list = g_list_first(tapinfo->schedules_list);
    GList *nodes_list = g_list_first(tapinfo->nodes_list);
    GList *tmp_list = NULL;
    GList *tmp_list2 = NULL;
    GList *common_schedules_list = NULL;
    smac_node_info_t *src_node = smac_find_existing_node(tapinfo->nodes_list, &(pinfo->src));
    smac_node_info_t *dst_node = smac_find_existing_node(tapinfo->nodes_list, &(pinfo->dst));
    smac_schedule_info_t *tmp_schedule = NULL;
    smac_node_info_t *tmp_node = NULL;
    guint32 *tmp_number = NULL;
    const int broadcast = MOTERADIO_BROADCAST;
    address tmp_addr;
    guint64 time_now = m_priv->rx_timestamp;
    guint64 known_phase;
    guint64 cur_phase;
    int i = 0;
    static int cnt = 0;

    if(reset_cnt) cnt = 0;

    /* Assumption: Nodes may send on many schedules, but only listen on one schedule */

    if (s_priv->type == TYPE_SYNC_PKT) {

        /* If it's a sync packet, the phase should match some known phase
           time within a tolerance. */

        cnt++;
        cur_phase = smac_schedules_phase(time_now + MSEC_TO_USEC(s_priv->next_sleep)); 

        for (tmp_list = g_list_first(schedules_list); tmp_list; tmp_list = g_list_next(tmp_list)) {
            known_phase = ((smac_schedule_info_t*)(tmp_list->data))->phase;

            if (llabs(cur_phase - known_phase) < SMAC_TIMING_TOLERANCE ||
                llabs(cur_phase - (smac_measured_period - known_phase)) < SMAC_TIMING_TOLERANCE)
                return (smac_schedule_info_t*)(tmp_list->data);
        }

        /*
          printf("\nNo known schedule for packet from node %u at time %llu with sleep time %llu and phase %llu. Existing schedules:\n\n",
          NODE_ID(src_node), time_now, 
          time_now + MSEC_TO_USEC(s_priv->next_sleep), cur_phase);
          for (tmp_list = g_list_first(schedules_list); tmp_list; tmp_list = g_list_next(tmp_list)) {
          printf("Schedule %u\tPhase %llu\tNodes:", 
          ((smac_schedule_info_t*)(tmp_list->data))->id, 
          ((smac_schedule_info_t*)(tmp_list->data))->phase);
          for (tmp_list2 = g_list_first(((smac_schedule_info_t*)(tmp_list->data))->nodes_list);
          tmp_list2; tmp_list2 = g_list_next(tmp_list2))
          printf(" %u", NODE_ID((smac_node_info_t*)(tmp_list2->data));
          printf("\n");
          }
        */

    } else if (s_priv->type == TYPE_DATA_PKT && smac_addr_is_broadcast(&(pinfo->dst))) {
  
        /* If it's a broadcast packet, we should land sometime within
           SMAC_CONTENTION_TIME before a known phase within a tolerance. */

        tmp_schedule = smac_find_schedule_for_contention_pkt(schedules_list, time_now, s_priv);
    
        if (tmp_schedule != NULL) {
            tmp_number = (guint32*)malloc(sizeof(guint32));
      
            *tmp_number = pinfo->fd->num;
            tmp_schedule->broadcast_packets_list = g_list_append(tmp_schedule->broadcast_packets_list, 
                                                                 tmp_number);
        }
    
        return tmp_schedule;  

    } else {  
    
        /* If it's anything else it's a unicast packet of some sort, which means it is
           necessarily on the receiving node's schedule. (for DATA/RTS/CTS receiving node
           is dest, for ACK, receiving node is source) */

        if (s_priv->type == TYPE_ACK_PKT) 
            tmp_list = g_list_first(src_node->schedules_list);
        else {
            if (!dst_node && !smac_addr_is_broadcast(&(pinfo->dst))) {
                dst_node = (smac_node_info_t*)g_malloc(sizeof(smac_node_info_t));
                tapinfo->nodes_list = g_list_append(tapinfo->nodes_list,  dst_node);
	
                COPY_ADDRESS(&(dst_node->addr), &(pinfo->dst));
                dst_node->schedules_list = NULL;
                dst_node->status_list = NULL;

            } else if (dst_node) {
                /* If we just created dst_node it will have no schedules! */
                tmp_list = g_list_first(dst_node->schedules_list);     
            }
        }

        if (tmp_list) 
            return smac_find_schedule_for_contention_pkt(tmp_list, time_now, s_priv);

    }

    return NULL;
}

static smac_schedule_info_t* smac_find_schedule_for_contention_pkt(GList *schedules_list, 
								   guint64 time_now,
								   smac_priv_data *s_priv)
{
    GList *list;
    smac_schedule_info_t *best_schedule = NULL;
    smac_schedule_info_t *cur_schedule = NULL;
    guint64 best_fit_metric = 0;
    guint64 fit_metric = 0;
    guint64 cur_phase = 0;
    guint64 known_phase = 0;
    guint64 phase_diff_fwd = 0;
    guint64 phase_diff_bkwd = 0;

    for (list = g_list_first(schedules_list); list; list = g_list_next(list)) {
        cur_schedule = list->data;
    
        known_phase = cur_schedule->phase;
        cur_phase = smac_schedules_phase(time_now + MSEC_TO_USEC(s_priv->next_sleep));

        phase_diff_fwd = llabs(cur_phase - known_phase); 
        phase_diff_bkwd = llabs(cur_phase - (smac_measured_period - known_phase));

        if (phase_diff_bkwd < SMAC_TIMING_TOLERANCE ||
            phase_diff_fwd < SMAC_CONTENTION_TIME + SMAC_TIMING_TOLERANCE) {
            best_schedule = cur_schedule;
            break;
        } else {
            fit_metric = MIN(phase_diff_bkwd - SMAC_TIMING_TOLERANCE,
                             phase_diff_fwd - SMAC_CONTENTION_TIME - SMAC_TIMING_TOLERANCE);
            if (fit_metric < best_fit_metric || best_fit_metric == 0) {
                best_schedule = cur_schedule;
                best_fit_metric = fit_metric;
            }
        }
    }

#ifdef SMAC_CONTENTION_PACKETS_EXACT
    if (best_fit_metric != 0) return NULL;
#endif

    return best_schedule;
}

int smac_schedules_packet(void *t, packet_info *pinfo, 
			  epan_dissect_t *edt, void *priv)
{
    /* Processing goes here */

    static gboolean first_packet = TRUE;
    smac_schedules_tapinfo_t *tapinfo = &the_tapinfo_struct;
    gboolean reset_schedule_num = FALSE;
    const smac_priv_data* s_priv = NULL;
    const mote_priv_data* m_priv = NULL;

    if (reset_cnt == TRUE) {
        reset_cnt = FALSE;
        first_packet = TRUE;
    }

    if  ((m_priv = (mote_priv_data*)p_get_proto_data(pinfo->fd, proto_moteradio)) == NULL) {
        /* printf("Rejecting packet: No MoteRadio data.\n"); */
        return 0; /* No useful data here! */
    }

    /* Partial collisions should have a valid (non-NULL) s_priv */
    
    s_priv = (smac_priv_data*)p_get_proto_data(pinfo->fd, proto_smac);
    tapinfo->time_now = m_priv->rx_timestamp;
    
    if (m_priv->initialized && m_priv->collision) {
        /* printf("Packet is collision!\n"); */
        smac_analyze_collision(pinfo, m_priv, s_priv);
    }

    if (s_priv == NULL) {
        /* printf("Cannot analyze packet: No S-MAC data.\n"); */
        return 0; /* No useful data here! */
    }
    
    /* If it is a non- or partial collision, perform further analysis */

    /* printf("Analyzing packet.\n"); */
    smac_visualization_packet(t, pinfo, edt, priv);
        
    if (first_packet == TRUE) {
        tapinfo->time_zero = tapinfo->time_now;
        reset_schedule_num = TRUE;
        first_packet = FALSE;
    }
    
    smac_analyze_packet(pinfo, m_priv, s_priv, reset_schedule_num);
    
    return 1;

}

static gboolean smac_lone_active_node(smac_node_info_t* node)
{
    GList* tmp_list;
    smac_schedule_info_t* primary_schedule = NULL;
    smac_node_info_t* test_node = NULL;
    smac_schedule_nodestatus_t* test_status = NULL;

    if (!node || !(node->schedules_list)) return TRUE;

    primary_schedule = g_list_last(node->schedules_list)->data;
  
    for (tmp_list = g_list_first(primary_schedule->nodes_list);
         tmp_list; tmp_list = g_list_next(tmp_list)) {

        test_node = tmp_list->data;
        if (test_node == node) continue;

        if (!(test_node->status_list)) continue;
        test_status = g_list_last(test_node->status_list)->data;

        if (test_status->schedule->id == primary_schedule->id) {
	
            if(test_status->active)  
                return FALSE;
     
            continue;
        }
    }

    return TRUE;

}

static gboolean smac_analyze_packet(packet_info *pinfo, mote_priv_data *m_priv, 
				    smac_priv_data *s_priv, gboolean reset)
{
    static int next_schedule_num = 0;
    smac_schedules_tapinfo_t *tapinfo = &the_tapinfo_struct;
    smac_schedule_info_t *schedule = NULL;
    smac_schedule_info_t *test_schedule = NULL;
    smac_schedule_nodestatus_t *status = NULL;
    smac_node_info_t *node = NULL;
    smac_loose_packet_t *loose_packet = NULL;
    GList* tmplist = NULL;
    guint64 time_now;
    guint64 cur_phase;
    guint64 delta_phase;

    if (reset) next_schedule_num = 0;

    time_now = m_priv->rx_timestamp;

    if ((node = smac_find_existing_node(tapinfo->nodes_list, &(pinfo->src))) == NULL) {
        node = (smac_node_info_t*)g_malloc(sizeof(smac_node_info_t));
        tapinfo->nodes_list = g_list_append(tapinfo->nodes_list,  node);
    
        COPY_ADDRESS(&(node->addr), &(pinfo->src));
        node->schedules_list = NULL;
        node->status_list = NULL;
    }

    /* According to the S-MAC protocol design, the only time a node will change 
       schedule is if:

       1. It is the only active node on its schedule AND
       2. It is changing to a known schedule

       Otherwise we can assume that all syncs are for the initial schedule of a 
       given node.
    */

    if (node->schedules_list)
        schedule = (smac_schedule_info_t*)(g_list_last(node->schedules_list)->data);    

    test_schedule = smac_find_existing_schedule(pinfo, s_priv, m_priv);

    if (!(node->schedules_list) || 
        (test_schedule && smac_lone_active_node(node)))
        schedule = test_schedule;

  

    if (schedule == NULL) {
  
        if (s_priv->type != TYPE_SYNC_PKT) {
            loose_packet = (smac_loose_packet_t*)malloc(sizeof(smac_loose_packet_t));
            loose_packet->pinfo = *pinfo;
            loose_packet->m_priv = m_priv;
            loose_packet->s_priv = s_priv;
            tapinfo->loose_packets_list = g_list_append(tapinfo->loose_packets_list,
                                                        loose_packet);

            return FALSE; /* Don't start schedules on non-SYNC packets. Too unreliable. */
        }
        /*
          if(node->status_list) {
          tmplist = g_list_append(tmplist, ((smac_schedule_nodestatus_t*)(g_list_last(node->status_list)))->schedule);
          if(!smac_find_schedule_for_contention_pkt(tmplist, time_now, s_priv)) 
          return FALSE; 
          }
        */

        tapinfo->num_sync_packets++;

        schedule = (smac_schedule_info_t*)g_malloc(sizeof(smac_schedule_info_t));
        tapinfo->schedules_list = g_list_append(tapinfo->schedules_list,  schedule);

        schedule->id = next_schedule_num++;
        schedule->total_nodes = 1;
        schedule->active_nodes = 1;
        schedule->origin_node = *((guint16*)(pinfo->src.data));
        schedule->packets = 1;
        schedule->bytes = pinfo->fd->pkt_len;
        schedule->periods = 0;
        schedule->sync_packets = 1;
        schedule->first_sync_s = pinfo->fd->rel_secs;
        schedule->first_sync_us = pinfo->fd->rel_usecs;
        schedule->last_sync_s = pinfo->fd->rel_secs;
        schedule->last_sync_us = pinfo->fd->rel_usecs;
        schedule->phase = smac_schedules_phase(time_now + MSEC_TO_USEC(s_priv->next_sleep));
        schedule->nodes_list = NULL;
        schedule->broadcast_packets_list = NULL;

        schedule->nodes_list = g_list_append(schedule->nodes_list, node);     
        node->schedules_list = g_list_append(node->schedules_list, schedule); 

        status = (smac_schedule_nodestatus_t*)g_malloc(sizeof(smac_schedule_nodestatus_t));
        status->schedule = schedule;
        status->last_sync = time_now;
        status->active = TRUE;

        node->status_list = g_list_append(node->status_list, status);

    } else {

        if (!(s_priv->type == TYPE_DATA_PKT && smac_addr_is_broadcast(&(pinfo->dst)))) {
            for (tmplist = g_list_first(node->schedules_list); tmplist != NULL;
                 tmplist = g_list_next(tmplist)) {
                if (((smac_schedule_info_t*)(tmplist->data))->id == schedule->id) break;
            }
      
            if (node->schedules_list == NULL) {
                schedule->total_nodes++;
                schedule->nodes_list = g_list_append(schedule->nodes_list, node);     
                node->schedules_list = g_list_append(node->schedules_list, schedule); 
                status = (smac_schedule_nodestatus_t*)g_malloc(sizeof(smac_schedule_nodestatus_t));
                status->schedule = schedule;
                status->active = FALSE; /* Will be set to true below */
                node->status_list = g_list_append(node->status_list, status);
            } 
        }

        if (s_priv->type == TYPE_SYNC_PKT) {

            for (tmplist = g_list_first(node->status_list); status == NULL && tmplist != NULL;
                 tmplist = g_list_next(tmplist)) {
                if (((smac_schedule_nodestatus_t*)(tmplist->data))->schedule->id == schedule->id)
                    status = ((smac_schedule_nodestatus_t*)(tmplist->data));
            }  

            if (!status) return FALSE;
            /* assert(status != NULL); */

            status->last_sync = time_now;
    
            if(status->active == FALSE) {
                status->active = TRUE;
                status->schedule->active_nodes++;
            }

            tapinfo->num_sync_packets++;
            schedule->sync_packets++;
            schedule->last_sync_s = pinfo->fd->rel_secs;
            schedule->last_sync_us = pinfo->fd->rel_usecs;
            /*            schedule->phase = smac_moving_approx(schedule->phase, 
                          smac_schedules_phase(time_now + MSEC_TO_USEC(s_priv->next_sleep)),
                          schedule->sync_packets, 
                          SMAC_MAX_APPROX_VALS); 
            */

            cur_phase = smac_schedules_phase(time_now + MSEC_TO_USEC(s_priv->next_sleep));  
            delta_phase = (cur_phase > schedule->phase) ? (cur_phase - schedule->phase) : (schedule->phase - cur_phase);
            if (delta_phase <= SMAC_TIMING_TOLERANCE/2) {
                /*	printf("%llu\t%u\t%llu\t%llu\t%llu\n", time_now, schedule->id,
                        schedule->phase, cur_phase, delta_phase);
                */
                schedule->phase = cur_phase;
            }



        }    

        schedule->packets++;
        schedule->bytes += pinfo->fd->pkt_len;
        schedule->periods = (guint16)((time_now - tapinfo->time_zero) / smac_measured_period);

    }


    return TRUE;
  
}

static gboolean smac_analyze_collision(packet_info *pinfo, mote_priv_data *m_priv, smac_priv_data *s_priv)
{
    smac_schedules_tapinfo_t *tapinfo = &the_tapinfo_struct;
    smac_loose_packet_t *collision_pkt = (smac_loose_packet_t*)(malloc(sizeof(smac_loose_packet_t)));
    smac_schedule_info_t *schedule, *cur_schedule;
    GList* list = NULL;
    gint votes_sync = 0;
    gint votes_rcb = 0;
    gint votes_data = 0;
    gboolean hit = FALSE;
    gboolean multiple_schedules = FALSE;
    guint8 type;
    
    collision_pkt->pinfo = *pinfo; /* COPY */
    collision_pkt->m_priv = m_priv;
    collision_pkt->s_priv = NULL;
    
    /* First simply record the presence of the collision */

    tapinfo->collision_packets_list = g_list_append(tapinfo->collision_packets_list,
                                                    collision_pkt);

    
    /* Then try to determine what kind of a collision this is, what kind of
       packet it probably was, and what schedule it was on (if possible) */
    
    if ((m_priv->meta_flags & MOTERADIO_FLAG_RSSI_DATA) == 0) {

        /* Partial collision */

        if (s_priv != NULL) { /* should always be true */ 

            schedule = smac_find_existing_schedule(pinfo, s_priv, m_priv);

            if (schedule != NULL) {
                /* printf("Calling SRC with sched 1 coll addr 0x%X\n", &(schedule->collisions)); */
                smac_record_collision(&(schedule->collisions), pinfo, s_priv->type);
            }

            /* printf("Calling SRC with tap 1 coll addr 0x%X\n", &(tapinfo->collisions)); */
            smac_record_collision(&(tapinfo->collisions), pinfo, s_priv->type);

        } else {

            /* printf("Calling SRC with tap 2 coll addr 0x%X\n", &(tapinfo->collisions)); */
            smac_record_collision(&(tapinfo->collisions), pinfo, TYPE_UNKNOWN);
        }
            
    } else {

        /* Total collision */

        schedule = NULL;
        
        for (list = g_list_first(tapinfo->schedules_list); list;
             list = g_list_next(list)) {

            if (list == NULL) {
                printf("wtf list is null. :( \n");
                break;
            }
            else printf("List is not null, list is 0x%X\n", list);

            hit = TRUE;
            cur_schedule = (smac_schedule_info_t*)(list->data);

            if (smac_test_collision_sync(pinfo, m_priv, cur_schedule)) {
                printf("+1 sync vote\n");
                votes_sync++;
            } else if (smac_test_collision_rcb(pinfo, m_priv, cur_schedule)) {
                printf("+1 rcb vote\n");
                votes_rcb++;
            } else if (smac_test_collision_data(pinfo, m_priv, cur_schedule)) {
                printf("+1 data vote\n");
                votes_data++;
            } else {
                hit = FALSE;
            }

            if (hit) {
                if (!multiple_schedules) {
                    if (!schedule)
                        schedule = cur_schedule;
                    else {
                        multiple_schedules = TRUE;
                        printf("Multiple schedules detected!\n");
                    }
                }
            }
        }

        if (votes_sync > votes_rcb && votes_sync > votes_data)
            type = TYPE_SYNC_PKT;
        else if (votes_rcb > votes_sync && votes_rcb > votes_data)
            type = TYPE_RTS_PKT;
        else if (votes_data > votes_sync && votes_data > votes_rcb)
            type = TYPE_ACK_PKT; /* circumvent the test for broadcast address */
        else
            type = TYPE_UNKNOWN;

        /* printf("Calling SRC with tap 3 coll addr 0x%X\n", &(tapinfo->collisions)); */
        smac_record_collision(&(tapinfo->collisions), pinfo, type);

        if (hit && schedule && !multiple_schedules) {
            /* printf("Calling SRC with sched 2 coll addr 0x%X\n", &(schedule->collisions)); */
            smac_record_collision(&(schedule->collisions), pinfo, type);
        }
    }

    printf("Analyzed a collision. Current totals: \n");
    printf("Total:\t %d \n", tapinfo->collisions.num_total);
    printf("Sync:\t %d \n", tapinfo->collisions.num_sync);
    printf("RCB:\t %d \n", tapinfo->collisions.num_rcb);
    printf("Data:\t %d \n", tapinfo->collisions.num_data);
    printf("\n");
    
    return TRUE;
}

static void smac_record_collision (smac_collisions* collisions, packet_info *pinfo,
                                   guint8 type)
{
    if (!collisions || !pinfo) return;

    /* printf("In SRC with addr 0x%X\n", collisions); */
    
    collisions->num_total++;
    
    switch (type) {
        
        case TYPE_DATA_PKT:
            if (smac_addr_is_broadcast(&(pinfo->dst)))
                collisions->num_rcb++;
            else {
                case TYPE_ACK_PKT:
                    collisions->num_data++;
            }
            break;
            
        case TYPE_RTS_PKT:
        case TYPE_CTS_PKT:
            collisions->num_rcb++;
            break;
            
        case TYPE_SYNC_PKT:
            collisions->num_sync++;
            break;
            
        default:
            break;
    };
}

static gboolean smac_test_collision_sync (packet_info *pinfo, mote_priv_data *m_priv,
                                          smac_schedule_info_t *schedule)
{
    smac_schedules_tapinfo_t *tapinfo = &the_tapinfo_struct;    
    guint64 phase, known_phase;

    phase = smac_schedules_phase(tapinfo->time_now);
    known_phase = schedule->phase;

    /* Return true IFF this packet arrived in the following interval:

       . . . Sleep ][  Sync Slots  ][ RTS/CTS/Bcast Contention ][ Data . . .
                  ^^^^^^^^^^^^^^^^^^^^
                      (Target Area)

      Where the bits outside of Sync Slots is equal to SMAC_TIMING_TOLERANCE
      on either side.
    */
    
    /* printf("Testing for SYNC p = %llu , kp = %llu , dp = %llu\n", 
       phase, known_phase, (phase > known_phase) ? phase - known_phase : known_phase - phase); */
           
    if (phase > known_phase)
        return ((phase - known_phase) < SMAC_SYNC_INTERVAL_MS + SMAC_TIMING_TOLERANCE);
    else
        return ((known_phase - phase) < SMAC_TIMING_TOLERANCE);

    return FALSE;
}

static gboolean smac_test_collision_rcb (packet_info *pinfo, mote_priv_data *m_priv,
                                         smac_schedule_info_t *schedule)
{
    smac_schedules_tapinfo_t *tapinfo = &the_tapinfo_struct;
    guint64 phase, known_phase;

    phase = smac_schedules_phase(tapinfo->time_now);
    known_phase = schedule->phase;

    /* Return true IFF this packet arrived in the following interval:

       . . . Sleep ][  Sync Slots  ][ RTS/CTS/Bcast Contention ][ Data . . .
                                  ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
                                            (Target Area)

      Where the bits outside of contention is equal to SMAC_TIMING_TOLERANCE
      on either side.
    */

    /* printf("Testing for RCB p = %llu , kp = %llu , dp = %llu\n",
       phase, known_phase, (phase > known_phase) ? phase - known_phase : known_phase - phase); */
    
    if (phase > known_phase) 
        return (((phase - known_phase) > SMAC_SYNC_INTERVAL_MS - SMAC_TIMING_TOLERANCE) &&
                ((phase - known_phase) < SMAC_CONTENTION_TIME + SMAC_TIMING_TOLERANCE));

    return FALSE;
}

static gboolean smac_test_collision_data (packet_info *pinfo, mote_priv_data *m_priv,
                                          smac_schedule_info_t *schedule)
{
    smac_schedules_tapinfo_t *tapinfo = &the_tapinfo_struct;
    guint64 phase, known_phase;
    guint64 last_data_send, last_data_recv, last_rts, last_listen;
    GList *list;
    gboolean precondition = FALSE;
    
    phase = smac_schedules_phase(tapinfo->time_now);
    known_phase = schedule->phase;

    /* Return true IFF this packet arrived in the following interval:

       . . . Contention ][       Data      ][     Sleep    ][ Sync . . .
                       ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
                                     (Target Area)

      Where the bits outside of Sync Slots is equal to SMAC_TIMING_TOLERANCE
      on either side.

      AND if this schedule had another data packet preceeding this one in the
      same period OR this schedule had a RTS in the preceeding contention interval.

      Note that we cannot distinguish the boundary between Data and Sleep, which
      is why the extra conditions are necessary.
    */

    last_data_send = smac_visualization_find_prev_type_by_schedule(schedule,
                                                                  tapinfo->time_now,
                                                                  event_send_data_end);
    last_data_recv = smac_visualization_find_prev_type_by_schedule(schedule,
                                                                  tapinfo->time_now,
                                                                  event_recv_data_end);
    last_rts = smac_visualization_find_prev_type_by_schedule(schedule,
                                                            tapinfo->time_now,
                                                            event_send_oneshot_rts);
    last_listen = smac_visualization_find_prev_type_by_schedule(schedule,
                                                               tapinfo->time_now,
                                                               event_listen);

    /* printf("LL: %llu , LRTS: %llu , LDS: %llu , LDR  %llu \n",
       last_listen, last_rts, last_data_send, last_data_recv); */
    
    if ((last_listen < last_rts) && (last_rts - last_listen < SMAC_CONTENTION_TIME + 2 * SMAC_TIMING_TOLERANCE))
        precondition = TRUE;
    else if ((last_listen < last_data_send) && (last_data_send - last_listen > SMAC_CONTENTION_TIME - SMAC_TIMING_TOLERANCE))
        precondition = TRUE;
    else if ((last_listen < last_data_recv) && (last_data_recv - last_listen > SMAC_CONTENTION_TIME - SMAC_TIMING_TOLERANCE))
        precondition = TRUE;

    /* printf("Testing for DATA p = %llu , kp = %llu , dp = %llu\n",
       phase, known_phase, (phase > known_phase) ? phase - known_phase : known_phase - phase); */
    
    /* if (precondition) { */
    if (1) {
        if (phase > known_phase)
            return ((phase - known_phase) > SMAC_CONTENTION_TIME - SMAC_TIMING_TOLERANCE);
        else
            return ((known_phase - phase) < SMAC_TIMING_TOLERANCE);
    } else {
        printf("DATA precondition failed.\n");
    }
    
    return FALSE;
}

static void smac_get_node_from_hash(gpointer key, gpointer value, gpointer node_list_ptr) 
{
    GList **list = (GList**)node_list_ptr;
    smac_extended_seq_num_t* cur_ext_seq_num = (smac_extended_seq_num_t*)value;
  
    if(cur_ext_seq_num->total_syncs >= SMAC_MIN_SYNC_POINTS)
        *list = g_list_append(*list, (guint16*)key); 
}

static guint64 smac_compute_default_period()
{
    guint64 default_period;

    /* Based on the S-MAC options, choose the period length */
  
    default_period = SMAC_PERIOD_LENGTH_DEFAULT;
  
    if (smac_fast_path_schedule)
        default_period += SMAC_PERIOD_SKEW_FAST_PATH_SCHEDULE;
  
    if (smac_tx_timestamp)
        default_period += SMAC_PERIOD_SKEW_TX_TIMESTAMP;
  
    if (smac_global_schedule)
        default_period += SMAC_PERIOD_SKEW_GLOBAL_SCHEDULE;
  
    if (smac_snooper_debug)
        default_period += SMAC_PERIOD_SKEW_SNOOPER_DEBUG;

    return default_period;
}

static guint64 smac_compute_period()
{
    GList *list = NULL;
    GList *node_list = NULL;
    smac_estimate_t estimated_period;
    unsigned int periods_per_sync = 0;
    guint64 min_delta_sleep = 0;
    guint64 computed_period = 0;
    guint64 previous_computed_period = 0;
    gdouble average_correlation = 0;
    unsigned int iterations = 0;

  
    g_hash_table_foreach(smac_last_seq_nums, &smac_get_node_from_hash, &node_list);

    /* Use static defaults if we lack a sufficient number of points */

    if (g_list_length(node_list) == 0)
        return smac_compute_default_period();
    
    min_delta_sleep = smac_find_min_delta_sleep(smac_sync_points);

    if (smac_sync_model == SYNC_MODEL_TIMER) {

        /* Unconditional first approximation */    
        periods_per_sync = smac_compute_periods_per_sync(node_list, smac_sync_points,
                                                         min_delta_sleep);
    } else {
        periods_per_sync = SMAC_SYNC_PERIOD_LENGTH; /* In periods */
    }

    computed_period = NEAREST_GUINT64(min_delta_sleep / periods_per_sync);
     printf("Computed period: %llu (rounded), %llu (intdiv)\n", computed_period, min_delta_sleep/periods_per_sync); 

    do {
        iterations++;

        smac_compute_all_effective_seq_nums(smac_sync_points, min_delta_sleep, computed_period);
    
        previous_computed_period = computed_period;
        computed_period = 0;
        average_correlation = 0;
    
        for (list = g_list_first(node_list); list; list = g_list_next(list)) {
            estimated_period = smac_compute_estimate_period(smac_sync_points, 
                                                            *((guint16*)(list->data)),
                                                            TRUE);
            computed_period += estimated_period.value;
            average_correlation += estimated_period.correlation;
        }
    
        computed_period /=  g_list_length(node_list);
        average_correlation /=  g_list_length(node_list);

    } while (average_correlation < SMAC_MIN_PERIOD_CORRELATION &&
             computed_period != previous_computed_period &&
             iterations < SMAC_MAX_PERIOD_ITERATIONS);

    g_list_free(node_list);

    return (computed_period == 0) ? smac_compute_default_period() : computed_period;
}

static gboolean smac_is_anomaly(smac_sync_point_t *point_delta)
{
    guint64 min_reasonable_time;

    if (smac_sync_model == SYNC_MODEL_PERIODS)
        min_reasonable_time = NEAREST_GUINT64((gdouble)(SMAC_PERIOD_LENGTH_DEFAULT * 
                                                        SMAC_SYNC_PERIOD_LENGTH) * 
                                              ((gdouble)SMAC_SYNC_ANOMALY_THRESHOLD)/100.0);
    else if (smac_sync_model == SYNC_MODEL_TIMER)
        min_reasonable_time = NEAREST_GUINT64((gdouble)SMAC_SYNC_PERIOD_LENGTH * 
                                              ((gdouble)SMAC_SYNC_ANOMALY_THRESHOLD)/100.0);				  
    else
        return FALSE;

    return (point_delta->sleep_time < min_reasonable_time);
					  			 
}

static guint64 smac_find_min_delta_sleep(GList* sync_points)
{
    GArray *node_syncs_found = g_array_new(FALSE, TRUE, sizeof(int));
    GList *list;
    smac_sync_point_t *cur_sync_point = NULL;
    smac_sync_point_t *gst_point = NULL;
    smac_sync_point_t *point_delta = NULL;
    guint64 adjusted_delta_sleep = 0;
    guint64 min_delta_sleep = 0;
    guint8 num_anomalies = 0;
    guint32 total_anomalies = 0;
    guint8 points_since_anomaly = 0;
    int syncs_found = 0;


    for (list = g_list_first(smac_sync_points); list; list = g_list_next(list)) {
        cur_sync_point = list->data;

        if ((gst_point = smac_find_greatest_sleep_time(list, cur_sync_point->node, cur_sync_point->seq_num)))
            cur_sync_point = gst_point;

        node_syncs_found = g_array_set_size(node_syncs_found, cur_sync_point->node); 
        syncs_found = g_array_index(node_syncs_found, int, cur_sync_point->node);

        if (syncs_found < 3) {
            syncs_found++;
            node_syncs_found = g_array_insert_val(node_syncs_found, cur_sync_point->node, syncs_found);
            continue;
        }

        point_delta = smac_compute_sync_delta(list, cur_sync_point);
        if (point_delta && point_delta->seq_num > 0 && point_delta->seq_num <= SMAC_SYNC_DELTA_SEQ_TOLERANCE) {

            /*
              if(point_delta->seq_num > 1) 

              printf("%d\t%u\t%llu\t%llu\t%llu\t%u\n", cur_sync_point->node,
              cur_sync_point->seq_num, cur_sync_point->recv_time, cur_sync_point->sleep_time,
              NEAREST_GUINT64(((gdouble)point_delta->sleep_time) / point_delta->seq_num),
              point_delta->seq_num);

              else

              printf("%d\t%u\t%llu\t%llu\t%llu\t%u\n", cur_sync_point->node,
              cur_sync_point->seq_num, cur_sync_point->recv_time, cur_sync_point->sleep_time,
              point_delta->sleep_time, point_delta->seq_num);
            */

            if ((num_anomalies < SMAC_SYNC_ANOMALY_PTS_NEEDED) && (smac_is_anomaly(point_delta))) {
                points_since_anomaly = 0;
                num_anomalies++;
                total_anomalies++;
                break;
            } else
                points_since_anomaly++;

            if ((num_anomalies > 0) && (points_since_anomaly > SMAC_SYNC_ANOMALY_PTS_RESET))
                num_anomalies = 0;

            if (point_delta->seq_num > 1)
                adjusted_delta_sleep = NEAREST_GUINT64(((gdouble)point_delta->sleep_time) / point_delta->seq_num);
            else
                adjusted_delta_sleep = point_delta->sleep_time;

            if ((adjusted_delta_sleep < min_delta_sleep && adjusted_delta_sleep > 0) || min_delta_sleep == 0) { 
                assert(adjusted_delta_sleep > 0 && point_delta->seq_num > 0);
                min_delta_sleep = adjusted_delta_sleep;
            }
        }
    }

    assert(min_delta_sleep != 0);

    return min_delta_sleep;
}

static smac_sync_point_t* smac_find_greatest_sleep_time(const GList * const points_list, guint16 node, guint32 seq_num)
{
    GList *list = NULL;
    smac_sync_point_t *cur_point = NULL;
    smac_sync_point_t *gst_point = NULL;

    for (list = points_list; list; list = g_list_next(list)) {
        cur_point = list->data;
    
        if (cur_point->node == node && cur_point->seq_num == seq_num && 
            (gst_point == NULL || cur_point->sleep_time > gst_point->sleep_time))
            gst_point = cur_point;
        else if (cur_point->node == node && cur_point->seq_num > seq_num) 
            break;
    }

    return gst_point;
}

static smac_sync_point_t* smac_find_least_sleep_time(const GList * const points_list, guint16 node, guint32 seq_num)
{
    GList *list = NULL;
    smac_sync_point_t *cur_point = NULL;
    smac_sync_point_t *lst_point = NULL;

    for (list = points_list; list; list = g_list_previous(list)) {
        cur_point = list->data;
    
        if (cur_point->node == node && cur_point->seq_num == seq_num && 
            (lst_point == NULL || cur_point->sleep_time < lst_point->sleep_time))
            lst_point = cur_point;
        else if (cur_point->node == node && cur_point->seq_num < seq_num) 
            break;
    }

    return lst_point;
}

static unsigned int smac_compute_periods_per_sync(GList *node_list, GList* sync_points,
						  guint64 min_delta_sleep) 
{
    GList *list;
    unsigned int n = 0;
    unsigned int best_n = 0;
    guint16 cur_node = 0;
    guint64 test_period = 0;
    guint32 num_schedules = 0;
    guint32 best_num_schedules = 0;

    for(n = 2; n < NEAREST_GUINT32((1 + 2 * (gdouble)smac_duty_cycle / 100.0) * ((double)min_delta_sleep) / 
                                   (min_delta_sleep - SMAC_SYNC_PERIOD_LENGTH)); n++) {
    
        test_period = NEAREST_GUINT64(min_delta_sleep / n);

        num_schedules = smac_compute_num_schedules(sync_points, min_delta_sleep, test_period);

        if (num_schedules < best_num_schedules || best_num_schedules == 0) {
            best_n = n;
            best_num_schedules = num_schedules;
        }

    }

    assert(best_n != 0);

    return best_n;
}

static guint32 smac_compute_num_schedules(GList *sync_points, guint64 min_delta_sleep, 
					  guint64 period)
{
    GList *list = NULL;
    GList *list2 = NULL;
    GList *tmp_schedules_list = NULL;
    smac_sync_point_t *cur_sync_point = NULL; 
    smac_schedule_info_t *cur_schedule = NULL; 
    smac_schedule_info_t *tmp_schedule = NULL;
    guint64 cur_phase = 0;
    guint32 num_schedules = 0;

    for (list = g_list_first(sync_points); list; list = g_list_next(list)) {

        cur_sync_point = list->data;
        cur_phase = smac_schedules_relative_phase(cur_sync_point->sleep_time, period);

        tmp_schedule = NULL;

        for (list2 = g_list_first(tmp_schedules_list); list2; list2 = g_list_next(list2)) {

            cur_schedule = list2->data;

            if (llabs(cur_phase - cur_schedule->phase) < SMAC_TIMING_TOLERANCE ||
                llabs(cur_phase - (period - cur_schedule->phase)) < SMAC_TIMING_TOLERANCE) {
                tmp_schedule = cur_schedule;
                break;
            }
        }
    
        if (tmp_schedule == NULL) {
            tmp_schedule = (smac_schedule_info_t*)malloc(sizeof(smac_schedule_info_t));
            tmp_schedule->phase = cur_phase;
            tmp_schedules_list = g_list_append(tmp_schedules_list, tmp_schedule);
        }
    }
    

    num_schedules = g_list_length(tmp_schedules_list);

    for (list = g_list_first(tmp_schedules_list); list; list = g_list_next(list))
        free(list->data);

    g_list_free(tmp_schedules_list);

    return num_schedules;

}

static guint64 smac_moving_approx(guint64 cur_val, guint64 new_val, 
				  guint16 num_vals, guint16 max_vals)
{
    return (guint64)((cur_val * MIN((num_vals - 1),(max_vals - 1)) + new_val) / 
                     MIN(num_vals, max_vals));
}

static smac_sync_point_t* smac_compute_sync_delta(GList *sync_points,
						  smac_sync_point_t* sync_point)
{
    GList* list = NULL;
    smac_sync_point_t *delta_sync_point = NULL;
    smac_sync_point_t *old_sync_point = NULL;
    smac_sync_point_t *lst_point = NULL;

    for (list = g_list_previous(sync_points); list; list = g_list_previous(list)) {
        old_sync_point = list->data;
        if (old_sync_point->node == sync_point->node && 
            old_sync_point->seq_num < sync_point->seq_num) break;
        else old_sync_point = NULL;
    }
  
    if (old_sync_point) {
    
        if ((lst_point = smac_find_least_sleep_time(list, old_sync_point->node, old_sync_point->seq_num)))
            old_sync_point = lst_point;

        delta_sync_point = (smac_sync_point_t*)malloc(sizeof(smac_sync_point_t));
        delta_sync_point->node = sync_point->node - old_sync_point->node;
        delta_sync_point->seq_num = sync_point->seq_num - old_sync_point->seq_num;
        delta_sync_point->eff_seq_num = sync_point->eff_seq_num - old_sync_point->eff_seq_num;
        delta_sync_point->sleep_time = sync_point->sleep_time - old_sync_point->sleep_time;
    }

    return delta_sync_point;
}

static void smac_compute_all_effective_seq_nums(GList *sync_points, guint64 min_delta_sleep,
						guint64 period) 
{
    GList *list = NULL;

    for (list = g_list_first(sync_points); list; list = g_list_next(list)) 
        smac_compute_effective_seq_num(list, (smac_sync_point_t*)(list->data), min_delta_sleep,
                                       period);
}

static void smac_compute_effective_seq_num(GList *sync_points, 
					   smac_sync_point_t* sync_point, 
					   guint64 min_delta_sleep, guint64 period)
{
    /* The effective sequence number is the number of periods (as opposed to SYNC intervals)
       that have passed between beginning to record and the time of the given point. The
       seq num in packets transmitted from the node count up each time one is successfully
       sent, but we also want to know how many SHOULD have been sent. */
    
    smac_sync_point_t* point_delta = smac_compute_sync_delta(sync_points, sync_point);
    guint64 delta_sleep_time;
    guint64 delta_eff_seq_num;

    if (point_delta) {

        delta_sleep_time = point_delta->sleep_time;
        delta_eff_seq_num = point_delta->eff_seq_num;
    
        free(point_delta);
 
        sync_point->eff_seq_num = sync_point->eff_seq_num - delta_eff_seq_num + 
            NEAREST_GUINT32((gdouble)delta_sleep_time/(gdouble)period);

        if (delta_sleep_time > (min_delta_sleep * SMAC_PHASE_DRIFT_TOLERANCE))
            sync_point->discard = TRUE;
  
    } else {
        sync_point->discard = TRUE;
    }

}

/* Undefined for use_effectives = FALSE at the moment */
#define SEQ_NUM(pt) (use_effectives ? pt->eff_seq_num : pt->seq_num)

static smac_estimate_t smac_compute_estimate_period(GList* sync_points, guint16 node,
						    gboolean use_effectives)
{
    /*
      This function performs a linear least-squares analysis on the data points 
      sleep time vs. sequence number (with the "time" values being the values 
      with error) and returns the slope which is an estimate of the period.
    
      Forumula is:
    
      (x,y) = (seq_num, time)
      N = number of points 
      D = -(Sum(x)^2) + N*Sum(x^2)
      m = Sum((N*Sum(x*y)/D - Sum(x)*Sum(y)/D)*y)
    
      Note that we compute this on a per-node basis. Even though the resultant
      periods of all nodes should be very similar, their sequence numbers can
      be wildly different. We must compute seperately and average later.
  
      Notice that everything is a double! The y values can be as large as 2^64.
      In fact, in only an hour or so, the timestamps will reach 2^32 which will
      overflow on squaring.

    */

    GList* points_list = NULL;
    smac_sync_point_t* cur_point = NULL;
    smac_estimate_t result;
    guint64 num_points = 0;
    gdouble x_sum = 0;         /* Sum(x)   */
    gdouble x_squared_sum = 0; /* Sum(x^2) */
    gdouble y_squared_sum = 0; /* Sum(y^2) */
    gdouble y_sum = 0;         /* Sum(y)   */
    gdouble xy_sum = 0;        /* Sum(x*y) */
    gdouble x_avg = 0;
    gdouble y_avg = 0;
    gdouble x_ss = 0;
    gdouble y_ss = 0;
    gdouble xy_ss = 0;

    for (points_list = g_list_first(sync_points); points_list;
         points_list = g_list_next(points_list)) {

        cur_point = (smac_sync_point_t*)(points_list->data);

        if (cur_point->discard == TRUE) continue;

        if (cur_point->node == node) {

            num_points++;

            x_sum += SEQ_NUM(cur_point);
            x_squared_sum += ((gdouble)(SEQ_NUM(cur_point)) * SEQ_NUM(cur_point));
      
            y_sum += cur_point->sleep_time;
            y_squared_sum += ((gdouble)(cur_point->sleep_time) * cur_point->sleep_time);

            xy_sum += ((gdouble)(SEQ_NUM(cur_point)) * cur_point->sleep_time); 
        }
    }

    if (num_points >= SMAC_MIN_SYNC_POINTS) {

        result.exact_value = (((num_points * xy_sum) - (x_sum * y_sum)) / 
                              ((num_points * x_squared_sum) - (x_sum * x_sum)));

        result.value = NEAREST_GUINT64(result.exact_value);

        /* Compute correlation coefficient r^2 for this data */

        x_avg = x_sum / num_points;
        y_avg = y_sum / num_points;
    
        x_ss = x_squared_sum - num_points * (x_avg * x_avg);
        y_ss = y_squared_sum - num_points * (y_avg * y_avg);

        xy_ss = xy_sum - num_points * (x_avg * y_avg);

        result.correlation = (gdouble)(xy_ss * xy_ss) / (x_ss * y_ss);

    } else {
        result.exact_value = 0;
        result.value = 0;
        result.correlation = 0;
    }

    return result;
}

#undef SEQ_NUM

