/* smac_visualization.c
 * Graphic Analysis addition for S-MAC
 *
 * $Id: smac_visualization.c 5384 2006-07-11 20:45:24Z smchenry $
 *
 * Copyright 2005, University of Southern California
 * By Tyler McHenry <smchenry@isi.edu>
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation,  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#include "smac_visualization.h"
#include "smac_schedules.h"

#include <epan/epan_dissect.h>

#include "util.h"
#include <epan/tap.h>
#include "register.h"
#include <epan/dissectors/packet-rtp.h>
#include <epan/addr_resolv.h>

/* in /gtk ... */
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include "gtkglobals.h"

#include "dlg_utils.h"
#include "ui_util.h"
#include "main.h"
#include "compat_macros.h"
#include "../color.h"
#include "epan/filesystem.h"

#include <string.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "simple_dialog.h"

#define SMAC_VIS_MEASURED_PERIOD SMAC_PERIOD_LENGTH_DEFAULT
#define SMAC_VIS_CONTENTION_TIME NEAREST_GUINT64((smac_duty_cycle / 100.0) * SMAC_VIS_MEASURED_PERIOD)

GList* smac_node_streams;

/* External functions */

void smac_visualization_reset();
gint smac_visualization_cmp_nodes(gconstpointer a,
                                  gconstpointer b);
gint smac_visualization_cmp_schedules(gconstpointer a,
                                      gconstpointer b);
int smac_visualization_packet(void *t, packet_info *pinfo, 
			      epan_dissect_t *edt, void *priv);
void smac_visualization_dump(FILE* ofile, smac_visualization_export_format_t format);
gdouble smac_visualization_type_value(smac_node_stream_event_type_t type);
guint64 smac_visualization_find_prev_type_by_schedule(smac_schedule_info_t* schedule,
                                                      guint64 current_time,
                                                      smac_node_stream_event_type_t type);
guint64 smac_visualization_find_next_type_by_schedule(smac_schedule_info_t* schedule,
                                                      guint64 current_time,
                                                      smac_node_stream_event_type_t type);

/* Internal functions */

static smac_node_stream_t* smac_visualization_find_existing_stream(GList *list, address node_addr);
static smac_node_stream_t* smac_visualization_find_or_create_stream(address node_addr);

static void smac_visualization_generate_event(packet_info *pinfo,
                                              const mote_priv_data* m_priv,
                                              const smac_priv_data* s_priv,
                                              const smac_schedules_tapinfo_t* schedules_tapinfo,
                                              smac_node_stream_t* stream);
static void smac_visualization_add_event(smac_node_stream_t* stream, 
                                         smac_node_stream_event_type_t type,
                                         guint64 time);
static void smac_visualization_handle_cts_send(smac_node_stream_t* dest_stream, guint64 time_now);
static void smac_visualization_handle_cts_recv(smac_node_stream_t* stream, guint64 time_now);
static void smac_visualization_add_implicit_sleeps(smac_node_stream_t* stream, guint64 time_now);
static GList* smac_visualization_find_prev_type(smac_node_stream_t* stream,
                                                guint64 timestamp,
                                                smac_node_stream_event_type_t type);
static GList* smac_visualization_find_next_type(smac_node_stream_t* stream,
                                                guint64 timestamp,
                                                smac_node_stream_event_type_t type);
static GList* smac_visualization_find_timestamp(smac_node_stream_t* stream,
                                                guint64 timestamp);
static void smac_visualization_dump_jdb(FILE* ofile);
static void smac_visualization_dump_gnuplot(FILE* ofile, gboolean points);
static void smac_visualization_write_stream_jdb(FILE* ofile, smac_node_stream_t* stream);
static void smac_visualization_write_stream_gnuplot(FILE* ofile, smac_node_stream_t* stream, gboolean points);

void smac_visualization_reset() 
{
    /* Clear data records for re-processing. This is called if the SMAC schedules
       analysis tool is re-run or is being repeatedly re-called during live capture. */

    GList* list;
    GList* list2;
    smac_node_stream_t* stream;

    for (list = g_list_first(smac_node_streams); list;
         list = g_list_next(list)) {
        stream = list->data;
    
        for (list2 = g_list_first(stream->events); list2;
             list2 = g_list_next(list2))
            g_free(list2->data);

        g_list_free(stream->events);

        g_free(stream);
    }

    g_list_free(smac_node_streams);

}

gint smac_visualization_cmp_nodes(gconstpointer a,
                                  gconstpointer b)
{
    /* Compare the TOS addresses of two node pointers.
       Return 0 if equal, -1 if a < b, 1 if a > b.
       For use as a callback by glib sorting functions. */

    smac_node_info_t const* na = (smac_node_info_t*)a;
    smac_node_info_t const* nb = (smac_node_info_t*)b;
    guint16 nia, nib;

    nia = NODE_ID(na);
    nib = NODE_ID(nb);

    if (nia == nib) return 0;
    else return (nia < nib) ? -1 : 1;
}

gint smac_visualization_cmp_schedules(gconstpointer a,
                                      gconstpointer b)
{
    /* Compare the TOS addresses of two schedule pointers.
       Return 0 if equal, -1 if sa < sb, 1 if sa > sb.
       For use as a callback by glib sorting functions. */
    
    smac_schedule_info_t const* sa = (smac_schedule_info_t*)a;
    smac_schedule_info_t const* sb = (smac_schedule_info_t*)b;

    if (sa->id == sb->id) return 0;
    else return (sa->id < sb->id) ? -1 : 1;
}


static smac_node_stream_t* smac_visualization_find_existing_stream(GList *list, address node_addr) 
{
    /* Given a node address search for any existing stream associated with this
       address, and return it (or NULL if no such stream) */

    GList* tmp_list;
    smac_node_stream_t* stream;

    for (tmp_list = g_list_first(list); tmp_list;
         tmp_list = g_list_next(tmp_list)) {
        stream = tmp_list->data;
        if (ADDRESSES_EQUAL(&(stream->addr), &node_addr)) return stream;
    }

    return NULL;
}

static smac_node_stream_t* smac_visualization_find_or_create_stream(address node_addr)
{
    smac_node_stream_t* stream;

    /* Given a node address search for any existing stream associated with this
       address, and return it (create the stream if no such stream presently exists) */

    if ((stream = smac_visualization_find_existing_stream(smac_node_streams, node_addr)) == NULL) {
        stream = (smac_node_stream_t*)g_malloc(sizeof(smac_node_stream_t));
        smac_node_streams = g_list_append(smac_node_streams, stream);
    
        COPY_ADDRESS(&(stream->addr), &node_addr);
        stream->events = NULL;
        stream->recv_cts = FALSE;
        stream->baseline = NODE_ID(stream);
        stream->prev_sleep.type = event_null;
        stream->counter = 0;
    }
  
    return stream;
}


int smac_visualization_packet(void *t, packet_info *pinfo, 
			      epan_dissect_t *edt, void *priv)
{
    /* Here is where packets come in from the schedule tap. We are guaranteed
       by that tap that the packets we get are complete (i.e. have an s_priv
       and an m_priv that are each not NULL). Our job here is to convert this
       packet into an event if it corresponds to one, and add in any implicit
       events that have occurred since the previous packet. */

    const smac_priv_data* s_priv = NULL;
    const mote_priv_data* m_priv = NULL;
    const smac_schedules_tapinfo_t *schedules_tapinfo = smac_schedules_get_info();
    smac_node_stream_t* stream;

    s_priv = (smac_priv_data*)p_get_proto_data(pinfo->fd, proto_smac);
    m_priv = (mote_priv_data*)p_get_proto_data(pinfo->fd, proto_moteradio);

    stream = smac_visualization_find_or_create_stream(pinfo->src);

    smac_visualization_generate_event(pinfo, m_priv, s_priv,
                                      schedules_tapinfo, stream);

    return TRUE;
}

static void smac_visualization_generate_event(packet_info *pinfo,
				       const mote_priv_data* m_priv,
				       const smac_priv_data* s_priv,
				       const smac_schedules_tapinfo_t* schedules_tapinfo,
				       smac_node_stream_t* stream)
{
    /* Precondition: m_priv and s_priv are not null (complete packet), the stream
       exists and points to the stream for the node whose address is in the
       source field of pinfo. schedules_tapinfo points to the schedule data from
       the schedule analysis module.

       Determine based on the packet type, source, destination, and the state of the
       stream, which kind of event this packet represents, and additionally which other
       events are implied by the existence of this packet (if any). Call the appropriate
       add function for each of these events.
    */
    

    guint64 time_now = m_priv->rx_timestamp;
    smac_node_stream_t* dest_stream;

    /* Each time we are about to add new event(s) for a stream, we should first add
       implicit sleeps so that the implicit sleeps function does not have to worry
       about stepping over/around other events when adding the sleeps. (notice that
       this is also done when modifiying the stream for the destination node under
       TYPE_CTS_PKT) */
    
    if (stream->prev_sleep.type == event_sleep &&
        (time_now - stream->prev_sleep.timestamp) > SMAC_VIS_MEASURED_PERIOD) 
        smac_visualization_add_implicit_sleeps(stream, time_now); 

    switch(s_priv->type) {
  
        case TYPE_DATA_PKT:
            if (smac_addr_is_broadcast(&(pinfo->dst)))
                smac_visualization_add_event(stream, event_send_oneshot_data, time_now);
            else {
                dest_stream = smac_visualization_find_or_create_stream(pinfo->dst);
                /* event_*_data_end overwrites the previous unless there is a
                   event_*_data_begin in between */
                if (dest_stream->prev_sleep.type == event_sleep &&
                    (time_now - dest_stream->prev_sleep.timestamp) > SMAC_VIS_MEASURED_PERIOD) 
                    smac_visualization_add_implicit_sleeps(dest_stream, time_now); 
                smac_visualization_add_event(stream, event_send_data_end, time_now);
                smac_visualization_add_event(dest_stream, event_recv_data_end, time_now);
            }
            break;

        case TYPE_CTS_PKT:

            /* As opposed to the other types, CTS packets imply an event for both the
               sender and the receiver, since they should both log the beginning of
               a data flow (in opposite directions) at the end of the current contention
               period */
            
            dest_stream = smac_visualization_find_or_create_stream(pinfo->dst);
            if (dest_stream->prev_sleep.type == event_sleep &&
                (time_now - dest_stream->prev_sleep.timestamp) > SMAC_VIS_MEASURED_PERIOD) 
                smac_visualization_add_implicit_sleeps(dest_stream, time_now); 
            smac_visualization_handle_cts_send(dest_stream, time_now);
            smac_visualization_handle_cts_recv(stream, time_now);

            /* The -(SMAC_VIS_CONTENTION_TIME/2) is a slight kludge to make it less likely that
               the notation for the contention period event will fall outside (after) the notation
               for the contention period when visualized. This happens sometimes due to clock skew. */
            
            smac_visualization_add_event(stream, event_send_oneshot_cts, time_now - (SMAC_VIS_CONTENTION_TIME/2));
            break;
            
        case TYPE_RTS_PKT:
            smac_visualization_add_event(stream, event_send_oneshot_rts, time_now - (SMAC_VIS_CONTENTION_TIME/2)); 
            break;
 
        case TYPE_SYNC_PKT:
            smac_visualization_add_event(stream, event_send_oneshot_sync, time_now - (SMAC_VIS_CONTENTION_TIME/2));
            smac_visualization_add_event(stream, event_listen, time_now + MSEC_TO_USEC(s_priv->next_sleep) - SMAC_VIS_CONTENTION_TIME);
            smac_visualization_add_event(stream, event_sleep, time_now + MSEC_TO_USEC(s_priv->next_sleep));

            break;
    };

    return;

}


static void smac_visualization_add_event(smac_node_stream_t* stream, 
				  smac_node_stream_event_type_t type,
				  guint64 time)
{
    /* Given a stream with an event and time, I want to insert the given event into
       the given stream at the given time, but this event may imply side-effects for
       pre-existing events in this stream. Deal with those too. */
    
    GList *next_time_pos = smac_visualization_find_timestamp(stream, time);
    GList *prev_time_pos = NULL;
    smac_node_stream_event_t* new_event = NULL;
    smac_node_stream_event_t* prev_event = NULL;
    char comment_tmp[40];

    /* Grab a reasonable prev_time_pos and prev_event */

//    printf("Adding event to stream 0x%X\n", stream);
    
    if (next_time_pos != g_list_first(stream->events) &&
        next_time_pos != NULL)
        prev_time_pos = g_list_previous(next_time_pos);
    else if (next_time_pos == NULL)
        prev_time_pos = g_list_last(stream->events);

    while (prev_time_pos && ((smac_node_stream_event_t*)(prev_time_pos->data))->timestamp > time) {
        next_time_pos = prev_time_pos;
        prev_time_pos = g_list_previous(prev_time_pos);
    }

    /*
      if (next_time_pos && ((smac_node_stream_event_t*)(next_time_pos->data))->timestamp == time)
      return;
      if (prev_time_pos && ((smac_node_stream_event_t*)(prev_time_pos->data))->timestamp == time)
      return;
    */

    if (prev_time_pos) prev_event = (smac_node_stream_event_t*)(prev_time_pos->data);

    if (prev_event && 
        ((type == event_send_data_end && prev_event->type == event_send_data_end) ||
         (type == event_recv_data_end && prev_event->type == event_recv_data_end))) {

        /* If we are not the first event and we are in the middle of a send or a
           receive ... */

        new_event = (smac_node_stream_event_t*)malloc(sizeof(smac_node_stream_event_t));
        new_event->type = type;
        new_event->timestamp = time;

        /* Change the previous "data end" event into a "data packet" event, since by
           virtue of this event we know that the previous packet was not the end.
           Assume that THIS event is the end until we are corrected by a future event */
        
        if (prev_event->type == event_send_data_end)
            prev_event->type = event_send_data_packet;
        if (prev_event->type == event_recv_data_end)
            prev_event->type = event_recv_data_packet;

        strncpy(comment_tmp, new_event->comment, 40);
        snprintf(new_event->comment, 40, "%s,%d", comment_tmp, stream->counter++);

        stream->events = g_list_insert_before(stream->events, next_time_pos, new_event);
 
    } else if (prev_event &&
               ((type == event_send_data_begin && prev_event->type == event_sleep) ||
                (type == event_recv_data_begin && prev_event->type == event_sleep))) {
 
        /* If we are not the first event as we are just starting a send/recv */

        prev_event->type = type;
        prev_event->timestamp = time;
        snprintf(prev_event->comment, 40, "added by SVAE: *%d", stream->counter++);
    
        smac_visualization_add_event(stream, event_sleep, time + SMAC_VIS_MEASURED_PERIOD);

    } else {

        /* Otherwise we are either the first event or a oneshot event */
    
        new_event = (smac_node_stream_event_t*)malloc(sizeof(smac_node_stream_event_t));
        new_event->type = type;
        new_event->timestamp = time;

        /* Account for the possibility that we may have missed noting the beginning
           of this transfer or may have begun recording in media res */
        
        if (prev_event && type == event_send_data_end && 
            prev_event->type != event_send_data_begin)
            new_event->type = event_send_data_begin;

        if (prev_event && type == event_recv_data_end && 
            prev_event->type != event_recv_data_begin)
            new_event->type = event_recv_data_begin;

        snprintf(new_event->comment, 40, "added by SVAE: %d", stream->counter++);
    
        stream->events = g_list_insert_before(stream->events, next_time_pos, new_event);

        /* Record the last time that we noted a sleep so that the implicit sleeps
           function (which is called frequently) doesn't have to go searching through
           the stream to find it every time */
        
        if (new_event->type == event_sleep) {
            stream->prev_sleep.type = new_event->type;
            stream->prev_sleep.timestamp = new_event->timestamp;
        }
  
    }
}

static void smac_visualization_handle_cts_send(smac_node_stream_t* dest_stream, guint64 time_now)
{
    /* This function will add a wake_up + event_send_data_start event at the next
       scheduled sleep time (instead of a sleep event) */

    GList* prev_listen_pos = smac_visualization_find_prev_type(dest_stream, time_now, event_listen);
    guint64 prev_listen_time;

    if(prev_listen_pos)
        prev_listen_time = ((smac_node_stream_event_t*)(prev_listen_pos->data))->timestamp;
    else
        prev_listen_time = time_now - (0.8 * SMAC_VIS_CONTENTION_TIME);

    smac_visualization_add_event(dest_stream, event_send_data_begin, prev_listen_time + SMAC_VIS_CONTENTION_TIME);
    smac_visualization_add_event(dest_stream, event_listen, prev_listen_time + SMAC_VIS_MEASURED_PERIOD);
    smac_visualization_add_event(dest_stream, event_sleep, prev_listen_time + SMAC_VIS_MEASURED_PERIOD + SMAC_VIS_CONTENTION_TIME);
    smac_visualization_add_event(dest_stream, event_listen, prev_listen_time);

}

static void smac_visualization_handle_cts_recv(smac_node_stream_t* stream, guint64 time_now)
{
    /* This function will add a wake_up + event_send_data_start event at the next
       scheduled sleep time (instead of a sleep event) */

    GList* prev_listen_pos = smac_visualization_find_prev_type(stream, time_now, event_listen);;
    guint64 prev_listen_time = 0;
  
    if(prev_listen_pos)
        prev_listen_time = ((smac_node_stream_event_t*)(prev_listen_pos->data))->timestamp;
    else
        prev_listen_time = time_now - (0.8 * SMAC_VIS_CONTENTION_TIME);

    smac_visualization_add_event(stream, event_listen, prev_listen_time + SMAC_VIS_MEASURED_PERIOD);
    smac_visualization_add_event(stream, event_listen, prev_listen_time);
    smac_visualization_add_event(stream, event_sleep, prev_listen_time + SMAC_VIS_MEASURED_PERIOD + SMAC_VIS_CONTENTION_TIME);
    smac_visualization_add_event(stream, event_sleep, prev_listen_time + SMAC_VIS_CONTENTION_TIME - SMAC_VIS_MEASURED_PERIOD); 
    smac_visualization_add_event(stream, event_recv_data_begin, prev_listen_time + SMAC_VIS_CONTENTION_TIME);
}


static void smac_visualization_add_implicit_sleeps(smac_node_stream_t* stream, guint64 time_now)
{
    /* Given a stream and a timestamp, find the previous sleep time in the stream, and insert
       additional sleep times at the appropriate intervals up to but not past the given timestamp
       This corrects for the reality that sleep times are only announced explictly by SYNCs
       every 10 sleeps or so. */
    
    guint64 sleep_time = 0;

    if (stream->prev_sleep.type != event_sleep) return;

    sleep_time = stream->prev_sleep.timestamp + SMAC_VIS_MEASURED_PERIOD;
 
    while (sleep_time < time_now) {
        smac_visualization_add_event(stream, event_listen, sleep_time - SMAC_VIS_CONTENTION_TIME); 
        smac_visualization_add_event(stream, event_sleep, sleep_time);
        sleep_time += SMAC_VIS_MEASURED_PERIOD;
    }

}

static GList* smac_visualization_find_prev_type(smac_node_stream_t* stream,
					 guint64 timestamp,
					 smac_node_stream_event_type_t type)

{
    /* Given a stream, a time, and a type, return the position in this stream which
       is an event of the given type and is closest to the time without
       going past/over */
    
    GList* next_time_pos = smac_visualization_find_timestamp(stream, timestamp); 
    GList* prev_type_pos = NULL; 

    if (next_time_pos == g_list_first(stream->events)) {
        printf("Given stream does not begin until after time %llu \n", timestamp);
        return NULL;
    }

//    printf("Looking for type %d in stream 0x%X ", type, stream);
    
    for (prev_type_pos = g_list_previous(next_time_pos);
         prev_type_pos; prev_type_pos = g_list_previous(prev_type_pos)) {
        if (((smac_node_stream_event_t*)(prev_type_pos->data))->type == type) {
//            printf(" ... found one. \n");
            return prev_type_pos;
          }
//          else
//            printf(" ... but this is type %d. \n", 
//                   ((smac_node_stream_event_t*)(prev_type_pos->data))->type);
    }

//    printf(" ... stopped looking.\n");

    return NULL;
}

guint64 smac_visualization_find_prev_type_by_schedule(smac_schedule_info_t* schedule,
                                                      guint64 current_time,
                                                      smac_node_stream_event_type_t type)
{
    /* Given a schedule, a time, and a type, return THE TIMESTAMP OF the event of the
       given type that is nearest the given time without going past for ANY node that
       is on this schedule.

       Note that this function returns a timestamp as opposed to a stream position. This
       function is intended to be called from other modules which are not aware of the
       event model and associated data structures. Functions local to SMAC visualization
       will probably want to use the functions without _by_schedule. */

    smac_node_stream_event_t* event;
    smac_node_stream_t* stream;
    smac_node_info_t* node;
    GList *list;
    GList *result;
    guint64 latest_time = 0;
    
    if (!schedule || !(schedule->nodes_list)) {
        printf("Given schedule is null or has no nodes, cannot operate.\n");
        return 0;
    }

//    printf("Finding previous type %d for schedule %d (time = %llu)\n",
//           type, schedule->id, current_time);
    
    for (list = g_list_first(schedule->nodes_list); list;
         list = g_list_next(list)) {
        node = (smac_node_info_t*)(list->data);
        stream = smac_visualization_find_existing_stream(smac_node_streams,
                                                         node->addr);
        if (!stream) {
            printf("Could not get stream for node in schedule.\n");
            continue;
        }
        result = smac_visualization_find_prev_type(stream, current_time, type);
        if (!result) {
            printf("No previous packet of this type for current node\n");
            continue;
        }
        event = (smac_node_stream_event_t*)(result->data);
        if (event->timestamp > latest_time)
            latest_time = event->timestamp;
    }
    
    return latest_time;
}

static GList* smac_visualization_find_next_type(smac_node_stream_t* stream,
					 guint64 timestamp,
					 smac_node_stream_event_type_t type)
{
    /* Analogous to find_prev_type except we look for an event that occurs
       at or AFTER the given time */

    GList *next_type_pos = NULL;

    for (next_type_pos = smac_visualization_find_timestamp(stream, timestamp); 
         next_type_pos; next_type_pos = g_list_next(next_type_pos)) 
        if (((smac_node_stream_event_t*)(next_type_pos->data))->type == type)
            return next_type_pos;

    return NULL;
}

guint64 smac_visualization_find_next_type_by_schedule(smac_schedule_info_t* schedule,
                                                      guint64 current_time,
                                                      smac_node_stream_event_type_t type)
{
    /* Analogous to find_prev_type_by_schedule except we look for an event that occurs
       at or AFTER the given time */
    
    smac_node_stream_event_t* event;
    smac_node_stream_t* stream;
    smac_node_info_t* node;
    GList *list;
    GList *result;
    guint64 earliest_time = 0;

    if (!schedule || !(schedule->nodes_list)) return 0;

    for (list = g_list_first(schedule->nodes_list); list;
         list = g_list_next(list)) {
        node = (smac_node_info_t*)(list->data);
        stream = smac_visualization_find_existing_stream(smac_node_streams,
                                                         node->addr);
        if (!stream) continue;
        result = smac_visualization_find_next_type(stream, current_time, type);
        if (!result) continue;
        event = (smac_node_stream_event_t*)(result->data);
        if (event->timestamp < earliest_time || earliest_time == 0)
            earliest_time = event->timestamp;
    }
    
    return earliest_time;
}


static GList* smac_visualization_find_timestamp(smac_node_stream_t* stream,
					 guint64 timestamp)
{
    /* Given a stream and a timestmap, return the position in the stream
       that is closest to this timestamp without going over */
    
    GList *prev_time_pos = NULL;
    
    for (prev_time_pos = g_list_last(stream->events); prev_time_pos;
         prev_time_pos = g_list_previous(prev_time_pos)) 
        if (((smac_node_stream_event_t*)(prev_time_pos->data))->timestamp <= timestamp) 
            return g_list_next(prev_time_pos);

    return g_list_first(stream->events);
}

void smac_visualization_dump(FILE* ofile, smac_visualization_export_format_t format)
{
    /* Disambiguator for dump function depending on the selected format */
    
    switch(format) {
        case exfmt_jdb_events:
            smac_visualization_dump_jdb(ofile);
            break;
        case exfmt_gnuplot_events:
            smac_visualization_dump_gnuplot(ofile, FALSE);
            break;
        case exfmt_gnuplot_points:
            smac_visualization_dump_gnuplot(ofile, TRUE);
            break;
    }
}

static void smac_visualization_dump_jdb(FILE* ofile)
{
    /* Print headers for JDB data, call dump function for each stream */
    
    GList* list;
    smac_node_stream_t* stream;

    fprintf(ofile, "\n");
    fprintf(ofile, "#h id\ttimestamp\ttype\tvalue \n");
    fprintf(ofile, "\n");
    fprintf(ofile, "#####################################################################################\n");
    fprintf(ofile, "# \n");
    fprintf(ofile, "# S-MAC Visualization output from Ethereal with Mote Network Visualization\n");
    fprintf(ofile, "# This file is suitable for processing with jdb or anything else that can\n");
    fprintf(ofile, "# parse a whitespace-delimited data file.\n");
    fprintf(ofile, "# \n");
    fprintf(ofile, "# Columns are as follows (the #h line below defines the column names for jdb)\n");
    fprintf(ofile, "# \n");
    fprintf(ofile, "#   id   \tS-MAC Node ID of the node that experienced the event\n");
    fprintf(ofile, "#   timestamp\tRelative timestamp (in microseconds) of the event, the x coordinate\n");
    fprintf(ofile, "#   type\tEvent type code (see below)\n");
    fprintf(ofile, "#   value\tThe y-coordinate of this event if it were to be plotted. This is equal to\n");
    fprintf(ofile, "#        \tthe node id plus or minus an offset that depends on the type.\n");
    fprintf(ofile, "# \n");
    fprintf(ofile, "# Event type codes are as follows:\n");
    fprintf(ofile, "# \n");
    fprintf(ofile, "# \t 0\tNULL Event (should not appear in this file)\n");
    fprintf(ofile, "# \t 1\tNode goes to sleep\n");
    fprintf(ofile, "# \t 2\tNode goes into listen mode\n");
    fprintf(ofile, "# \t 3\tNode sends a broadcast data packet\n");
    fprintf(ofile, "# \t 4\tNode sends a SYNC packet\n");
    fprintf(ofile, "# \t 5\tNode sends an RTS packet\n");
    fprintf(ofile, "# \t 6\tNode sends a CTS packet\n");
    fprintf(ofile, "# \t 7\tNode goes into transmit mode and begins a unicast transfer (first packet)\n");
    fprintf(ofile, "# \t 8\tNode sends a packet in a unicast transfer\n");
    fprintf(ofile, "# \t 9\tNode finishes a unicast transfer (last packet) and goes to sleep\n");
    fprintf(ofile, "# \t10\tNode goes into listen mode and begins to receive a unicast transfer (first packet)\n");
    fprintf(ofile, "# \t11\tNode recieves a packet in a unicast transfer\n");
    fprintf(ofile, "# \t12\tNode finishes receiving a unicast transfer (last packet) and goes to sleep\n");
    fprintf(ofile, "# \n");
    fprintf(ofile, "#####################################################################################\n");
    fprintf(ofile, "\n");
    fprintf(ofile, "#h id\ttimestamp\ttype\tvalue \n");
    fprintf(ofile, "\n");

    for (list = g_list_first(smac_node_streams); list;
         list = g_list_next(list)) {
        stream = list->data;
    
        smac_visualization_write_stream_jdb(ofile, stream);
    }
}

static void smac_visualization_dump_gnuplot(FILE* ofile, gboolean points)
{
    /* Print headers for GNUPlot data, call dump function for each stream */
    
    GList* sorted_streams;
    GList* list;
    smac_node_stream_t* stream;
    guint16 last_id = 0;
    
    fprintf(ofile, "#####################################################################################\n");
    fprintf(ofile, "# \n");
    fprintf(ofile, "# S-MAC Visualization output from Ethereal with Mote Network Visualization\n");
    fprintf(ofile, "# This file is suitable for processing with gnuplot.\n");
    fprintf(ofile, "# \n");
    fprintf(ofile, "# Columns are simply X and Y coordinates.\n");
    fprintf(ofile, "# \n");
    fprintf(ofile, "# The most simplest command to plot this data in gnuplot is:\n");
    fprintf(ofile, "# \t gnuplot> plot \"data_file\" with lines\n");
    fprintf(ofile, "# \n");
    fprintf(ofile, "# Data from different nodes is seperated by two blank lines, making the gnuplot\n");
    fprintf(ofile, "# \"index\" equal to the note number. (see gnuplot documentation on the plot\n");
    fprintf(ofile, "# command's index modifier. Dummy data is inserted to ensure that the index\n");
    fprintf(ofile, "# of a node is always the same as the node's S-MAC ID (in case you have data\n");
    fprintf(ofile, "# captured from nodes n and n+2 but not n+1).\n");
    fprintf(ofile, "# \n");

    if (points) {
        fprintf(ofile, "# POINTS MODE: This file was generated in \"points mode\" which means that intermediate\n");
        fprintf(ofile, "# coordinates which are not data points have been inserted. The effect of these extra\n");
        fprintf(ofile, "# coordinates is that if you plot this file in gnuplot without any modification, it\n");
        fprintf(ofile, "# should look reasonably similar to the visualization generated by Ethereal+MNV.\n");
    } else {
        fprintf(ofile, "# EVENTS MODE: This file was generated in \"events mode\" which means that the coordinates\n");
        fprintf(ofile, "# in this file are data points that represent only the events themselves. Plotting this file\n");
        fprintf(ofile, "# without any modification or customization will result in a sawtooth graph rather than\n");
        fprintf(ofile, "# the square wave that Ethereal+MNV presents since there are no intermediate coordinates.\n");
    }

    fprintf(ofile, "# \n");
    fprintf(ofile, "#####################################################################################\n");
    fprintf(ofile, "\n");

    sorted_streams = g_list_sort(smac_node_streams, smac_visualization_cmp_nodes);

    for (list = g_list_first(sorted_streams); list; list = g_list_next(list)) {

        stream = list->data;

        while (last_id > 0 && last_id < (NODE_ID(stream) - 1)) {
            fprintf(ofile, "# Index: Node %d (dummy)\n", last_id);
            fprintf(ofile, "0\t0\n");
            fprintf(ofile, "\n\n");
            last_id++;
        }
        
        fprintf(ofile, "# Index: Node %d\n", NODE_ID(stream));
        smac_visualization_write_stream_gnuplot(ofile, stream, points);
        fprintf(ofile, "\n\n");

        last_id = NODE_ID(stream);
    }
    
}

static void smac_visualization_write_stream_jdb(FILE* ofile, smac_node_stream_t* stream)
{
    /* JDB dumps are basically just tab-delimited dumps of the interesting parts of
       the event data structure for each event */

    GList* events;
    smac_node_stream_event_type_t prev_type = event_null;
    smac_node_stream_event_t* cur_event = NULL;
    
    for (events = g_list_first(stream->events); events;
         events = g_list_next(events)) {
        cur_event = events->data;

        fprintf(ofile, "%u\t%llu\t%u\t%f\n", NODE_ID(stream),
                cur_event->timestamp, cur_event->type,
                stream->baseline + smac_visualization_type_value(cur_event->type));

        if (!IS_ONESHOT(cur_event->type)) prev_type = cur_event->type;
    }
}

static void smac_visualization_write_stream_gnuplot(FILE* ofile, smac_node_stream_t* stream, gboolean points)
{
    /* GNUPlot dumps are value-only dumps, suitable for immediate plotting. If
       points is true, then this also adds in "end of value" events right before
       the value changes in order to facilitate a square-wave plot. */
    
    GList* events;
    smac_node_stream_event_type_t prev_type = event_null;
    smac_node_stream_event_t* cur_event = NULL;
    
    for (events = g_list_first(stream->events); events;
         events = g_list_next(events)) {
        cur_event = events->data;
        
        if (points && (prev_type != event_null))
            fprintf(ofile, "%llu\t%f\n", cur_event->timestamp - 1, 
                    stream->baseline + smac_visualization_type_value(prev_type));

        fprintf(ofile, "%llu\t%f\n", cur_event->timestamp, 
                stream->baseline + smac_visualization_type_value(cur_event->type));

        if (IS_ONESHOT(cur_event->type)) {
           if (points) {
              fprintf(ofile, "%llu\t%f\n",
                      cur_event->timestamp + SMAC_VIS_ONESHOT_DURATION,
                      stream->baseline + smac_visualization_type_value(cur_event->type));
           
              fprintf(ofile, "%llu\t%f\n", 
                      cur_event->timestamp + SMAC_VIS_ONESHOT_DURATION+1,
                      stream->baseline + (prev_type != event_null ? smac_visualization_type_value(prev_type) :
                                          smac_visualization_type_value(event_listen)));
           }
        } else 
            prev_type = cur_event->type;
    }
}

gdouble smac_visualization_type_value(smac_node_stream_event_type_t type) 
{
    /* Assign a value in (-0.5,0.5) to each type of event for plotting purposes.
       This value is scaled appropriately before display so assume that the baseline is at 0
       and we can vary up to half a unit away in either direction.

       Logically, these events are grouped into "classes" since breaking it down
       any more finely would just be confusing to the user - data overload */
    
    switch(type) {

        /* "Sleep" events */
        case event_sleep:
        case event_send_data_end:
        case event_recv_data_end:
            return 0;

        /* "Transmit" events */    
        case event_send_data_begin:
        case event_send_data_packet:
        case event_send_oneshot_data:
        case event_send_oneshot_sync:
        case event_send_oneshot_rts:
        case event_send_oneshot_cts:
            return SMAC_VIS_SEND_HEIGHT;

        /* "Passive receive" events */    
        case event_listen:
            return - (SMAC_VIS_SEND_HEIGHT / 3.0);

        /* "Active receive" events */
        case event_recv_data_begin:
        case event_recv_data_packet:
            return - ((2.0 * SMAC_VIS_SEND_HEIGHT) / 3.0);

        /* This is bad */
        case event_null:
            exit(1);
    }
    
    return 0.0;
}
 
   
