/* smac_visualization_dlg.c
 * Graphic Analysis addition for S-MAC
 *
 * $Id: smac_visualization_dlg.c 5199 2006-04-21 00:38:19Z smchenry $
 *
 * Copyright 2005, University of Southern California
 * By Tyler McHenry <smchenry@isi.edu>
 *
 * based on graph_analysis.c
 * Copyright 2004, Verso Technologies Inc.
 * By Alejandro Vaquero <alejandrovaquero@yahoo.com>
 *
 * based on rtp_analysis.c and io_stat
 *
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation,  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>
#include <ctype.h>
#include <stdio.h>

#include "register.h"

#include "smac_schedules.h"
#include "smac_visualization.h"
#include "smac_visualization_dlg.h"

#include "globals.h"
#include "epan/filesystem.h"

#include <epan/tap.h>
#include "tap_menu.h"
#include "dlg_utils.h"
#include "ui_util.h"
#include "compat_macros.h"
#include "gtkglobals.h"

#include "image/clist_ascend.xpm"
#include "image/clist_descend.xpm"
#include "simple_dialog.h"

#define SCALE_FACTOR_Y 50.0

#define SCALE_X(x, s, d, w) (guint32)(((x-s)/d)*w)
#define SCALE_Y(b, d, h) (guint32)(h-(SCALE_FACTOR_Y*(b+d)))

#define BASE_VIEW_WIDTH 700
#define BASE_VIEW_HEIGHT 500
#define INITIAL_ZOOM_RATIO 0.005 /* Zoom levels per second of data */
#define MAX_INIT_ZOOM 40

typedef struct _smac_pair {
    void* first;
    void* second;
} smac_pair_t;

/* External Functions */

smac_visualization_dlg_t* smac_visualization_dlg_open(smac_visualization_dlg_t* dlg);
void smac_visualization_dlg_draw(smac_visualization_dlg_t* dlg);

/* Internal Functions */

static void smac_visualization_dlg_draw_line_wrapper(smac_visualization_dlg_t* dlg, GdkPixmap* bm, GdkGC* gc,
                                                     guint32 x1, guint32 y1, guint32 x2, guint32 y2);
static const char* smac_visualization_dlg_get_oneshot_lbl(smac_node_stream_event_type_t type);
static gint smac_visualization_dlg_get_oneshot_lbl_offset(smac_node_stream_event_type_t type);
static GdkColor* smac_visualization_dlg_get_oneshot_lbl_color(smac_node_stream_event_type_t type,
                                                              smac_visualization_dlg_t* dlg);
static gboolean smac_visualization_dlg_suppress_event(smac_visualization_dlg_filters_t filters,
						      smac_node_stream_event_type_t type);
static void smac_visualization_dlg_draw_tic(smac_visualization_dlg_t* dlg,
                                            guint32 x, guint32 y,
                                            smac_node_stream_event_type_t type);
static void smac_visualization_dlg_draw_stream(smac_visualization_dlg_t* dlg,
                                               smac_node_stream_t* stream);
static gint smac_visualization_dlg_cmp_guint16(gconstpointer a, gconstpointer b);
static GList* smac_visualization_dlg_show_nodes(smac_visualization_dlg_params_t params);
static void smac_visualization_dlg_redraw(smac_visualization_dlg_t* dlg);
static gboolean smac_visualization_dlg_configure_event(GtkWidget *widget, 
						       GdkEventConfigure *event,
						       smac_visualization_dlg_t* dlg );
static gboolean smac_visualization_dlg_export_filename_clicked(GtkWidget *widget, 
							       GtkWidget *txt_filename);
static gboolean smac_visualization_dlg_export_export_clicked(GtkWidget *widget, 
							     smac_visualization_dlg_export_dlg_t* exdlg);
static gboolean smac_visualization_dlg_export_destroy(GtkWidget *widget, 
						      smac_visualization_dlg_export_dlg_t* exdlg);
static GtkWidget* smac_visualization_dlg_export_dlg_create();
static gboolean smac_visualization_dlg_expose_event( GtkWidget *widget,
						     GdkEventExpose *event,
						     smac_visualization_dlg_t* dlg);
static gboolean smac_visualization_dlg_options_clicked( GtkWidget *widget, 
							GtkWidget *opt_panel);
static gboolean smac_visualization_dlg_export_clicked( GtkWidget *widget, 
						       smac_visualization_dlg_t* dlg );
static gboolean smac_visualization_dlg_print_clicked( GtkWidget *widget, 
                                                      smac_visualization_dlg_t* dlg );
static gboolean smac_visualization_dlg_close_clicked( GtkWidget *widget, 
						      smac_visualization_dlg_t* dlg );
static gboolean smac_visualization_dlg_clear_clicked( GtkWidget *widget, 
						      smac_visualization_dlg_t* dlg );
static gboolean smac_visualization_dlg_source_chk_event(GtkWidget *button, 
							smac_visualization_dlg_t* dlg);
static gboolean smac_visualization_dlg_source_add_to_selection(GtkTreeModel *model,
							       GtkTreePath *path,
							       GtkTreeIter *iter,
							       gpointer sel_list_ptr_arg);
static void smac_visualization_dlg_source_lst_event(GtkTreeView* lst_sources,
						    smac_visualization_dlg_t* dlg);
static void smac_visualization_dlg_source_lst_sel_event(GtkTreeSelection* sel_sources,
							smac_visualization_dlg_t* dlg);
static gboolean smac_visualization_dlg_filter_event(GtkWidget *button, 
						    smac_pair_t* dlg_and_toggle);
static gboolean smac_visualization_dlg_zoom_event( GtkWidget *widget, smac_visualization_dlg_t* dlg );
static guint smac_visualization_dlg_compute_needed_height(smac_visualization_dlg_params_t params);
static void smac_visualization_dlg_populate_sources(smac_visualization_dlg_params_t params, GtkWidget* lst_schedules, GtkWidget* lst_nodes);
static void smac_visualization_dlg_sources_setup_controls(GtkWidget* box, smac_visualization_dlg_t* dlg);
static void smac_visualization_dlg_filters_setup_controls(GtkWidget* box, smac_visualization_dlg_t* dlg);
static void smac_visualization_dlg_setup_colors(smac_visualization_dlg_t* dlg);
static void smac_visualization_dlg_setup(smac_visualization_dlg_t* dlg);
static smac_visualization_dlg_t* smac_visualization_dlg_create();
static smac_pair_t* smac_visualization_make_pair(void* first, void* second);

static char oneshot_lbl_rts[2] = "R";
static char oneshot_lbl_cts[2] = "C";
static char oneshot_lbl_sync[2] = "S";
static char oneshot_lbl_bcast[2] = "B";

static void smac_visualization_dlg_draw_line_wrapper(smac_visualization_dlg_t* dlg,
                                                     GdkPixmap* bm, GdkGC* gc,
                                                     guint32 x1, guint32 y1, 
                                                     guint32 x2, guint32 y2)
{
    if (!gc) gc = dlg->graph_area->style->black_gc;
    
    gdk_draw_line(bm, gc, x1, y1, x2, y2);
}

static const char* smac_visualization_dlg_get_oneshot_lbl(smac_node_stream_event_type_t type)
{
    switch (type) {
        case event_send_oneshot_rts:
            return oneshot_lbl_rts;
        case event_send_oneshot_cts:
            return oneshot_lbl_cts;
        case event_send_oneshot_sync:
            return oneshot_lbl_sync;
        case event_send_oneshot_data:
            return oneshot_lbl_bcast;
        default:
            return NULL;
    }
}

static gint smac_visualization_dlg_get_oneshot_lbl_offset(smac_node_stream_event_type_t type)
{
    switch (type) {
        case event_send_oneshot_rts:
        case event_send_oneshot_sync:
            return SMAC_VIS_LBL_CORRECTION - SMAC_VIS_LBL_OFFSET;
        case event_send_oneshot_cts:
        case event_send_oneshot_data:
            return SMAC_VIS_LBL_CORRECTION + SMAC_VIS_LBL_OFFSET;
        default:
            return 0;
    }
}

static GdkColor* smac_visualization_dlg_get_oneshot_lbl_color(smac_node_stream_event_type_t type,
                                                              smac_visualization_dlg_t* dlg)
{
    switch (type) {
        case event_send_oneshot_rts:
            return &(dlg->colors.rts);
        case event_send_oneshot_cts:
            return &(dlg->colors.cts);
        case event_send_oneshot_sync:
            return &(dlg->colors.sync);
        case event_send_oneshot_data:
            return &(dlg->colors.bcast);
        default:
            return NULL;
    }
}

static gboolean smac_visualization_dlg_suppress_event(smac_visualization_dlg_filters_t filters,
						      smac_node_stream_event_type_t type)
{
    switch (type) {

        case event_listen:
            return !(filters.listen_cycles);

        case event_send_oneshot_sync:
            return !(filters.sync_packets);
            
        case event_send_oneshot_rts:
        case event_send_oneshot_cts:
            return !(filters.rts_cts_packets);

        case event_send_oneshot_data:
        case event_send_data_begin:
        case event_recv_data_begin:
            return !(filters.data_transfers);

        case event_send_data_packet:
        case event_recv_data_packet:
            return !(filters.data_packets) || !(filters.data_transfers);

        default:
            return FALSE;
    }
}

static void smac_visualization_dlg_draw_tic(smac_visualization_dlg_t* dlg,
                                            guint32 x, guint32 y,
                                            smac_node_stream_event_type_t type)
{
    guint32 change = (SCALE_FACTOR_Y * SMAC_VIS_TIC_HEIGHT) / 2.0;

    if (!IS_DATAPACKET(type)) return;

    smac_visualization_dlg_draw_line_wrapper(dlg, dlg->backing_map, dlg->gc,
                                             x, y - change,
                                             x, y + change);
}
		
static void smac_visualization_dlg_draw_stream(smac_visualization_dlg_t* dlg,
                                               smac_node_stream_t* stream)
{
    GList* events;
    smac_node_stream_event_type_t prev_type = event_null;
    gdouble prev_time = dlg->params.start_time;
    gdouble cur_time = 0;
    gdouble dduration = dlg->params.duration;
    gdouble dwidth = dlg->graph_area->allocation.width;
    gdouble dheight = dlg->graph_area->allocation.height;
    gdouble dstart_time = dlg->params.start_time;
    gdouble disp, baseline;
    smac_node_stream_event_t* cur_event = NULL;
    guint32 prev_x = 0;
    guint32 cur_x = 0;
    guint32 prev_y = 0;
    guint32 cur_y = 0;
    guint32 oneshot_x = SCALE_X((gdouble)SMAC_VIS_ONESHOT_DURATION,
				0, dduration, dwidth);
    GdkFont* fnt = gdk_font_load(SMAC_VIS_FONT_DESCRIPTOR);
    char n_label[8];
    
    baseline = stream->baseline;

    if (dlg->params.filters.h_gridlines) {

        gdk_gc_set_foreground(dlg->gc, &(dlg->colors.gridlines));
    
        gdk_draw_line(dlg->backing_map, dlg->gc,
                      0,
                      SCALE_Y(baseline, 0, dheight),
                      dlg->graph_area->allocation.width,
                      SCALE_Y(baseline, 0, dheight));

        gdk_gc_set_foreground(dlg->gc, &(dlg->colors.gridlabels));
        
        snprintf(n_label, 8, "Node %d", NODE_ID(stream));
        gdk_draw_string(dlg->backing_map, fnt,
                        dlg->gc, SMAC_VIS_EDGE_H_OFFSET,
                        SCALE_Y(baseline + SMAC_VIS_GRID_V_OFFSET, 0, dheight),
                        (char*)(n_label));
        
    }

    for (events = g_list_first(stream->events); events;
         events = g_list_next(events)) {

        cur_event = events->data;
        cur_time = cur_event->timestamp;

        if (smac_visualization_dlg_suppress_event(dlg->params.filters, cur_event->type)) continue;

        if (prev_type == event_null || prev_time == 0) {
            prev_type = cur_event->type;
            prev_time = cur_event->timestamp;
            continue;
        }

        if (cur_event->timestamp < dlg->params.start_time || cur_event->type == event_null) continue;
        if (cur_event->timestamp > (dlg->params.start_time + dlg->params.duration)) break;

        if (prev_x == 0) 
            prev_x = SCALE_X(prev_time, dstart_time, dduration, dwidth);
        if (prev_y == 0) {
            disp = smac_visualization_type_value(prev_type);
            prev_y = SCALE_Y(baseline, disp, dheight);
        }

        cur_x = SCALE_X(cur_time, dstart_time, dduration, dwidth);
        disp = smac_visualization_type_value(cur_event->type);
        cur_y = SCALE_Y(baseline, disp, dheight);

        if (IS_DATAPACKET(cur_event->type)) {
            smac_visualization_dlg_draw_tic(dlg, cur_x, cur_y, cur_event->type);
            continue;
        }

        gdk_gc_set_foreground(dlg->gc, &(dlg->colors.lines));
   
        smac_visualization_dlg_draw_line_wrapper(dlg, dlg->backing_map, dlg->gc,
                                                 prev_x, prev_y, cur_x, prev_y);

        smac_visualization_dlg_draw_line_wrapper(dlg, dlg->backing_map, dlg->gc,
                                                 cur_x, prev_y, cur_x, cur_y);

        if (IS_ONESHOT(cur_event->type)) {

            if (dlg->params.filters.use_color)
                gdk_gc_set_foreground(dlg->gc, smac_visualization_dlg_get_oneshot_lbl_color(cur_event->type, dlg));
            
            smac_visualization_dlg_draw_line_wrapper(dlg, dlg->backing_map, dlg->gc,
                                                     cur_x, cur_y, cur_x + oneshot_x, cur_y);

            smac_visualization_dlg_draw_line_wrapper(dlg, dlg->backing_map, dlg->gc,
                                                     cur_x + oneshot_x, cur_y, cur_x + oneshot_x, prev_y);

            if (dlg->params.filters.use_text) {

                gdk_draw_string(dlg->backing_map, fnt,
                                dlg->gc, cur_x + smac_visualization_dlg_get_oneshot_lbl_offset(cur_event->type),
                                SCALE_Y(baseline + SMAC_VIS_GRID_V_OFFSET, 0, dheight),
                                smac_visualization_dlg_get_oneshot_lbl(cur_event->type));
            }
            
            prev_time = cur_event->timestamp + SMAC_VIS_ONESHOT_DURATION;
            prev_x = cur_x + oneshot_x;
        } else {
            prev_type = cur_event->type;
            prev_time = cur_time;
            prev_x = cur_x;
            prev_y = cur_y;
        }
    }
}


static gint smac_visualization_dlg_cmp_guint16(gconstpointer a, gconstpointer b)
{
    guint16 ai, bi;

    ai = *((guint16*)a);
    bi = *((guint16*)b);

    return !(ai == bi);
}

static GList* smac_visualization_dlg_show_nodes(smac_visualization_dlg_params_t params)
{
    GList* retval = NULL;
    GList* list = NULL;
    GList* sch_list = NULL;
    smac_node_info_t* node;
    smac_schedule_info_t* schedule;

    for (list = g_list_first(params.nodes_list); list; list = g_list_next(list)) {
        node = (smac_node_info_t*)(list->data);
    
        if (g_list_find_custom(params.sources.nodes, &(NODE_ID(node)), 
                               smac_visualization_dlg_cmp_guint16)) {
            retval = g_list_append(retval, &(NODE_ID(node)));
        } else {
      
            for (sch_list = g_list_first(node->schedules_list); sch_list;
                 sch_list = g_list_next(sch_list)) {
                schedule = (smac_schedule_info_t*)(sch_list->data);
		
                if (g_list_find_custom(params.sources.schedules, &(schedule->id),
                                       smac_visualization_dlg_cmp_guint16)) {
                    retval = g_list_append(retval, &(NODE_ID(node)));
                    break;
                }
            }
        }
    }
  
    return retval;
}

static void smac_visualization_dlg_redraw(smac_visualization_dlg_t* dlg)

{
    GList* stream_item;
    GList* show_list = NULL;
    guint16 node_id;
    guint64 vg_time;
    guint32 vg_x;
    GdkFont* fnt = gdk_font_load(SMAC_VIS_FONT_DESCRIPTOR);
    char s_label[8];

    if (!dlg || !(dlg->schedules) || !(dlg->nodes)) return;
    
    gtk_widget_show(dlg->schedules);
    gtk_widget_show(dlg->nodes);
 
    gdk_draw_rectangle (dlg->backing_map,
                        dlg->graph_area->style->white_gc,
                        TRUE,
                        0, 0,
                        dlg->graph_area->allocation.width,
                        dlg->graph_area->allocation.height);

    if(dlg->params.filters.v_gridlines) {
      
        for (vg_time = dlg->params.start_time; 
             vg_time < (dlg->params.start_time + dlg->params.duration);
             vg_time += SEC_TO_USEC(SMAC_VIS_VGRID_SPACING)) {

            gdk_gc_set_foreground(dlg->gc, &(dlg->colors.gridlines));

            
            vg_x = SCALE_X(vg_time, (gdouble)(dlg->params.start_time), 
                           (gdouble)(dlg->params.duration),
                           (gdouble)(dlg->graph_area->allocation.width));

            smac_visualization_dlg_draw_line_wrapper(dlg, dlg->backing_map, dlg->gc,
                                                     vg_x, 0, vg_x, 
                                                     dlg->graph_area->allocation.height);			       

            gdk_gc_set_foreground(dlg->gc, &(dlg->colors.gridlabels));
            
            if (vg_x + 15 < dlg->graph_area->allocation.width) {
                snprintf(s_label, 8, "%ds", USEC_TO_SEC(vg_time - dlg->params.start_time));
                gdk_draw_string(dlg->backing_map, fnt,
                                dlg->gc, vg_x + SMAC_VIS_GRID_H_OFFSET,
                                dlg->graph_area->allocation.height - SMAC_VIS_EDGE_V_OFFSET,
                                (char*)(s_label));
            }
        }
    }

    if(!(dlg->params.sources.show_all)) 
        show_list = smac_visualization_dlg_show_nodes(dlg->params);

    for (stream_item = g_list_first(dlg->params.node_streams); stream_item; 
         stream_item = g_list_next(stream_item)) {

        node_id = NODE_ID((smac_node_stream_t*)(stream_item->data));

        if (dlg->params.sources.show_all || 
            g_list_find_custom(show_list, &node_id, smac_visualization_dlg_cmp_guint16))
            smac_visualization_dlg_draw_stream(dlg, (smac_node_stream_t*)(stream_item->data));
    }

    if (show_list)
        g_list_free(show_list);

    gdk_draw_drawable(dlg->graph_area->window,
                      dlg->graph_area->style->fg_gc[GTK_WIDGET_STATE (dlg->graph_area)],
                      dlg->backing_map,
                      0,0,
                      0,0,
                      dlg->graph_area->allocation.width,
                      dlg->graph_area->allocation.height);  
}

void smac_visualization_dlg_draw(smac_visualization_dlg_t* dlg)
{
    guint32 init_zoom;
                
    if (!dlg || dlg->magic != SMAC_VISUALIZATION_MAGIC) return;

    smac_visualization_dlg_populate_sources(dlg->params, dlg->schedules, dlg->nodes);

    if (dlg->window && dlg->graph_area) {

        gtk_drawing_area_size(GTK_DRAWING_AREA(dlg->graph_area),
                              dlg->graph_area->allocation.width,
                              smac_visualization_dlg_compute_needed_height(dlg->params));
        
        
        init_zoom = NEAREST_GUINT32((gdouble)(USEC_TO_SEC(dlg->params.duration)) *
                                    INITIAL_ZOOM_RATIO);

        init_zoom = (init_zoom > MAX_INIT_ZOOM) ? MAX_INIT_ZOOM : init_zoom;
        
        gtk_adjustment_set_value(dlg->zoom, (gdouble)(init_zoom));
        
    }

    smac_visualization_dlg_redraw(dlg);
}


static gboolean smac_visualization_dlg_configure_event( GtkWidget *widget, 
							GdkEventConfigure *event,
							smac_visualization_dlg_t* dlg)
{
    if (!(dlg->window)) return;

    if (dlg->backing_map)
        g_object_unref(dlg->backing_map);
    

    dlg->backing_map = gdk_pixmap_new(widget->window,
                                      widget->allocation.width,
                                      widget->allocation.height,
                                      -1);


  
    if (dlg->gc)
        g_object_unref(dlg->gc);
  
    dlg->gc = gdk_gc_new(dlg->backing_map);

    gdk_draw_rectangle (dlg->backing_map,
                        widget->style->white_gc,
                        TRUE,
                        0, 0,
                        widget->allocation.width,
                        widget->allocation.height);

  
    smac_visualization_dlg_redraw(dlg);  

    return FALSE;

}

static gboolean smac_visualization_dlg_export_filename_clicked(GtkWidget *widget, 
							       GtkWidget *txt_filename)
{
    GtkWidget* file_dlg;

    file_dlg = gtk_file_chooser_dialog_new ("Choose Export File",
                                            GTK_WINDOW(gtk_widget_get_parent_window(widget)),
                                            GTK_FILE_CHOOSER_ACTION_SAVE,
                                            GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                            GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
                                            NULL);
    //  gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);

    if (gtk_dialog_run (GTK_DIALOG (file_dlg)) == GTK_RESPONSE_ACCEPT)
    {
        char *filename;
    
        filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (file_dlg));
        gtk_entry_set_text(GTK_ENTRY(txt_filename), filename);
        g_free (filename);
    }

    gtk_widget_destroy (file_dlg);

    return FALSE;
}

static gboolean smac_visualization_dlg_export_export_clicked( GtkWidget *widget, 
							      smac_visualization_dlg_export_dlg_t* exdlg )
{
    GtkWidget* errmsg;
    FILE* ofile;
    const gchar* filename = gtk_entry_get_text(GTK_ENTRY(exdlg->txt_filename));

    if (strlen(filename) < 1) {

        errmsg = gtk_message_dialog_new (GTK_WINDOW(exdlg->window),
                                         GTK_DIALOG_MODAL,
                                         GTK_MESSAGE_ERROR,
                                         GTK_BUTTONS_OK,
                                         "You must provide an export file.");

        gtk_dialog_run (GTK_DIALOG (errmsg));
        gtk_widget_destroy (errmsg);
        return FALSE;
    }

    ofile = fopen(filename, "w");
    smac_visualization_dump(ofile, gtk_combo_box_get_active(GTK_COMBO_BOX(exdlg->cb_format)));
    fclose(ofile);
    gtk_widget_destroy(exdlg->window);

    return FALSE;
}

static gboolean smac_visualization_dlg_export_destroy(GtkWidget *widget, 
						      smac_visualization_dlg_export_dlg_t* exdlg)
{
    free(exdlg);
    return TRUE;
}

static GtkWidget* smac_visualization_dlg_export_dlg_create()
{
    smac_visualization_dlg_export_dlg_t* exdlg = (smac_visualization_dlg_export_dlg_t*)malloc(sizeof(smac_visualization_dlg_export_dlg_t));
    GtkWidget* vbox;
    GtkWidget* infogrid;
    GtkWidget* buttonbox;
    GtkWidget* lbl_info;
    GtkWidget* lbl_format;
    GtkWidget* lbl_filename;
    GtkWidget* bt_filename;
    GtkWidget* bt_export;
    GtkWidget* bt_cancel;

    exdlg->window = window_new(GTK_WINDOW_TOPLEVEL, "Export Visualization Data");
    vbox = gtk_vbox_new(FALSE, 10);
    infogrid = gtk_table_new(2, 3, FALSE);
    buttonbox = gtk_hbutton_box_new();
    lbl_info = gtk_label_new("For some formats you may select to output either 'points' or 'events'. A 'points' output is event data plus artificial null events inserted to give the visualization data a square-wave shape when plotted directly with a straight line between each event. Plotting an 'events' ouput directly with result in a less comprehensible sawtooth-like pattern.");
    lbl_format = gtk_label_new("Export format: ");
    lbl_filename = gtk_label_new("Export file: ");
    exdlg->cb_format = gtk_combo_box_new_text();
    exdlg->txt_filename = gtk_entry_new();
    bt_filename = gtk_button_new_with_label("Browse...");
    bt_export = gtk_button_new_with_label("Export");
    bt_cancel = BUTTON_NEW_FROM_STOCK(GTK_STOCK_CANCEL);

    gtk_label_set_line_wrap(GTK_LABEL(lbl_info), TRUE);
    gtk_widget_set_size_request(lbl_info, 400, -1);

    GTK_WIDGET_SET_FLAGS(bt_cancel, GTK_CAN_DEFAULT);
    window_set_cancel_button(exdlg->window, bt_cancel, window_cancel_button_cb);
    gtk_container_set_border_width(GTK_CONTAINER(exdlg->window), 10);

    gtk_combo_box_append_text (GTK_COMBO_BOX (exdlg->cb_format), "jdb tab-delimited ASCII (Events)");
    gtk_combo_box_append_text (GTK_COMBO_BOX (exdlg->cb_format), "gnuplot tab-delimited ASCII (Events)");
    gtk_combo_box_append_text (GTK_COMBO_BOX (exdlg->cb_format), "gnuplot tab-delimited ASCII (Points)");
    gtk_combo_box_set_active (GTK_COMBO_BOX (exdlg->cb_format), exfmt_jdb_events);
 
    gtk_table_attach(GTK_TABLE(infogrid), lbl_format, 0, 1, 0, 1, GTK_FILL, GTK_SHRINK, 5, 5); 
    gtk_table_attach(GTK_TABLE(infogrid), exdlg->cb_format, 1, 2, 0, 1, GTK_FILL, GTK_SHRINK, 5, 5); 
    gtk_table_attach(GTK_TABLE(infogrid), lbl_filename, 0, 1, 1, 2, GTK_FILL, GTK_SHRINK, 5, 5); 
    gtk_table_attach(GTK_TABLE(infogrid), exdlg->txt_filename, 1, 2, 1, 2, GTK_FILL, GTK_SHRINK, 5, 5); 
    gtk_table_attach(GTK_TABLE(infogrid), bt_filename, 2, 3, 1, 2, GTK_FILL, GTK_SHRINK, 5, 5); 
    gtk_box_pack_start(GTK_BOX(buttonbox), bt_cancel, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(buttonbox), bt_export, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), lbl_info, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), infogrid, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), buttonbox, FALSE, FALSE, 0);

    gtk_container_add(GTK_CONTAINER(exdlg->window), vbox);

    gtk_widget_show(lbl_info);
    gtk_widget_show(lbl_format);
    gtk_widget_show(lbl_filename);
    gtk_widget_show(exdlg->cb_format);
    gtk_widget_show(exdlg->txt_filename);
    gtk_widget_show(bt_filename);
    gtk_widget_show(bt_cancel);
    gtk_widget_show(bt_export);
    gtk_widget_show(infogrid);
    gtk_widget_show(buttonbox);
    gtk_widget_show(vbox);

    SIGNAL_CONNECT(bt_filename, "clicked", 
                   smac_visualization_dlg_export_filename_clicked, exdlg->txt_filename);
    SIGNAL_CONNECT(bt_export, "clicked", 
                   smac_visualization_dlg_export_export_clicked, exdlg);
    SIGNAL_CONNECT(exdlg->window, "destroy", smac_visualization_dlg_export_destroy, exdlg);

    return exdlg->window;
}

static gboolean smac_visualization_dlg_expose_event( GtkWidget *widget, 
						     GdkEventExpose *event,
						     smac_visualization_dlg_t* dlg)
{

    gdk_draw_drawable(widget->window,
                      widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                      dlg->backing_map,
                      event->area.x, event->area.y,
                      event->area.x, event->area.y,
                      event->area.width, event->area.height);

    return FALSE;
}

static gboolean smac_visualization_dlg_options_clicked( GtkWidget *widget, 
							GtkWidget *opt_panel )
{
    if (GTK_WIDGET_VISIBLE(opt_panel)) {
        gtk_widget_hide(opt_panel);
        gtk_button_set_label(GTK_BUTTON(widget), "Show Options");
    } else {
        gtk_widget_show(opt_panel);
        gtk_button_set_label(GTK_BUTTON(widget), "Hide Options");
    }
    return FALSE;
}


static gboolean smac_visualization_dlg_export_clicked( GtkWidget *widget, 
                                                       smac_visualization_dlg_t* dlg )
{
    GtkWidget* export_window = smac_visualization_dlg_export_dlg_create();

    gtk_widget_show(export_window);
    gtk_window_set_modal(GTK_WINDOW(export_window), TRUE);

    return FALSE;
}

static gboolean smac_visualization_dlg_print_clicked( GtkWidget *widget, 
                                                      smac_visualization_dlg_t* dlg )
{
    GtkWidget* file_dlg;
    GdkPixbuf* pixbuf = NULL;
   
    file_dlg = gtk_file_chooser_dialog_new ("Save As (PNG format)",
                                            GTK_WINDOW(gtk_widget_get_parent_window(widget)),
                                            GTK_FILE_CHOOSER_ACTION_SAVE,
                                            GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                            GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
                                            NULL);
    //  gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
    
    if (gtk_dialog_run (GTK_DIALOG (file_dlg)) == GTK_RESPONSE_ACCEPT)
    {
        char *filename;
    
        filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (file_dlg));

        pixbuf = gdk_pixbuf_get_from_drawable(NULL, dlg->backing_map,
                                              gtk_widget_get_colormap(dlg->window),
                                              0, 0, 0, 0, -1, -1);
                                                     
        gdk_pixbuf_save(pixbuf, filename, "png", NULL, NULL);

        g_free (filename);
    }

    gtk_widget_destroy (file_dlg);

    return FALSE;
}

static gboolean smac_visualization_dlg_close_clicked( GtkWidget *widget, 
						      smac_visualization_dlg_t* dlg )
{
    gtk_widget_hide(dlg->window);

    return FALSE;
}

static gboolean smac_visualization_dlg_clear_clicked( GtkWidget *widget, 
						      smac_visualization_dlg_t* dlg )
{
    gtk_tree_selection_unselect_all(gtk_tree_view_get_selection(GTK_TREE_VIEW(dlg->schedules)));
    gtk_tree_selection_unselect_all(gtk_tree_view_get_selection(GTK_TREE_VIEW(dlg->nodes)));
    smac_visualization_dlg_source_lst_event(GTK_TREE_VIEW(dlg->schedules), dlg);
    smac_visualization_dlg_source_lst_event(GTK_TREE_VIEW(dlg->nodes), dlg);

    return FALSE;
} 

static gboolean smac_visualization_dlg_source_chk_event(GtkWidget *button, 
							smac_visualization_dlg_t* dlg)
{
    dlg->params.sources.show_all = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button));

    smac_visualization_dlg_redraw(dlg);

    return FALSE;
}

static gboolean smac_visualization_dlg_source_add_to_selection(GtkTreeModel *model,
							       GtkTreePath *path,
							       GtkTreeIter *iter,
							       gpointer sel_list_ptr_arg)
{
    GList** sel_list_ptr = sel_list_ptr_arg;
    guint16 *id = (guint16*)malloc(sizeof(guint16));

    gtk_tree_model_get (model, iter, LC_ID, id, -1);
    *sel_list_ptr = g_list_append(*sel_list_ptr, id);

    return TRUE;
}

static void smac_visualization_dlg_source_lst_event(GtkTreeView* lst_sources,
						    smac_visualization_dlg_t* dlg)
{
    GList* sel_list = NULL;
    GList* list = NULL;
    guint16* id = NULL;
    GtkTreeIter iter;
    GtkTreeModel *model;
    GtkTreeSelection *selection;

    model = gtk_tree_view_get_model(GTK_TREE_VIEW(lst_sources));
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(lst_sources));

    if (!selection) return;

    sel_list = gtk_tree_selection_get_selected_rows(selection, NULL);
  
    if (lst_sources == GTK_TREE_VIEW(dlg->schedules)) {
        if (dlg->params.sources.schedules)
            g_list_free(dlg->params.sources.schedules);
        dlg->params.sources.schedules = (GList*)NULL;
    } else {
        if (dlg->params.sources.nodes)
            g_list_free(dlg->params.sources.nodes);
        dlg->params.sources.nodes = (GList*)NULL;
    }
      
    for (list = g_list_first(sel_list); list; list = g_list_next(list)) {

        id = (guint16*)malloc(sizeof(guint16));

        gtk_tree_model_get_iter(model, &iter, (GtkTreePath*)(list->data));
        gtk_tree_model_get(model, &iter, LC_ID, id, -1);
    
        if (lst_sources == GTK_TREE_VIEW(dlg->schedules)) {
            dlg->params.sources.schedules = g_list_append(dlg->params.sources.schedules, id);
        } else {
            dlg->params.sources.nodes = g_list_append(dlg->params.sources.nodes, id);
        }
    }

    if (sel_list) {
        g_list_foreach (sel_list, (GFunc*)(gtk_tree_path_free), NULL);
        g_list_free (sel_list);
    }


    if (!(dlg->params.sources.show_all))
        smac_visualization_dlg_redraw(dlg);
}

static void smac_visualization_dlg_source_lst_sel_event(GtkTreeSelection* sel_sources,
							smac_visualization_dlg_t* dlg)
{
    smac_visualization_dlg_source_lst_event(gtk_tree_selection_get_tree_view(sel_sources), dlg);
}   

static gboolean smac_visualization_dlg_filter_event(GtkWidget *button, 
						    smac_pair_t* dlg_and_toggle)
{
    smac_visualization_dlg_t* dlg = dlg_and_toggle->first;
    gboolean* toggle_val = dlg_and_toggle->second;

    *toggle_val = !(*toggle_val);

    smac_visualization_dlg_redraw(dlg);

    return FALSE;
}

static gboolean smac_visualization_dlg_zoom_event( GtkWidget *widget,
						   smac_visualization_dlg_t* dlg)
{
    guint zoom = (gint)(gtk_adjustment_get_value(GTK_ADJUSTMENT(widget)));

    gtk_drawing_area_size(GTK_DRAWING_AREA(dlg->graph_area), 
                          BASE_VIEW_WIDTH*zoom, 
                          dlg->graph_area->allocation.height);

    return FALSE;
}

static guint smac_visualization_dlg_compute_needed_height(smac_visualization_dlg_params_t params)
{
    gint max_height = 0;
    gint cur_height = 0;
    GList* list;
    smac_node_stream_t* stream;
    
    for(list = g_list_first(params.node_streams); list; list = g_list_next(list)) {
        stream = (smac_node_stream_t*)(list->data);
        cur_height = -SCALE_Y(stream->baseline + 1, 0, 0);
        max_height = (max_height > cur_height) ? max_height : cur_height;
    }

    return (max_height > BASE_VIEW_HEIGHT) ? max_height : BASE_VIEW_HEIGHT;
}

static void smac_visualization_dlg_populate_sources(smac_visualization_dlg_params_t params,
						    GtkWidget* lst_schedules, GtkWidget* lst_nodes)
{
    GList* list;
    GtkListStore* schedules_store;
    GtkListStore* nodes_store;
    GtkTreeIter s_iter, n_iter;
    smac_schedule_info_t* schedule;
    smac_node_info_t* node;
    gchar* desc_string;
    gboolean new_model = FALSE;
    guint16 node_id;

    schedules_store = GTK_LIST_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(lst_schedules)));

    if (!schedules_store) {
        schedules_store = gtk_list_store_new (LC_NUM_COLUMNS, G_TYPE_STRING, G_TYPE_INT);
        new_model = TRUE;
    }

    for (list = g_list_first(g_list_sort(params.schedules_list,
                                         smac_visualization_cmp_schedules));
         list; list = g_list_next(list)) {

        schedule = (smac_schedule_info_t*)(list->data);

        desc_string = (gchar*)malloc(13);
        sprintf(desc_string, "Schedule %u", schedule->id);

        gtk_list_store_append(schedules_store, &s_iter);
        gtk_list_store_set(schedules_store, &s_iter,
                           LC_DESCRIPTION, desc_string, -1);
        gtk_list_store_set(schedules_store, &s_iter,
                           LC_ID, schedule->id, -1);

        g_free(desc_string);
    }

    nodes_store = GTK_LIST_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(lst_nodes)));

    if (!nodes_store || new_model) {
        nodes_store = gtk_list_store_new (LC_NUM_COLUMNS, G_TYPE_STRING, G_TYPE_INT);
        new_model = TRUE;
    }

  
    for (list = g_list_first(g_list_sort(g_list_first(params.nodes_list),
                                         smac_visualization_cmp_nodes)); 
         list; list = g_list_next(list)) {
  
        node = (smac_node_info_t*)(list->data);
        node_id = NODE_ID(node);

        desc_string = (gchar*)malloc(9);
        sprintf(desc_string, "Node %u", node_id);

        gtk_list_store_append(nodes_store, &n_iter);
        gtk_list_store_set(nodes_store, &n_iter,
                           LC_DESCRIPTION, desc_string, -1);
        gtk_list_store_set(nodes_store, &n_iter,
                           LC_ID, node_id, -1);

        g_free(desc_string);
    }

    if (new_model) {
        gtk_tree_view_set_model(GTK_TREE_VIEW(lst_schedules), GTK_TREE_MODEL(schedules_store));
        gtk_tree_view_set_model(GTK_TREE_VIEW(lst_nodes), GTK_TREE_MODEL(nodes_store));
    
        gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(lst_schedules), 
                                                    -1, "Schedules", gtk_cell_renderer_text_new(),
                                                    "text", LC_DESCRIPTION, NULL);
        gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(lst_schedules), 
                                                    -1, "Schedule IDs", gtk_cell_renderer_text_new(),
                                                    "text", LC_ID, NULL);

        gtk_tree_view_column_set_visible(gtk_tree_view_get_column(GTK_TREE_VIEW(lst_schedules), LC_ID), FALSE);

        gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(lst_nodes), 
                                                    -1, "Nodes", gtk_cell_renderer_text_new(),
                                                    "text", LC_DESCRIPTION, NULL);	
        gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(lst_nodes), 
                                                    -1, "Node IDs", gtk_cell_renderer_text_new(),
                                                    "text", LC_ID, NULL);	

        gtk_tree_view_column_set_visible(gtk_tree_view_get_column(GTK_TREE_VIEW(lst_nodes), LC_ID), FALSE);

        g_object_unref(schedules_store);
        g_object_unref(nodes_store);
    }
}

static void smac_visualization_dlg_sources_setup_controls(GtkWidget* box, smac_visualization_dlg_t* dlg)
{
    GtkWidget* hbox_schedules;
    GtkWidget* hbox_nodes;
    GtkAdjustment* adj_schedules;
    GtkAdjustment* adj_nodes;
    GtkWidget* vsb_schedules;
    GtkWidget* vsb_nodes;
    GtkWidget* chk_all;
    GtkWidget* lst_schedules;
    GtkWidget* lst_nodes;
    GtkWidget* bt_clear;
    GtkTreeSelection* sel_schedules;
    GtkTreeSelection* sel_nodes;
  
    hbox_schedules = gtk_hbox_new(FALSE, 0);
    hbox_nodes = gtk_hbox_new(FALSE, 0);
    adj_schedules = gtk_adjustment_new(0, 0, 100, 1.0, 1.0, 0.0);
    adj_nodes = gtk_adjustment_new(0, 0, 100, 1.0, 1.0, 0.0);
    vsb_schedules = gtk_vscrollbar_new(GTK_ADJUSTMENT(adj_schedules));
    vsb_nodes = gtk_vscrollbar_new(GTK_ADJUSTMENT(adj_nodes));
    chk_all = gtk_check_button_new_with_label("Everything");
    lst_schedules = gtk_tree_view_new();
    lst_nodes = gtk_tree_view_new();
    bt_clear = gtk_button_new_with_label("Clear Selections");

    gtk_tree_view_set_vadjustment(GTK_TREE_VIEW(lst_schedules), GTK_ADJUSTMENT(adj_schedules));
    gtk_tree_view_set_vadjustment(GTK_TREE_VIEW(lst_nodes), GTK_ADJUSTMENT(adj_nodes));

    sel_schedules = gtk_tree_view_get_selection(GTK_TREE_VIEW(lst_schedules));
    sel_nodes = gtk_tree_view_get_selection(GTK_TREE_VIEW(lst_nodes));

    gtk_tree_selection_set_mode(sel_schedules, GTK_SELECTION_MULTIPLE);
    gtk_tree_selection_set_mode(sel_nodes, GTK_SELECTION_MULTIPLE);

    dlg->schedules = lst_schedules;
    dlg->nodes = lst_nodes;
  
    gtk_box_pack_start(GTK_BOX(hbox_schedules), lst_schedules, TRUE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(hbox_schedules), vsb_schedules, FALSE, FALSE, 0);

    gtk_box_pack_start(GTK_BOX(hbox_nodes), lst_nodes, TRUE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(hbox_nodes), vsb_nodes, FALSE, FALSE, 0);

    gtk_box_pack_start(GTK_BOX (box), chk_all, FALSE, FALSE, 0);  
    gtk_box_pack_start(GTK_BOX (box), hbox_schedules, FALSE, FALSE, 0);  
    gtk_box_pack_start(GTK_BOX (box), hbox_nodes, FALSE, FALSE, 0);  
    gtk_box_pack_start(GTK_BOX (box), bt_clear, FALSE, FALSE, 0);

    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chk_all), TRUE);
  
    dlg->params.sources.show_all = TRUE;

    gtk_widget_set_usize(hbox_schedules, -1, 75);
    gtk_widget_set_usize(hbox_nodes, -1, 100);
   
    gtk_widget_show(chk_all);
    gtk_widget_show(lst_schedules);
    gtk_widget_show(lst_nodes);
    gtk_widget_show(vsb_schedules);
    gtk_widget_show(vsb_nodes);
    gtk_widget_show(hbox_schedules);
    gtk_widget_show(hbox_nodes);
    gtk_widget_show(bt_clear);

    SIGNAL_CONNECT(chk_all, "toggled", 
                   smac_visualization_dlg_source_chk_event, 
                   dlg);

    SIGNAL_CONNECT(lst_schedules, "cursor-changed", 
                   smac_visualization_dlg_source_lst_event, 
                   dlg);
  
    SIGNAL_CONNECT(lst_nodes, "cursor-changed", 
                   smac_visualization_dlg_source_lst_event, 
                   dlg);
  
    SIGNAL_CONNECT(gtk_tree_view_get_selection(GTK_TREE_VIEW(lst_schedules)), "changed", 
                   smac_visualization_dlg_source_lst_sel_event, 
                   dlg);
  
    SIGNAL_CONNECT(gtk_tree_view_get_selection(GTK_TREE_VIEW(lst_nodes)), "changed", 
                   smac_visualization_dlg_source_lst_sel_event, 
                   dlg);
  
    SIGNAL_CONNECT(bt_clear, "clicked", 
                   smac_visualization_dlg_clear_clicked, dlg);

}

static void smac_visualization_dlg_filters_setup_controls(GtkWidget* box, smac_visualization_dlg_t* dlg)
{
    GtkWidget* chk_axes;
    GtkWidget* chk_baseline;
    GtkWidget* chk_listen_cycles;
    GtkWidget* chk_sync_packets;
    GtkWidget* chk_rts_cts_packets;
    GtkWidget* chk_data_packets;
    GtkWidget* chk_send_packets;
    GtkWidget* chk_use_color;
    GtkWidget* chk_use_text;

    chk_baseline = gtk_check_button_new_with_label("Horizontal Gridlines");
    chk_axes = gtk_check_button_new_with_label("Vertical Gridlines");
    chk_listen_cycles = gtk_check_button_new_with_label("Scheduled Listens");
    chk_sync_packets = gtk_check_button_new_with_label("SYNC packets");
    chk_rts_cts_packets = gtk_check_button_new_with_label("RTS/CTS packets");
    chk_data_packets = gtk_check_button_new_with_label("Data transfers");
    chk_send_packets = gtk_check_button_new_with_label("Each data packet");
    chk_use_color = gtk_check_button_new_with_label("Colors");
    chk_use_text = gtk_check_button_new_with_label("Text labels");

    gtk_box_pack_start (GTK_BOX (box), chk_baseline, FALSE, FALSE, 0);  
    gtk_box_pack_start (GTK_BOX (box), chk_axes, FALSE, FALSE, 0);  
    gtk_box_pack_start (GTK_BOX (box), chk_listen_cycles, FALSE, FALSE, 0);  
    gtk_box_pack_start (GTK_BOX (box), chk_sync_packets, FALSE, FALSE, 0);  
    gtk_box_pack_start (GTK_BOX (box), chk_rts_cts_packets, FALSE, FALSE, 0);  
    gtk_box_pack_start (GTK_BOX (box), chk_data_packets, FALSE, FALSE, 0);  
    gtk_box_pack_start (GTK_BOX (box), chk_send_packets, FALSE, FALSE, 0);  
    gtk_box_pack_start (GTK_BOX (box), chk_use_color, FALSE, FALSE, 0);  
    gtk_box_pack_start (GTK_BOX (box), chk_use_text, FALSE, FALSE, 0);  

    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chk_baseline), TRUE);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chk_axes), TRUE);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chk_listen_cycles), TRUE);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chk_sync_packets), TRUE);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chk_rts_cts_packets), TRUE);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chk_data_packets), TRUE);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chk_use_color), TRUE);
 
    dlg->params.filters.h_gridlines = TRUE;
    dlg->params.filters.v_gridlines = TRUE; 
    dlg->params.filters.listen_cycles = TRUE;
    dlg->params.filters.sync_packets = TRUE;
    dlg->params.filters.rts_cts_packets = TRUE;
    dlg->params.filters.data_transfers = TRUE;
    dlg->params.filters.data_packets = FALSE;
    dlg->params.filters.use_color = TRUE;
    dlg->params.filters.use_text = FALSE;

    gtk_widget_show(chk_baseline);
    gtk_widget_show(chk_axes);
    gtk_widget_show(chk_listen_cycles);
    gtk_widget_show(chk_sync_packets);
    gtk_widget_show(chk_rts_cts_packets);
    gtk_widget_show(chk_data_packets);
    gtk_widget_show(chk_send_packets);
    gtk_widget_show(chk_use_color);
    gtk_widget_show(chk_use_text);

    SIGNAL_CONNECT(chk_baseline, "toggled", 
                   smac_visualization_dlg_filter_event, 
                   smac_visualization_make_pair(dlg, &(dlg->params.filters.h_gridlines)));
    SIGNAL_CONNECT(chk_axes, "toggled", 
                   smac_visualization_dlg_filter_event, 
                   smac_visualization_make_pair(dlg, &(dlg->params.filters.v_gridlines)));
    SIGNAL_CONNECT(chk_listen_cycles, "toggled", 
                   smac_visualization_dlg_filter_event, 
                   smac_visualization_make_pair(dlg, &(dlg->params.filters.listen_cycles)));
    SIGNAL_CONNECT(chk_sync_packets, "toggled", 
                   smac_visualization_dlg_filter_event, 
                   smac_visualization_make_pair(dlg, &(dlg->params.filters.sync_packets)));
    SIGNAL_CONNECT(chk_rts_cts_packets, "toggled", 
                   smac_visualization_dlg_filter_event, 
                   smac_visualization_make_pair(dlg, &(dlg->params.filters.rts_cts_packets)));
    SIGNAL_CONNECT(chk_data_packets, "toggled", 
                   smac_visualization_dlg_filter_event, 
                   smac_visualization_make_pair(dlg, &(dlg->params.filters.data_transfers)));
    SIGNAL_CONNECT(chk_send_packets, "toggled", 
                   smac_visualization_dlg_filter_event, 
                   smac_visualization_make_pair(dlg, &(dlg->params.filters.data_packets)));
    SIGNAL_CONNECT(chk_use_color, "toggled", 
                   smac_visualization_dlg_filter_event, 
                   smac_visualization_make_pair(dlg, &(dlg->params.filters.use_color)));
    SIGNAL_CONNECT(chk_use_text, "toggled", 
                   smac_visualization_dlg_filter_event, 
                   smac_visualization_make_pair(dlg, &(dlg->params.filters.use_text)));


}

static void smac_visualization_dlg_setup_colors(smac_visualization_dlg_t* dlg)
{
    GdkColormap* cmap = gtk_widget_get_colormap(dlg->window);

    gdk_color_parse("Black", &(dlg->colors.lines));
    gdk_color_parse("Grey", &(dlg->colors.gridlines));
    gdk_color_parse("Grey", &(dlg->colors.gridlabels));
    gdk_color_parse("Blue", &(dlg->colors.rts));
    gdk_color_parse("Green", &(dlg->colors.cts));
    gdk_color_parse("Red", &(dlg->colors.sync));
    gdk_color_parse("Black", &(dlg->colors.bcast));

    gdk_colormap_alloc_color(cmap, &(dlg->colors.lines), FALSE, TRUE);
    gdk_colormap_alloc_color(cmap, &(dlg->colors.gridlines), FALSE, TRUE);
    gdk_colormap_alloc_color(cmap, &(dlg->colors.gridlabels), FALSE, TRUE);
    gdk_colormap_alloc_color(cmap, &(dlg->colors.rts), FALSE, TRUE);
    gdk_colormap_alloc_color(cmap, &(dlg->colors.cts), FALSE, TRUE);
    gdk_colormap_alloc_color(cmap, &(dlg->colors.sync), FALSE, TRUE);
    gdk_colormap_alloc_color(cmap, &(dlg->colors.bcast), FALSE, TRUE);
}

static void smac_visualization_dlg_setup(smac_visualization_dlg_t* dlg)
{
    GtkWidget *vbox;
    GtkWidget *hbox;
    GtkWidget *hbuttonbox;
    GtkWidget *inner_vbox;
    GtkWidget *sidebox;
    GtkWidget *zoom_box;
    GtkWidget *sframe_sources;
    GtkWidget *sframe_filters;
    GtkWidget *sframe_sources_vbox;
    GtkWidget *sframe_filters_vbox;
    GtkTooltips *tooltips = gtk_tooltips_new();
    GtkWidget *lbl_instructions;
    GtkAdjustment *adj_graph_view_h;
    GtkAdjustment *adj_graph_view_v;
    GtkWidget *scw_graph_view;  
    GtkWidget *bt_close;
    GtkWidget *bt_export;
    GtkWidget *bt_print;
    GtkWidget *bt_options;
    GtkWidget *lbl_zoom;
    GtkWidget *spn_zoom;
    
    vbox = gtk_vbox_new(FALSE, 10);
    hbox = gtk_hbox_new(FALSE, 10);

    gtk_container_add(GTK_CONTAINER(dlg->window), vbox);
    gtk_widget_show(vbox);

    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    sidebox = gtk_vbox_new(FALSE,20);
    gtk_container_set_border_width(GTK_CONTAINER(sidebox), 10);

    sframe_sources = gtk_frame_new("Sources");
    sframe_sources_vbox = gtk_vbox_new(FALSE, 10);
    gtk_container_set_border_width(GTK_CONTAINER(sframe_sources_vbox), 10);
    gtk_container_add(GTK_CONTAINER(sframe_sources), sframe_sources_vbox);
    smac_visualization_dlg_sources_setup_controls(sframe_sources_vbox, dlg);

    sframe_filters = gtk_frame_new("Show");
    sframe_filters_vbox = gtk_vbox_new(FALSE, 10);
    gtk_container_set_border_width(GTK_CONTAINER(sframe_filters_vbox), 10);
    gtk_container_add(GTK_CONTAINER(sframe_filters), sframe_filters_vbox);
    smac_visualization_dlg_filters_setup_controls(sframe_filters_vbox, dlg);
 
    gtk_window_set_resizable(GTK_WINDOW(dlg->window), FALSE);
    smac_visualization_dlg_populate_sources(dlg->params, dlg->schedules, dlg->nodes);

    gtk_box_pack_start (GTK_BOX (sidebox), sframe_sources, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (sidebox), sframe_filters, FALSE, FALSE, 0);

    gtk_box_pack_start (GTK_BOX (hbox), sidebox, FALSE, FALSE, 0);

    gtk_widget_show(sframe_sources);
    gtk_widget_show(sframe_sources_vbox);
    gtk_widget_show(sframe_filters);
    gtk_widget_show(sframe_filters_vbox);
    gtk_widget_show(sidebox);

    inner_vbox = gtk_vbox_new(FALSE, 10);
    gtk_container_set_border_width(GTK_CONTAINER(inner_vbox), 10);

    lbl_instructions = gtk_label_new("Flat lines indicate sleep. Spikes above the baseline indicate transmission. Spikes below the baseline indicate listening or reception (lower). Single packets are differentiated as follows: RTS = R (blue), CTS = C (green), SYNC = S (red), Broadcast Data = B (black).");
    gtk_label_set_line_wrap(GTK_LABEL(lbl_instructions), TRUE);
    gtk_widget_set_size_request(lbl_instructions, BASE_VIEW_WIDTH-10, -1);

    adj_graph_view_h = gtk_adjustment_new(50, 0, 100, 5.0, 10.0, 0.0);
    adj_graph_view_v = gtk_adjustment_new(100, 0, 100, 5.0, 10.0, 0.0);

    scw_graph_view = gtk_scrolled_window_new(GTK_ADJUSTMENT(adj_graph_view_h),
                                             GTK_ADJUSTMENT(adj_graph_view_v));

    gtk_widget_set_usize(scw_graph_view, BASE_VIEW_WIDTH, BASE_VIEW_HEIGHT);

    dlg->graph_area = gtk_drawing_area_new();
    gtk_drawing_area_size(GTK_DRAWING_AREA(dlg->graph_area), BASE_VIEW_WIDTH,
                          smac_visualization_dlg_compute_needed_height(dlg->params));

    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scw_graph_view),
                                          dlg->graph_area);

    gtk_box_pack_start (GTK_BOX (inner_vbox), lbl_instructions, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (inner_vbox), scw_graph_view, TRUE, TRUE, 0);

    gtk_box_pack_start (GTK_BOX (hbox), inner_vbox, FALSE, FALSE, 0);
 
    gtk_widget_show(lbl_instructions);
    gtk_widget_show(scw_graph_view);
    gtk_widget_show(inner_vbox);
    gtk_widget_show(dlg->graph_area);

    dlg->backing_map = gdk_pixmap_new(dlg->graph_area->window,
                                      dlg->graph_area->allocation.width,
                                      dlg->graph_area->allocation.height,
                                      -1);

    /* button row */

    hbuttonbox = gtk_hbutton_box_new ();
    gtk_box_pack_start (GTK_BOX (vbox), hbuttonbox, FALSE, FALSE, 0);
    gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox), GTK_BUTTONBOX_SPREAD);
    gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbuttonbox), 5);
    gtk_container_set_border_width(GTK_CONTAINER(hbuttonbox), 10);
    gtk_widget_show(hbuttonbox);
  
    bt_options = gtk_button_new_with_label("Hide Options");
    gtk_container_add(GTK_CONTAINER(hbuttonbox), bt_options);
    gtk_widget_show(bt_options);
    /*  SIGNAL_CONNECT(bt_export, "clicked", on_save_bt_clicked, user_data); */
    gtk_tooltips_set_tip (tooltips, bt_options, "Show/Hide the left-hand panel with viewing options", NULL);

    zoom_box = gtk_hbox_new(FALSE, 5);
        
    lbl_zoom = gtk_label_new("Zoom:");
    gtk_box_pack_start(GTK_BOX(zoom_box), lbl_zoom, FALSE, FALSE, 0);
    gtk_widget_show(lbl_zoom);

    dlg->zoom = gtk_adjustment_new(1.0, 1.0, 200.0, 1.0, 10.0, 0.0);
    spn_zoom = gtk_spin_button_new(GTK_ADJUSTMENT(dlg->zoom), 0.5, 0);
    gtk_box_pack_start(GTK_BOX(zoom_box), spn_zoom, FALSE, FALSE, 0);
    gtk_widget_show(spn_zoom);

    gtk_box_pack_start(GTK_BOX(hbuttonbox), zoom_box, TRUE, FALSE, 0);
    gtk_widget_show(zoom_box);

    bt_export = gtk_button_new_with_label("Export Data");
    gtk_container_add(GTK_CONTAINER(hbuttonbox), bt_export);
    gtk_widget_show(bt_export);
    gtk_tooltips_set_tip (tooltips, bt_export, "Save an image or text dump of the visualization", NULL);

    bt_print = gtk_button_new_with_label("Save as Image");
    gtk_container_add(GTK_CONTAINER(hbuttonbox), bt_print);
    gtk_widget_show(bt_print);
    gtk_tooltips_set_tip (tooltips, bt_print, "Export the visualization graph as an image", NULL);
  
    bt_close = BUTTON_NEW_FROM_STOCK(GTK_STOCK_CLOSE);
    gtk_container_add (GTK_CONTAINER (hbuttonbox), bt_close);
    GTK_WIDGET_SET_FLAGS(bt_close, GTK_CAN_DEFAULT);
    gtk_widget_show(bt_close);
    gtk_tooltips_set_tip (tooltips, bt_close, "Close this dialog", NULL);
    //  window_set_cancel_button(dlg->window, bt_close, window_cancel_button_cb);
 

    SIGNAL_CONNECT(dlg->graph_area, "expose_event", 
                   smac_visualization_dlg_expose_event, dlg);
    SIGNAL_CONNECT(dlg->graph_area, "configure_event", 
                   smac_visualization_dlg_configure_event, dlg);
    SIGNAL_CONNECT(dlg->zoom, "value_changed",
                   smac_visualization_dlg_zoom_event, dlg);
    SIGNAL_CONNECT(bt_options, "clicked", 
                   smac_visualization_dlg_options_clicked, sidebox);
    SIGNAL_CONNECT(bt_export, "clicked", 
                   smac_visualization_dlg_export_clicked, dlg);
    SIGNAL_CONNECT(bt_print, "clicked", 
                   smac_visualization_dlg_print_clicked, dlg);
    SIGNAL_CONNECT(bt_close, "clicked", 
                   smac_visualization_dlg_close_clicked, dlg);

    gtk_widget_set_events (dlg->graph_area, GDK_EXPOSURE_MASK);
  
}

static smac_visualization_dlg_t* smac_visualization_dlg_create()
{
    smac_visualization_dlg_t* dialog = (smac_visualization_dlg_t*)malloc(sizeof(smac_visualization_dlg_t));

    dialog->magic = SMAC_VISUALIZATION_MAGIC;
    dialog->schedules = NULL;
    dialog->nodes = NULL;
    dialog->params.nodes_list = NULL;
    dialog->params.schedules_list = NULL;
    dialog->params.sources.schedules = NULL;
    dialog->params.sources.nodes = NULL;
    dialog->params.node_streams = NULL;
    dialog->gc = NULL;
    dialog->graph_area = NULL;

    /* create the main window */
    dialog->window = window_new(GTK_WINDOW_TOPLEVEL, "S-MAC Network Activity Visualization");
    smac_visualization_dlg_setup(dialog);
    smac_visualization_dlg_setup_colors(dialog);

    /*
      SIGNAL_CONNECT(dialog->window, "delete_event", window_delete_event_cb, NULL);
      SIGNAL_CONNECT(dialog->window, "destroy", on_destroy, user_data);
    */

    return dialog;

}

smac_visualization_dlg_t* smac_visualization_dlg_open(smac_visualization_dlg_t* dlg) 
{
    if (!dlg || !(dlg->window))
        dlg = smac_visualization_dlg_create();
  
    gtk_widget_show(dlg->window);
    /*  window_present(dlg->window); */

    return dlg;
}

static smac_pair_t* smac_visualization_make_pair(void* first, void* second)
{
    smac_pair_t* pair = (smac_pair_t*)malloc(sizeof(smac_pair_t));

    pair->first = first;
    pair->second = second;

    return pair;
}

