/* 
 * $Id: moteradio.c 5884 2006-10-30 22:59:55Z smchenry $
 *
 * moteradio.c -- Present a TinyOS-based mote attached through
 *                a serial port as an network interface device
 *
 * Author: Tyler McHenry (ISI) <smchenry@isi.edu>
 *
 * Note that this is different from SLIP in that it does not
 * expect that the packets are TCP/IP (or even IP!). It will
 * read the packet-begin char, the len, and the CRC, and 
 * deliver everything in between as it is (raw)
 *
 * Based on the SLIP driver by Laurence Culhane and Fred N. 
 * van Kempen. Basically it is a pared-down simplified 
 * version of SLIP that is protocol-agnostic.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/version.h>

#ifdef MOTERADIO_KERNEL_2_4
#include <asm/system.h>
#ifndef MOTERADIO_UML
#include <asm/uaccess.h> 
#endif
#endif

#include <linux/sched.h>
#include <linux/kernel.h>    /* printk()    */
#include <linux/slab.h>      /* kmalloc()   */
#include <linux/errno.h>     /* error codes */
#include <linux/types.h>     /* size_t      */
#include <linux/interrupt.h> /* mark_bh     */

#include <linux/in.h>
#include <linux/in6.h>
#include <linux/netdevice.h>   /* struct device, and other headers */
#include <linux/etherdevice.h> /* eth_type_trans */
#include <linux/tty.h>
#include <linux/rtnetlink.h>
#include <linux/if_arp.h>

#include "moteradio.h" /* parameters and macros are in here */

/* --- Forward Function declarations ---   */
/* For organization and because C is silly */

static int __init moteradio_init_module(void);
static void __exit moteradio_exit_module(void);

static void moteradio_process_char(struct moteradio_priv *priv, unsigned char s);
static void moteradio_process_packet(struct moteradio_priv *priv, struct radio_packet *pkt);

#ifdef MOTERADIO_KERNEL_2_4
static int moteradio_nic_setup(struct net_device *dev);
static int moteradio_nic_ioctl(struct net_device *dev,struct ifreq *rq,int cmd);
#else
static void moteradio_nic_setup(struct net_device *dev);
#endif
static int moteradio_nic_init(struct net_device *dev);
static void moteradio_nic_uninit(struct net_device *dev); 
static int moteradio_nic_open(struct net_device *dev);
static int moteradio_nic_close(struct net_device *dev); 
static int moteradio_nic_change_mtu(struct net_device *dev, int new_mtu);
static struct net_device_stats* moteradio_nic_stats(struct net_device *dev);
static int moteradio_nic_tx(struct sk_buff *skb, struct net_device *dev);

static int moteradio_tty_open(struct tty_struct *tty);
static void moteradio_tty_close(struct tty_struct *tty);
static int moteradio_tty_ioctl(struct tty_struct *tty, struct file *file, unsigned int cmd, unsigned long arg);
#ifdef MOTERADIO_USE_RECEIVE_ROOM
static int moteradio_tty_receive_room(struct tty_struct *tty);
#endif
static void moteradio_tty_receive_buf(struct tty_struct *tty, const unsigned char *cp, char *fp, int count);
static void moteradio_tty_write_wakeup(struct tty_struct *tty);

static void moteradio_help_synchronize(void);
static struct moteradio_priv* moteradio_help_alloc_devs(dev_t line);
static int moteradio_help_alloc_pkt(struct moteradio_priv *priv);
static struct radio_packet* moteradio_help_create_pkt(struct moteradio_priv *priv);
static int moteradio_help_free_pkt(struct moteradio_priv *priv, struct radio_packet *pkt);
static int moteradio_help_crc_check(struct radio_packet *pkt);


/* --- Global Data --- */

struct net_device *moteradio_devs[MOTERADIO_DEVS];

static struct tty_ldisc moteradio_ldisc = {
#ifdef MOTERADIO_KERNEL_2_4
  .flags           = 0,
  .read            = NULL,
  .write           = NULL,
  .poll            = NULL,
#else
  .owner           = THIS_MODULE,
#endif
  .magic           = TTY_LDISC_MAGIC,
  .name            = "moteradio",
  .open            = moteradio_tty_open,
  .close           = moteradio_tty_close,
  .ioctl           = moteradio_tty_ioctl,
  .receive_buf     = moteradio_tty_receive_buf,
#ifdef MOTERADIO_USE_RECEIVE_ROOM
  .receive_room    = moteradio_tty_receive_room,
#endif
  .write_wakeup    = moteradio_tty_write_wakeup,
};

#ifndef MOTERADIO_NO_TIMESTAMP
static struct timeval time_zero;
#endif

/* --- NIC Operations --- */

#ifdef MOTERADIO_KERNEL_2_4
static int
#else
static void 
#endif
moteradio_nic_setup(struct net_device *dev)
{

    MUST_NOT_BE_NULL(dev, NET_DEV_DESC);

/* Required */

  dev->init                = moteradio_nic_init;
  dev->uninit              = moteradio_nic_uninit;
  dev->open                = moteradio_nic_open;
  dev->stop                = moteradio_nic_close;
#ifndef MOTERADIO_KERNEL_2_4
  dev->destructor          = free_netdev;
#else
  dev->do_ioctl            = moteradio_nic_ioctl;
#endif
  dev->hard_start_xmit     = moteradio_nic_tx;
  dev->get_stats           = moteradio_nic_stats; 
  dev->change_mtu          = moteradio_nic_change_mtu;

  /* Parameters */

  dev->type                = ARPHRD_SLIP;
  dev->hard_header_len     = RADIO_HARD_HEADER_LEN;
  dev->addr_len            = 0;
  dev->tx_queue_len        = 0; /* Read-only! */
  dev->mtu                 = RADIO_MTU;

  /* Flags */

  SET_MODULE_OWNER(dev);

  dev->flags              |= IFF_NOARP | IFF_POINTOPOINT | IFF_MULTICAST;
  dev->features           |= NETIF_F_HW_CSUM;

#ifdef MOTERADIO_KERNEL_2_4
  return 0;
#endif

}

static int moteradio_nic_init(struct net_device *dev) 
{

    MUST_NOT_BE_NULL_R(dev, NET_DEV_DESC, -1);
    
#ifdef MOTERADIO_KERNEL_2_4
  return moteradio_nic_setup(dev);
#endif
  
  dev->mtu = RADIO_MTU;

  return 0;
}

static void moteradio_nic_uninit(struct net_device *dev) 
{
    struct moteradio_priv *priv = NULL;

    MUST_NOT_BE_NULL(dev, NET_DEV_DESC);
    
    priv = netdev_priv(dev);

    MUST_NOT_BE_NULL(priv, PRIV_DESC);
    MUST_MATCH_MAGIC(priv, PRIV_DESC);
    MUST_NOT_BE_NULL(priv->nic.rx_queue, RXQ_DESC);
    MUST_NOT_BE_NULL(&priv->lock, LOCK_DESC);
    
    MR_LOCK;

    if (priv->nic.rx_queue != NULL)
        moteradio_help_free_pkt(priv, priv->nic.rx_queue);

    MR_UNLOCK;
}

static int moteradio_nic_open(struct net_device *dev)
{
    struct moteradio_priv* priv = NULL;

  MUST_NOT_BE_NULL_R(dev, NET_DEV_DESC, -ENODEV);

  priv = netdev_priv(dev);
  
  MUST_NOT_BE_NULL_R(priv, PRIV_DESC, -ENODEV);
  MUST_MATCH_MAGIC_R(priv, PRIV_DESC, -ENODEV);
  MUST_NOT_BE_NULL_R(priv->tty.dev, TTY_DEV_DESC, -ENODEV);

  priv->flags &= (1 << SLF_INUSE);
  
  netif_start_queue(dev);
  return 0;
}

static int moteradio_nic_change_mtu(struct net_device *dev,
				    int new_mtu) 
{
  /* Static MTU for simplicity ATM */

  if (new_mtu != RADIO_MTU) return -EINVAL;
  else return 0;
}

static int moteradio_nic_close(struct net_device *dev) 
{
    struct moteradio_priv *priv = NULL;

  MUST_NOT_BE_NULL_R(dev, NET_DEV_DESC, -1);

  priv = netdev_priv(dev);
  
  MUST_NOT_BE_NULL_R(priv, PRIV_DESC, -1);
  MUST_MATCH_MAGIC_R(priv, PRIV_DESC, -1);
  
  MR_LOCK;
  if (priv->tty.dev) {
    priv->flags &= ~(1 << TTY_DO_WRITE_WAKEUP);
  }
  MR_UNLOCK;
  
  netif_stop_queue(dev);

  MR_LOCK;
  while(priv->nic.rx_queue != NULL)
    moteradio_help_free_pkt(priv, priv->nic.rx_queue);
  MR_UNLOCK;


  return 0;
}

static struct net_device_stats * 
moteradio_nic_stats(struct net_device *dev)
{
	static struct net_device_stats stats;
	struct moteradio_priv *priv = NULL;

        MUST_NOT_BE_NULL_R(dev, NET_DEV_DESC, NULL);
        
        priv = netdev_priv(dev);

        MUST_NOT_BE_NULL_R(priv, PRIV_DESC, NULL);
        MUST_MATCH_MAGIC_R(priv, PRIV_DESC, NULL);
        
	memset(&stats, 0, sizeof(struct net_device_stats));

	/* Don't just return a pointer to our internal stats
	   because they are volitile as far as the calling
	   function is concerned */

	stats.rx_packets     = priv->nic.stats.rx_packets;
	stats.tx_packets     = priv->nic.stats.tx_packets;
	stats.rx_bytes	     = priv->nic.stats.rx_bytes;
	stats.tx_bytes	     = priv->nic.stats.tx_bytes;
	stats.rx_dropped     = priv->nic.stats.rx_dropped;
	stats.tx_dropped     = priv->nic.stats.tx_dropped;
	stats.tx_errors      = priv->nic.stats.tx_errors;
	stats.rx_errors      = priv->nic.stats.rx_errors;
	stats.rx_over_errors = priv->nic.stats.rx_over_errors;

	return (&stats);
}

static int moteradio_nic_tx(struct sk_buff *skb, struct net_device *dev)
{
  /* Read-only ATM */
  return -EINVAL;
}


#ifdef MOTERADIO_KERNEL_2_4 

static int moteradio_nic_ioctl(struct net_device *dev,struct ifreq
			       *rq,int cmd)
{
  debug_printk_args("Received NIC ioctl: %d", cmd);

  return -ENOIOCTLCMD;
}

#endif


/* --- Line Discipline (TTY) Operations --- */

static int moteradio_tty_open(struct tty_struct *tty) 
{
  struct moteradio_priv *priv;
  int err;

  MUST_NOT_BE_NULL_R(tty, TTY_DESC, -ENFILE);
  
#ifndef MOTERADIO_KERNEL_2_4
  if (!capable(CAP_NET_ADMIN)) 
    return -EPERM;
#else
  MOD_INC_USE_COUNT;
#endif

  rtnl_lock();

  moteradio_help_synchronize();

  priv = (struct moteradio_priv *)(tty->disc_data);

  if(priv && priv->magic == MOTERADIO_MAGIC) {
    err = -EEXIST;
    goto err_exit; /* Already connected */
  }

  if((priv = moteradio_help_alloc_devs(tty_devnum(tty))) 
     == NULL) {
    printk(KERN_ERR "Could not grab a TTY!\n"); 
    err = -ENFILE;
    goto err_exit; /* Could not grab a TTY */
  }

  MUST_NOT_BE_NULL_R(priv, PRIV_DESC, -1);
  MUST_MATCH_MAGIC_R(priv, PRIV_DESC, -1);
  
  priv->tty.dev = tty;
  tty->disc_data = priv;
  priv->tty.line = tty_devnum(tty);
  priv->tty.pid = current->pid;

#ifndef MOTERADIO_KERNEL_2_4
  if(priv->tty.dev->driver->flush_buffer)
    tty->driver->flush_buffer(tty);
#else
  if(priv->tty.dev->driver.flush_buffer)
    tty->driver.flush_buffer(tty);
#endif

  if(!test_bit(SLF_INUSE, &priv->flags)) {
     
    set_bit(SLF_INUSE, &priv->flags);

    if((err = register_netdevice(priv->nic.dev)))
      goto err_free_chan; /* Kernel not accepting new NICs? */
  }

  rtnl_unlock();
  return priv->nic.dev->base_addr;

 err_free_chan:
  priv->tty.dev = NULL;
  tty->disc_data = NULL;
  clear_bit(SLF_INUSE, &priv->flags);

 err_exit:
  rtnl_unlock();

#ifdef MOTERADIO_KERNEL_2_4
  MOD_DEC_USE_COUNT;
#endif

  return err;
}

static void moteradio_tty_close(struct tty_struct *tty) 
{
  struct moteradio_priv *priv = (struct moteradio_priv *)(tty->disc_data);

  MUST_NOT_BE_NULL(tty, TTY_DESC);
  MUST_NOT_BE_NULL(priv, PRIV_DESC);
  MUST_MATCH_MAGIC(priv, PRIV_DESC);
  
  /* Sanity check */
  if(priv->tty.dev != tty) return;

  tty->disc_data = NULL;
  priv->tty.dev = NULL;
  priv->tty.line = 0;

#ifdef MOTERADIO_KERNEL_2_4
  MOD_DEC_USE_COUNT;
#endif

}

static int moteradio_tty_ioctl(struct tty_struct *tty,
			       struct file *file,
			       unsigned int cmd,
			       unsigned long arg) 
{
  struct moteradio_priv *priv = (struct moteradio_priv *)(tty->disc_data);
  unsigned int tmp;

  debug_printk_args("Received TTY ioctl: %d", cmd);

  MUST_NOT_BE_NULL_R(tty, TTY_DESC, -EINVAL);
  MUST_NOT_BE_NULL_R(file, FILE_DESC, -EINVAL);
  MUST_NOT_BE_NULL_R(priv, PRIV_DESC, -EINVAL);
  MUST_MATCH_MAGIC_R(priv, PRIV_DESC, -EINVAL);

  switch(cmd) {
  
  case SIOCGIFNAME:
    tmp = strlen(priv->nic.dev->name) + 1;
#ifndef MOTERADIO_KERNEL_2_4
    if (copy_to_user((void __user*)(arg), priv->nic.dev->name, tmp))
      return -EFAULT;
#else
#ifndef MOTERADIO_UML
    if (copy_to_user((void*)(arg), priv->nic.dev->name, tmp))
      return -EFAULT;
#else
    return -EINVAL;
#endif
#endif
    return 0;

  case SIOCGIFENCAP:  /* No encapsulation */
  case SIOCSIFENCAP:
  case SIOCSIFHWADDR: /* NIC is fake, remember? */
    return -EINVAL;

  case TCGETS:
  case TCGETA:
    return n_tty_ioctl(tty, file, cmd, arg);

  default:
    return -ENOIOCTLCMD;
  }
}

#ifdef MOTERADIO_USE_RECEIVE_ROOM

static int moteradio_tty_receive_room(struct tty_struct *tty)
{
  return 65536; /* Inf */
}

#endif

static void moteradio_tty_receive_buf(struct tty_struct *tty,
					const unsigned char *cp,
					char *fp, int count)
{
    struct moteradio_priv *priv = NULL;
    
    MUST_NOT_BE_NULL(tty, TTY_DESC);

    priv = (struct moteradio_priv*)(tty->disc_data);

    MUST_NOT_BE_NULL(priv, PRIV_DESC);
    MUST_MATCH_MAGIC(priv, PRIV_DESC);
    
    if (!netif_running(priv->nic.dev)) return;

      
    while(count--) {
        if (fp && *fp++) {
            
            if (!test_and_set_bit(SLF_ERROR, &priv->flags)) 
                priv->nic.stats.rx_errors++;
        
            cp++;
            continue;
        }

        moteradio_process_char(priv, *cp++);
    }
}

static void moteradio_tty_write_wakeup(struct tty_struct *tty)
{
  /* Read-only ATM */
}

/* --- Helper Functions --- */

static void moteradio_help_synchronize(void)
{
  int i;
  struct moteradio_priv *priv;

  for (i=0; i < MOTERADIO_DEVS && moteradio_devs[i] != NULL; i++) {
    priv = netdev_priv(moteradio_devs[i]);

    MUST_NOT_BE_NULL(priv, PRIV_DESC);
    MUST_MATCH_MAGIC(priv, PRIV_DESC);
    
    if (!(priv->tty.dev) && moteradio_devs[i]->flags&IFF_UP)
      dev_close(moteradio_devs[i]);
  }
}

static struct moteradio_priv* moteradio_help_alloc_devs(dev_t line)
{
  int i;
  int sel = -1;
  int score = -1;
  struct net_device *dev = NULL;
  struct moteradio_priv *priv = NULL;

#ifdef MOTERADIO_KERNEL_2_4
  moteradio_ctrl_t *ctrl = NULL;
#endif

  /* Selection algorithm is same as SLIP */

  for (i = 0; i < MOTERADIO_DEVS; i++) {
    if((dev = moteradio_devs[i]) == NULL) break;
    
    priv = netdev_priv(dev);

    MUST_NOT_BE_NULL_R(priv, PRIV_DESC, NULL);
    MUST_MATCH_MAGIC_R(priv, PRIV_DESC, NULL);
    
    if (priv->tty.dev) continue; /* already alloc'ed */

    if (current->pid == priv->tty.pid) {
      if (priv->tty.line == line && score < 3) {
	sel = i;
	score = 3;
	continue;
      }
      if (score < 2) {
	sel = i;
	score = 2;
      }
      continue;
    }
    
    if (priv->tty.line == line && score < 1) {
      sel = i;
      score = 1;
      continue;
    }

    if (score < 0) {
      sel = i;
      score = 0;
    }
  }

  if (sel >= 0) {
    i = sel;
    dev = moteradio_devs[i];

    if (score > 1) {
      priv = netdev_priv(dev);
      priv->flags &= (1 << SLF_INUSE);
      return priv;
    }
  }

  if (i >= MOTERADIO_DEVS)
    return NULL; /* out of slots */

  if (dev) {
    priv = netdev_priv(dev);

    MUST_NOT_BE_NULL_R(priv, PRIV_DESC, NULL);
    MUST_MATCH_MAGIC_R(priv, PRIV_DESC, NULL);
    
    if (test_bit(SLF_INUSE, &priv->flags)) {

      /* currently, the unregister call causes a segfault
	 and the module fails to unload (in 2.4). Temporarily removed. */
        debug_printk("Unregistering net device...");
        unregister_netdevice(dev);
        debug_printk("Net device successfully unregistered");

      dev = NULL;
      moteradio_devs[i] = NULL;
    }
  } 

  if (!dev) {

#ifndef MOTERADIO_KERNEL_2_4

    char name[IFNAMSIZ];
    sprintf(name, "mr%d", i);
    dev = alloc_netdev(sizeof(*priv), name, moteradio_nic_setup);
    if (!dev) return NULL;

#else

    if ((ctrl = (moteradio_ctrl_t *)kmalloc(sizeof(moteradio_ctrl_t),GFP_KERNEL)) 
	== NULL)
      return NULL;

    memset(ctrl, 0, sizeof(moteradio_ctrl_t));

    priv = &ctrl->priv;
    dev = &ctrl->dev;

    sprintf(dev->name, "mr%d", i);
    
    dev->priv = (void*)priv;
    dev->init = moteradio_nic_init;

#endif 

    dev->base_addr = i;
  }
  
  priv = netdev_priv(dev);

  memset(priv, 0, sizeof(struct moteradio_priv));

#ifdef MOTERADIO_KERNEL_2_4
  priv->container = (void*)ctrl;
#endif

  priv->magic = MOTERADIO_MAGIC;
  priv->nic.dev = dev;
 
  spin_lock_init(&priv->lock);

  moteradio_devs[i] = dev;

  return priv;

}

/* 
   It is important to note a fundamental difference between the
   way that this driver buffers incoming packets versus the way
   that SLIP does so. SLIP has one giant array buffer that holds
   raw TCP/IP data.

   This driver, on the other hand, uses a linked-list of packet
   structs which hold separated data since we must do a small
   bit of decapsulation and CRC checking before passing up the 
   protocol stack.

   The linked list approach allows this driver (on SMP or
   pre-emptive systems) to simultaneously process one packet
   and be receiving subsequent packets.
*/

static int moteradio_help_alloc_pkt(struct moteradio_priv *priv)
{
  struct radio_packet *new_pkt;
  struct radio_packet *cur_pkt;
#ifndef MOTERADIO_NO_TIMESTAMP
  struct timeval tv;
#endif

  /* !! THIS FUNCTION ASSUMES WE HAVE THE LOCK !! */
  
  MUST_NOT_BE_NULL_R(priv, PRIV_DESC, -1);
  MUST_MATCH_MAGIC_R(priv, PRIV_DESC, -1);

  MR_UNLOCK;
  
  new_pkt = kmalloc(sizeof(struct radio_packet), GFP_KERNEL);
  if (new_pkt == NULL) {
    debug_printk("Alloc packet couldn't, uh, alloc a packet.");
    return -ENOBUFS;
  }

  MR_LOCK;
  
  new_pkt->magic = MOTERADIO_MAGIC;
  
  new_pkt->next = NULL;
  new_pkt->dev = priv->nic.dev;

  memset(&new_pkt->data, 0, RADIO_MAX_DATA_LEN);
  memset(&new_pkt->radio_crc, 0, RADIO_RADIO_CRC_LEN);

  new_pkt->flags = 0;
  new_pkt->datalen = 0;
  new_pkt->rcount = 0;
  
#ifndef MOTERADIO_NO_TIMESTAMP
  do_gettimeofday(&tv);
  new_pkt->timestamp = (uint64_t)(((tv.tv_sec - time_zero.tv_sec) *
				   USECS_PER_SEC) + 
				  ((int64_t)(tv.tv_usec) - 
				   (int64_t)(time_zero.tv_usec)));
#endif

  if (priv->nic.rx_queue == NULL) {
      new_pkt->prev = NULL;
      priv->nic.rx_queue = new_pkt;

  } else {

      MUST_NOT_BE_NULL_R(priv->nic.rx_queue, RXQ_DESC, -1);
          
      cur_pkt = priv->nic.rx_queue;
      while(cur_pkt->next != NULL) cur_pkt = cur_pkt->next;
      new_pkt->prev = cur_pkt;
      cur_pkt->next = new_pkt;
  }

  return 0;
}

static struct radio_packet* moteradio_help_create_pkt(struct moteradio_priv *priv)
{
    /* This is a wrapper function that calls alloc_pkt, prints a debug
       message on error, and initializes a few basic fields */
    
    struct radio_packet* pkt = NULL;

    MUST_NOT_BE_NULL_R(priv, PRIV_DESC, NULL);
    MUST_MATCH_MAGIC_R(priv, PRIV_DESC, NULL);
    
    if(moteradio_help_alloc_pkt(priv) != 0) return NULL;
    pkt = priv->nic.rx_queue;
    
    MUST_NOT_BE_NULL_R(pkt, PKT_DESC, NULL);
    
    pkt->flags = 0;
    set_bit(PKF_BEGUN, &pkt->flags);

    return pkt;
}

static int moteradio_help_free_pkt(struct moteradio_priv *priv,
				   struct radio_packet *pkt)
{

    /* !! THIS FUNCTION ASSUMES WE HAVE THE LOCK !! */

    MUST_NOT_BE_NULL_R(priv, PRIV_DESC, -1);
    MUST_NOT_BE_NULL_R(pkt, PKT_DESC, -1);
    MUST_MATCH_MAGIC_R(priv, PRIV_DESC, -1);

#ifndef MOTERADIO_NO_MAGIC_CHECKS
    if (pkt && pkt->magic != MOTERADIO_MAGIC) {
        if (pkt == priv->nic.rx_queue)
            priv->nic.rx_queue = NULL;
        return -1;
    }
    
    if (pkt == priv->nic.rx_queue ||
        (pkt->prev && pkt->prev->magic != MOTERADIO_MAGIC))
        pkt->prev = NULL;
    else if (pkt->next && pkt->next->magic != MOTERADIO_MAGIC)
        pkt->next = NULL;
#endif
    
    if(pkt->prev == NULL)
        priv->nic.rx_queue = pkt->next;
    else
        pkt->prev->next = pkt->next;
    
    if(pkt->next != NULL) 
        pkt->next->prev = pkt->prev;

    MUST_MATCH_MAGIC_R(pkt, PKT_DESC, -1);

    pkt->magic = 0;

    MR_UNLOCK;
    if (pkt) kfree(pkt);
    MR_LOCK;
    
    pkt = NULL;
    
    return 0;
}

static int moteradio_help_crc_check(struct radio_packet *pkt) 
{
  int i,j;
  u16 computed_crc = 0;
  u16 given_crc;
  u16 tmp;

  MUST_NOT_BE_NULL_R(pkt, PKT_DESC, -1);
  
  given_crc = ntohl(*((u16*)(pkt->radio_crc)));

  for (i = -1; i < pkt->datalen; i++) {
    if (i < 0) {
      tmp = (u16)(pkt->datalen + RADIO_SOFT_HEADER_LEN + RADIO_RADIO_CRC_LEN +
                  RADIO_SOFT_TRAILER_LEN + RADIO_HARD_TRAILER_LEN);
    } else {
      tmp = (u16)(pkt->data[i]);
    }

    computed_crc = computed_crc ^ (tmp << 8);

    for (j = 0; j < 8; j++) {
      if (computed_crc & 0x8000)
	computed_crc = computed_crc << 1 ^ 0x1021; /* << before ^ */
      else
	computed_crc = computed_crc << 1;
    }
  }

  return (computed_crc == given_crc) ? 0 : -1;
}

/* --- Packet Processing --- */
/* Here are the main differences from SLIP! */

static void moteradio_process_char(struct moteradio_priv *priv,
				   unsigned char s) 
{

  /* Note that we need not lock the structure since the current
     incoming packet will always be the head of the list, and
     two packets cannot arrive at once. */

  struct radio_packet *pkt;
  int signed_datalen;

  MUST_NOT_BE_NULL(priv, PRIV_DESC);
  MUST_MATCH_MAGIC(priv, PRIV_DESC);

  debug_verbose_printk_args("Received byte: 0x%X.", s);

  MUST_NOT_BE_NULL(&priv->lock, LOCK_DESC);

  // rx_queue is allowed to be null here. pkt is checked before being
  // referenced and is constructed if necessary.
  
  MR_LOCK; // Lock should stay on while in this function

  pkt = priv->nic.rx_queue;
  priv->nic.stats.rx_bytes++;

  /* First process control characters */
  
#ifdef MOTERADIO_BYTE_STUFFING
  
  if (!test_bit(STF_STUFFED, &priv->stream_flags)) {
      if (s == CHAR_STUFF) {
          debug_verbose_printk("Byte interpreted as STUFF byte.");
          set_bit(STF_STUFFED, &priv->stream_flags);
          MR_UNLOCK_AND_RETURN;
      }
  } else {
      if (s == CHAR_STUFF) 
          debug_verbose_printk("Got stuffed STUFF byte.");
  }
  
#endif
  
  if (s == CHAR_BEGIN) {
      
    debug_verbose_printk("Byte interpreted as BEGIN byte.");
    
    /* If we get a BEGIN byte there are several possibilities. This is a
       map to the complex series of if statements below.

       A. There is no current packet

          1. STUFF is on  : in-media-res of old packet, discard byte
          2. STUFF is off : clean BEGIN, start new packet
          3. No Stuffing  : clean BEGIN, start new packet

       B. There is a current packet

          1. ERROR is on

             a. STUFF is on  : Persist error, discard byte
             b. STUFF is off : Discard packet, start new packet
             c. No Stuffing  : Discard packet, start new packet

          2. ERROR is off

             a. STUFF is on  : Process as data
             b. STUFF is off : Discard packet, start new packet
             c. No Stuffing  : Process as data

      Some of the if conditions are dependant on the snooper features
      that the driver is built to support. Sorry about all of the ifdefs.
      
    */

    if(pkt != NULL) {
      if(test_and_set_bit(PKF_BEGUN, &pkt->flags) &&
#ifndef MOTERADIO_META_FLAGS
  	  !test_bit(PKF_GOTRADIOCRC, &pkt->flags)) {
#else
          !test_bit(PKF_GOTSERIALCRC, &pkt->flags)) {
#endif

          /* B */
          
	  if(test_and_clear_bit(PKF_ERROR, &pkt->flags)) {

              /* B-1 */


#ifdef MOTERADIO_BYTE_STUFFING
#define MR_B1A_TEST test_and_clear_bit(STF_STUFFED, &priv->stream_flags)
#else
#define MR_B1A_TEST 0
#endif
              if (MR_B1A_TEST) {

                  /* B-1-a */

                  debug_printk("Error in cur packet, stream may be screwy.");
                  set_bit(PKF_ERROR, &pkt->flags);
                  goto begin_fallthrough;
                  
              } else {

                  /* B-1-b and B-1-c */
                  
                  debug_printk("Error in prev packet, stream may be screwy; discarding current packet");

                  moteradio_help_free_pkt(priv, pkt);
                  pkt = moteradio_help_create_pkt(priv);
                  if (!pkt) MR_UNLOCK_AND_RETURN;
                  
              }
              
#ifdef MOTERADIO_BYTE_STUFFING
#define MR_B2A_TEST test_and_clear_bit(STF_STUFFED, &priv->stream_flags)
#else
#define MR_B2A_TEST 0
#endif
          } else if (MR_B2A_TEST) {

              /* B-2-a */
              
              debug_verbose_printk("Got stuffed BEGIN byte");
              
              goto begin_fallthrough;

	  } else {

#ifdef MOTERADIO_BYTE_STUFFING
              
              /* B-2-b */

              debug_printk("Got an (unstuffed) BEGIN byte inside of packet; discarding current packet.");
              priv->nic.stats.rx_errors++;
              
              moteradio_help_free_pkt(priv, pkt);
              pkt = moteradio_help_create_pkt(priv);
              if (!pkt) MR_UNLOCK_AND_RETURN;
              
#else
              
              /* B-2-c */
              
              debug_verbose_printk("This BEGIN seems to be DATA.");
              
              goto begin_fallthrough;
#endif

          }
      }
    }

/* A */
#ifdef MOTERADIO_BYTE_STUFFING    
    if (test_and_clear_bit(STF_STUFFED, &priv->stream_flags)) {

        /* A-1 */

        debug_verbose_printk("Got a stuffed BEGIN outside of a packet; discarding.");
        MR_UNLOCK_AND_RETURN;
    }
#endif

    /* A-2 and A-3 */

    pkt = moteradio_help_create_pkt(priv);
    MR_UNLOCK_AND_RETURN;
    
  }
  
/* Below here is the processing of non-control characters */
  
begin_fallthrough:

  
  // we have the lock at this point!!
    
#ifdef MOTERADIO_BYTE_STUFFING
  clear_bit(STF_STUFFED, &priv->stream_flags);
#endif

  if (pkt == NULL) {
    debug_printk("Recieved data without packet header.");
    MR_UNLOCK_AND_RETURN; /* Zuh? Data without BEGIN */
  }

  if (test_bit(PKF_ERROR, &pkt->flags)) {
    debug_printk("Discarding byte due to previous errors in this packet.");
    MR_UNLOCK_AND_RETURN;
  }

  if (!test_bit(PKF_BEGUN, &pkt->flags)) {
    debug_printk("Somehow got to a state in which we have a non-null packet pointer but have not received the begin byte. Halp!");
    set_bit(PKF_ERROR, &pkt->flags);
    MR_UNLOCK_AND_RETURN; /* Now this really doesn't make sense, but eh */
  }

  if (!test_bit(PKF_GOTLEN, &pkt->flags)) {
    /* This must be the length byte */

    debug_verbose_printk_args("Byte interpreted as length byte. (length: %d) \n", s);

    signed_datalen = (unsigned int)(s) - RADIO_SOFT_HEADER_LEN - RADIO_RADIO_CRC_LEN -
        RADIO_SOFT_TRAILER_LEN;
    pkt->datalen = (signed_datalen > 0) ? signed_datalen : 0;
    set_bit(PKF_GOTLEN, &pkt->flags);

    if (pkt->datalen > RADIO_MAX_DATA_LEN || pkt->datalen == 0) {
        debug_printk("Packet reported unreasonable length. Ignoring.");
        set_bit(PKF_ERROR, &pkt->flags);
    }
    
    MR_UNLOCK_AND_RETURN;
  }
  
  if (!test_bit(PKF_GOTDATA, &pkt->flags)) {
      /* This must be the payload data */
      
      debug_verbose_printk_args("Byte interpreted as data byte (%u of %u).", pkt->rcount + 1, pkt->datalen);
      
      if (!test_bit(PKF_ERROR, &pkt->flags)) {
          if (pkt->rcount < RADIO_MAX_DATA_LEN) {
              pkt->data[pkt->rcount++] = s;
          } else {
              priv->nic.stats.rx_over_errors++;
              set_bit(PKF_ERROR, &pkt->flags);
          }
          
          if(pkt->rcount == pkt->datalen) 
              set_bit(PKF_GOTDATA, &pkt->flags);
      }
      MR_UNLOCK_AND_RETURN;
  }

  if (!test_bit(PKF_GOTRADIOCRC, &pkt->flags)) {
    /* This must be one of the CRC bytes */

    debug_verbose_printk("Byte interpreted as Radio CRC byte.");
    
    if (!test_bit(PKF_ERROR, &pkt->flags)) {
      if (pkt->rcount < RADIO_MAX_DATA_LEN + RADIO_RADIO_CRC_LEN) {
	pkt->radio_crc[(pkt->rcount++) - pkt->datalen] = s;
      } else {
	priv->nic.stats.rx_over_errors++;
	set_bit(PKF_ERROR, &pkt->flags);
      }
      
      if(pkt->rcount == pkt->datalen + RADIO_RADIO_CRC_LEN) {
	set_bit(PKF_GOTRADIOCRC, &pkt->flags);
#ifndef MOTERADIO_META_FLAGS
        moteradio_process_packet(priv, pkt);
#endif
      } 
    }
    MR_UNLOCK_AND_RETURN;
  }

#ifdef MOTERADIO_META_FLAGS
  if (!test_bit(PKF_GOTTRAILER, &pkt->flags)) {
      /* This must be a trailer byte */

      debug_verbose_printk("Byte interpreted as trailer byte.");
      
    if (!test_bit(PKF_ERROR, &pkt->flags)) {
        if (pkt->rcount < RADIO_MAX_DATA_LEN + RADIO_RADIO_CRC_LEN +
            RADIO_SOFT_TRAILER_LEN + RADIO_HARD_TRAILER_LEN) {
            pkt->trailer[(pkt->rcount++) - pkt->datalen - RADIO_RADIO_CRC_LEN] = s;
        } else {
            priv->nic.stats.rx_over_errors++;
            set_bit(PKF_ERROR, &pkt->flags);
        }        
    }

    if(pkt->rcount == pkt->datalen + RADIO_RADIO_CRC_LEN +
       RADIO_SOFT_TRAILER_LEN + RADIO_HARD_TRAILER_LEN) {
        set_bit(PKF_GOTTRAILER, &pkt->flags);
        if (RADIO_SERIAL_CRC_LEN == 0) {
            moteradio_process_packet(priv, pkt);
        }
    } else {
        debug_verbose_printk_args("NOT setting PKF_GOTTRAILER because %d != %d",
                                  pkt->rcount, pkt->datalen + RADIO_RADIO_CRC_LEN +
                                  RADIO_SOFT_HEADER_LEN + RADIO_HARD_HEADER_LEN);
    }

    MR_UNLOCK_AND_RETURN;
  }

  if (!test_bit(PKF_GOTSERIALCRC, &pkt->flags)) {
      /* This must be a serial CRC byte */

      debug_verbose_printk("Byte interpreted as Serial CRC byte.");
      
    if (!test_bit(PKF_ERROR, &pkt->flags)) {
        if (pkt->rcount < RADIO_MAX_DATA_LEN + RADIO_RADIO_CRC_LEN +
            RADIO_SOFT_TRAILER_LEN + RADIO_HARD_TRAILER_LEN +
            RADIO_SERIAL_CRC_LEN) {
            pkt->serial_crc[(pkt->rcount++) - pkt->datalen - RADIO_RADIO_CRC_LEN -
                            RADIO_SOFT_TRAILER_LEN - RADIO_HARD_TRAILER_LEN] = s;
        } else {
            priv->nic.stats.rx_over_errors++;
            set_bit(PKF_ERROR, &pkt->flags);
        }        
    }
    
    if(pkt->rcount == pkt->datalen + RADIO_RADIO_CRC_LEN +
       RADIO_SOFT_TRAILER_LEN + RADIO_HARD_TRAILER_LEN +
       RADIO_SERIAL_CRC_LEN) {
        set_bit(PKF_GOTSERIALCRC, &pkt->flags);
        moteradio_process_packet(priv, pkt);
    } else {
        debug_verbose_printk_args("NOT setting PKF_GOTSERIALCRC because %d != %d",
                                  pkt->rcount, pkt->datalen + RADIO_RADIO_CRC_LEN +
                                  RADIO_SOFT_HEADER_LEN + RADIO_HARD_HEADER_LEN +
                                  RADIO_SERIAL_CRC_LEN);
    }
    
    MR_UNLOCK_AND_RETURN;
   
  }
#endif
  
  MR_UNLOCK_AND_RETURN;
  
}

static void moteradio_process_packet(struct moteradio_priv *priv,
				     struct radio_packet *pkt)
{
  char begin_char = CHAR_BEGIN;
  u8 real_soft_len = pkt->datalen + RADIO_SOFT_HEADER_LEN + RADIO_RADIO_CRC_LEN +
      RADIO_SOFT_TRAILER_LEN;
  u8 real_total_len = RADIO_HARD_HEADER_LEN + real_soft_len +
      RADIO_HARD_TRAILER_LEN + RADIO_SERIAL_CRC_LEN;
  u8 fullpacket[real_total_len];

  /* !! THIS FUNCTION ASSUMES WE HAVE THE LOCK !! */
  
  MUST_NOT_BE_NULL(pkt, PKT_DESC);
  MUST_NOT_BE_NULL(priv, PRIV_DESC);
  MUST_MATCH_MAGIC(priv, PRIV_DESC);
  MUST_NOT_BE_NULL(&priv->lock, LOCK_DESC);
  
  /* In the future, if we implement a full stack for this (like a new ARPHRD value
     and everything) we will probably want to strip the hard header byte (and possibly the
     soft header and CRC too depending on what we want to do with this), but fir the time
     being the Ethereal dissector uses the hard header's magic byte to differentiate
     real SLIP captures from our spoofed SLIP captures */
  
  /*  priv->nic.skb = dev_alloc_skb(pkt->datalen); */
  priv->nic.skb = dev_alloc_skb(real_total_len);
  
  debug_verbose_printk("Received entire packet.");

  if (priv->nic.skb == NULL) {
    printk(KERN_WARNING "%s: memory squeeze, dropping packet.\n",
           priv->nic.dev->name); 
    priv->nic.stats.rx_dropped++;
    return;
  }

  if (moteradio_help_crc_check(pkt) != 0) {
      printk(KERN_INFO "%s: bad crc, dropping packet (DEBUG: Not actually dropping!).\n",
             priv->nic.dev->name);

    priv->nic.stats.rx_crc_errors++;
    priv->nic.stats.rx_dropped++;

#ifdef MOTERADIO_DROP_BAD_RADIO_CRCS
    return;
#endif
    
  }

  priv->nic.skb->dev = priv->nic.dev;

  memcpy(fullpacket, &begin_char, sizeof(char));

#ifndef MOTERADIO_NO_TIMESTAMP
  memcpy(fullpacket+sizeof(char), &(pkt->timestamp),
	 RADIO_HARD_HEADER_LEN - sizeof(char));
#endif

  memcpy(fullpacket+RADIO_HARD_HEADER_LEN, &real_soft_len, 
	 RADIO_SOFT_HEADER_LEN);
  memcpy(fullpacket+RADIO_HARD_HEADER_LEN+RADIO_SOFT_HEADER_LEN, 
	 pkt->data, pkt->datalen);
  memcpy(fullpacket+RADIO_HARD_HEADER_LEN+RADIO_SOFT_HEADER_LEN
	 +(pkt->datalen), pkt->radio_crc, RADIO_RADIO_CRC_LEN);

#if (RADIO_SOFT_TRAILER_LEN + RADIO_HARD_TRAILER_LEN) > 0
  memcpy(fullpacket+RADIO_HARD_HEADER_LEN+RADIO_SOFT_HEADER_LEN
         +(pkt->datalen)+RADIO_RADIO_CRC_LEN, pkt->trailer,
         RADIO_SOFT_TRAILER_LEN + RADIO_HARD_TRAILER_LEN);
#endif

#if (RADIO_SERIAL_CRC_LEN > 0)
  memcpy(fullpacket+RADIO_HARD_HEADER_LEN+RADIO_SOFT_HEADER_LEN
         +(pkt->datalen)+RADIO_RADIO_CRC_LEN+RADIO_SOFT_TRAILER_LEN
         +RADIO_HARD_TRAILER_LEN, pkt->serial_crc,
         RADIO_SERIAL_CRC_LEN);
#endif

  memcpy(skb_put(priv->nic.skb, real_total_len), 
	 fullpacket, real_total_len);

  priv->nic.skb->mac.raw = priv->nic.skb->data;

  priv->nic.skb->protocol = htons(ETH_P_ALL); /* <-- !!! */
  
  netif_rx(priv->nic.skb);
 
  priv->nic.dev->last_rx = jiffies;
  priv->nic.stats.rx_packets++;

  moteradio_help_free_pkt(priv, pkt);

  return;
}

/* --- Kernel Interface --- */

static int __init moteradio_init_module(void)
{
  int i, retval;

  retval = -ENOMEM;

/* Print identifying information, including which options are enabled */
  
  printk(KERN_INFO "moteradio: version %s (dynamic channels, maxium of %d devices, mtu=%d).\n",
	 MOTERADIO_VERSION, MOTERADIO_DEVS, RADIO_MTU);
  
#ifdef MOTERADIO_DEBUG
#ifdef MOTERADIO_DEBUG_VERBOSE
    printk(KERN_INFO "moteradio: verbose debugging output enabled.\n");
#else
    printk(KERN_INFO "moteradio: debugging output enabled.\n");
#endif
#endif

#ifndef MOTERADIO_NO_TIMESTAMP
    printk(KERN_INFO "moteradio: timestamping enabled.\n");
#endif
#ifdef MOTERADIO_STEAL_SLIP_LDISC
    printk(KERN_INFO "moteradio: hijacking SLIP line discipline code (instead of serial mouse).\n");
#endif
#ifdef MOTERADIO_BYTE_STUFFING
    printk(KERN_INFO "moteradio: byte stuffing support enabled (stuff byte = 0x%X).\n", CHAR_STUFF);
#endif
#ifdef MOTERADIO_META_FLAGS
    printk(KERN_INFO "moteradio: meta flag support enabled.\n");
#endif
#ifdef MOTERADIO_NO_NULL_CHECKS
    printk(KERN_INFO "moteradio: null pointer checks disabled (dangerous!).\n");
#endif
#ifdef MOTERADIO_NO_MAGIC_CHECKS
    printk(KERN_INFO "moteradio: magic number checks disabled (dangerous!).\n");
#endif
#ifdef MOTERADIO_NO_LOCKS
    printk(KERN_INFO "moteradio: spinlocks disabled (dangerous!).\n");
#endif
#ifdef MOTERADIO_FORCE_OLD_LDISC_API
    printk(KERN_INFO "moteradio: old line discipline API forced (dangerous for kernels >= 2.6.13!).\n");
#endif
#ifdef MOTERADIO_WAIT_FOR_BUSY_DEVICES
    printk(KERN_INFO "moteradio: will wait for busy devices.\n");
#endif
#ifdef MOTERADIO_DROP_BAD_RADIO_CRCS
    printk(KERN_INFO "moteradio: will drop packets with bad radio CRC.\n");
#endif
    
    
  /* Set the "zero" time */
#ifndef MOTERADIO_NO_TIMESTAMP
  do_gettimeofday(&time_zero);
#endif

  /* NICs are allocated dynamically */

  for (i = 0; i < MOTERADIO_DEVS; i++)
    moteradio_devs[i] = NULL;
  
  /* Register the line discipline */
  
  if((retval = tty_register_ldisc(N_MOTERADIO, &moteradio_ldisc)) != 0) {
      printk(KERN_ERR "moteradio: fatal error %i registering line discipline \"%s\" (fatal)\n",
	   retval, moteradio_ldisc.name);
    moteradio_exit_module();
  }

  return retval;
}

static void __exit moteradio_exit_module(void)
{
  int i;
  struct net_device *dev;
  struct moteradio_priv *priv;
#ifdef MOTERADIO_KERNEL_2_4
  moteradio_ctrl_t *ctrl;
#endif
  unsigned long timeout;
  int busy = 0;

  timeout = jiffies + HZ;

  /* Hang up on all TTYs */

  do {


#ifdef MOTERADIO_WAIT_FOR_BUSY_DEVICES

      if (busy) {
          set_current_state(TASK_INTERRUPTIBLE);
          schedule_timeout(HZ/10);
      }
      
#endif

    busy = 0;

    for (i = 0; i < MOTERADIO_DEVS; i++) {
      dev = moteradio_devs[i];
      if (!dev) continue;
      
      priv = netdev_priv(dev);

#ifndef MOTERADIO_NO_NULL_CHECKS
      if (!priv) {
          debug_printk("Priv is NULL for some goddamned reason (in TTY shutdown).");
          continue;
      }
#endif

#ifndef MOTERADIO_NO_MAGIC_CHECKS
      if (priv->magic != MOTERADIO_MAGIC) {
          debug_printk("Priv does not match magic for some goddamned reason (in TTY shutdown).");
          continue;
      }
#endif
      
      if (&priv->lock == NULL) {
          debug_printk("Device lock is NULL for some goddamned reason (in TTY shutdown).");
      } else {

	MR_LOCK;
	
	if(priv->tty.dev) {
	  busy++;
	  tty_hangup(priv->tty.dev);
	}
	
	MR_UNLOCK;
	
      }
    }
  } while (busy && time_before(jiffies, timeout));

  /* Unregister all NICs */

  for (i = 0; i < MOTERADIO_DEVS; i++) {
    dev = moteradio_devs[i];
    if (!dev) continue;

    moteradio_devs[i] = NULL;
    priv = netdev_priv(dev);
    
    if (!priv) continue;

    if (priv->magic != MOTERADIO_MAGIC) {
        debug_printk("Priv does not match magic for some goddamned reason (in NIC shutdown).");
        continue;
    }
    
    if (&priv->lock == NULL) {
      debug_printk("Device lock is NULL for some goddamned reason (in NIC shutdown).");
    } else {

      if (priv->tty.dev) {
          printk(KERN_ERR "%s: tty discipline still running.\n",
                 dev->name);
        MR_LOCK;
	dev->destructor = NULL;
        MR_UNLOCK;
      }
      
    }

    unregister_netdev(dev);

#ifdef MOTERADIO_KERNEL_2_4
    if (dev != NULL) {
        priv = netdev_priv(dev);
        
        if (priv != NULL && priv->magic == MOTERADIO_MAGIC) {
          
            ctrl = (moteradio_ctrl_t*)(priv->container);
            if (ctrl != NULL)
                kfree(ctrl);
        }
    }
#endif

  }    

  /* Clear our line discipline */
  
  if ((i = MR_UNREGISTER_LDISC(N_MOTERADIO))) {      
      printk(KERN_ERR "moteradio: can't unregister line discipline (err = %d).\n", i); 
  }


}

module_init(moteradio_init_module);
module_exit(moteradio_exit_module);

MODULE_AUTHOR("Tyler McHenry");
MODULE_LICENSE("GPL");

#ifndef MOTERADIO_KERNEL_2_4
MODULE_ALIAS_LDISC(N_MOTERADIO);
#endif
