/*
 * $Id: moteradio.h 5884 2006-10-30 22:59:55Z smchenry $
 *
 * See moteradio.c for information
 *
 */

#ifndef MOTERADIO_H
#define MOTERADIO_H

#include <linux/in.h>
#include <linux/types.h>
#include <linux/netdevice.h>   /* struct device, and other headers */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/version.h>

#define MOTERADIO_VERSION "0.6"

/* --- Fundamental module parameters --- */

#ifndef MOTERADIO_DEVS
#define MOTERADIO_DEVS 1
#endif

#ifndef MOTERADIO_STEAL_SLIP_LDISC
#define N_MOTERADIO 2 /* Hijacking N_MOUSE */
#else
#define N_MOTERADIO 1 /* Hijacking N_SLIP */
#endif

#define MOTERADIO_MAGIC 0x1016B5FC

#ifndef ARPHRD_RADIO
#define ARPHRD_RADIO 20
#endif

/* --- All sorts of debugging macros --- */

#ifdef MOTERADIO_DEBUG
#define debug_printk(msg) printk(KERN_INFO "[DEBUG] %s: " #msg "\n", (priv && priv->nic.dev ? priv->nic.dev->name : "moteradio"))
#define debug_printk_args(msg,...) printk(KERN_INFO "[DEBUG] %s: " #msg "\n", (priv && priv->nic.dev ? priv->nic.dev->name : "moteradio"), __VA_ARGS__)
#define debug_printk_nopriv(msg) printk(KERN_INFO "[DEBUG] moteradio: " #msg "\n")
#define debug_printk_nopriv_args(msg,...) printk(KERN_INFO "[DEBUG] moteradio: " #msg "\n", __VA_ARGS__)
#else
#define debug_printk(msg)
#define debug_printk_args(msg,...)
#define debug_printk_nopriv(msg)
#define debug_printk_nopriv_args(msg,...) 
#endif

#if defined(MOTERADIO_DEBUG) && defined(MOTERADIO_DEBUG_VERBOSE)
#define debug_verbose_printk(msg) debug_printk(msg)
#define debug_verbose_printk_args(msg,...) debug_printk_args(msg, __VA_ARGS__)
#define debug_verbose_printk_nopriv(msg) debug_printk_nopriv(msg)
#define debug_verbose_printk_nopriv_args(msg,...) debug_printk_nopriv_args(msg, __VA_ARGS__)
#else
#define debug_verbose_printk(msg)
#define debug_verbose_printk_args(msg,...)
#define debug_verbose_printk_nopriv(msg)
#define debug_verbose_printk_nopriv_args(msg,...)
#endif

#define DEBUG_MARK debug_verbose_printk_nopriv_args("Markpoint at line %d reached", __LINE__)

#define PRIV_DESC "private data pointer"
#define PKT_DESC "incoming packet pointer"
#define NET_DEV_DESC "network interface device pointer"
#define TTY_DESC "serial terminal info pointer"
#define TTY_DEV_DESC "serial terminal device pointer"
#define RXQ_DESC "packet recieve queue"
#define LOCK_DESC "master lock"
#define FILE_DESC "ioctl file pointer"

#ifndef MOTERADIO_NO_NULL_CHECKS
#define MUST_NOT_BE_NULL(x,desc) { if (!(x)) { debug_printk_nopriv_args("%s is null in %s!", (desc), __func__); return; } }
#define MUST_NOT_BE_NULL_R(x,desc,r) { if (!(x)) { debug_printk_nopriv_args("%s is null in %s!", (desc), __func__); return (r); } }
#else
#define MUST_NOT_BE_NULL(x,desc)
#define MUST_NOT_BE_NULL_R(x,desc,r)
#endif

#ifndef MOTERADIO_NO_MAGIC_CHECKS
#define MUST_MATCH_MAGIC(x,desc) { if ((x)->magic != MOTERADIO_MAGIC) { debug_printk_nopriv_args("%s did not match magic in %s!", (desc), __func__); return; } }
#define MUST_MATCH_MAGIC_R(x,desc,r){ if ((x)->magic != MOTERADIO_MAGIC) { debug_printk_nopriv_args("%s did not match magic in %s!", (desc), __func__); return (r); } }
#else
#define MUST_MATCH_MAGIC(x,desc)
#define MUST_MATCH_MAGIC_R(x,desc,r)
#endif

/* --- Debug wrappers for spinlocks --- */

#ifndef MOTERADIO_NO_LOCKS
#define MR_LOCK { debug_verbose_printk_args("Locking in %s", __func__); spin_lock_irqsave(&priv->lock, priv->irq_flags); }
#define MR_UNLOCK { debug_verbose_printk_args("Unlocking in %s", __func__); spin_unlock_irqrestore(&priv->lock, priv->irq_flags); }
#define MR_UNLOCK_AND_RETURN { MR_UNLOCK; return; }
#else
#define MR_LOCK
#define MR_UNLOCK
#define MR_UNLOCK_AND_RETURN { return; }
#endif

/* --- Compatability with various kernel versions --- */

#ifndef LINUX_VERSION_CODE
#warning "Could not detect kernel version - see README to avoid serious errors"
#elif !defined(KERNEL_VERSION)
#warning "Build environment lacks kernel version comparisons - see README to avoid serious errors"
#elif (LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0))
#define MOTERADIO_KERNEL_2_4
#endif
 
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 13)) || defined(MOTERADIO_FORCE_OLD_LDISC_API) || defined(MOTERADIO_KERNEL_2_4)
// Using this interface in a post-2.6.12 kernel causes a PANIC on rmmod
#define MR_UNREGISTER_LDISC(n) (tty_register_ldisc(n, NULL))
#else
#define MR_UNREGISTER_LDISC(n) (tty_unregister_ldisc(n))
#endif


#ifdef MOTERADIO_KERNEL_2_4
// I think some 2.6 versions use this too, but I don't think it is vital for them
#define MOTERADIO_USE_RECEIVE_ROOM
#endif

/* --- Data structures and associated parameters --- */

/* Packet Description:

   There is a distinction between packets received "over the air"
   i.e. the bytes that are actually transmitted and received at the MAC
   layer or higher using the mote's radio (soft), and bytes that are
   added on after the packet is received by the snooper or by
   this driver (hard). The max payload (data) length is the MTU of
   the radio minus any extra soft bytes. Hard bytes do not affect the max
   payload since they are not transmitted, and anything included in the
   soft bytes (length, CRC) does not take into account anything in the
   hard bytes.

   A full packet goes in this order:

   [H_HEADER] [S_HEADER] [PAYLOAD] [CRC] [S_TRAILER] [H_TRAILER]

   Where any of these is conceivably zero length.

   The default TinyOS message format has a 1 byte soft header (which
   indicates the length of the soft portion of the packet), up to 509
   bytes of payload, and a 2 byte CRC-16. There are no soft trailer bytes.
   The radio MTU is therefore 512 bytes.

   The snooper adds one hard header byte (start symbol), and optionally
   adds one hard trailer byte (meta flags).

   This driver optionally adds an extra 4 hard header bytes (timestamp).

   This driver should not rely or modify any data contained in the payload
   portion of the packet.
*/

#ifndef RADIO_MTU
#define RADIO_MTU             512  /* Max bytes receivable over the air */
#endif

/* Hard header = bytes added in front of packet, not recv'd over air */

#ifndef MOTERADIO_NO_TIMESTAMP
#define RADIO_HARD_HEADER_LEN   sizeof(char) + sizeof(uint64_t)
#else
#define RADIO_HARD_HEADER_LEN   sizeof(char)
#endif

#define RADIO_SOFT_HEADER_LEN   1 /* Bytes before payload, over air */
#define RADIO_RADIO_CRC_LEN     2 /* CRC length (immediately after payload), over air */
#define RADIO_SOFT_TRAILER_LEN  0 /* Bytes after CRC, over air */

/* Hard trailer = bytes added after packet, not recv'd over air */

#ifndef MOTERADIO_META_FLAGS
#define RADIO_HARD_TRAILER_LEN  0
#define RADIO_SERIAL_CRC_LEN    0
#else
#define RADIO_HARD_TRAILER_LEN  1
#define RADIO_SERIAL_CRC_LEN    2
#endif

/* Not affected by things that do not
   actually come over the radio, such as
   timestamps and trailer bytes */

#define RADIO_MAX_DATA_LEN RADIO_MTU - RADIO_SOFT_HEADER_LEN - RADIO_RADIO_CRC_LEN - RADIO_SOFT_TRAILER_LEN


/* Serial line flags */

#define SLF_INUSE	0		/* Channel in use               */
#define SLF_ERROR	2               /* Parity, etc. error           */

/* Packet flags */

#define PKF_BEGUN        1  /* Mark position in reading packet */
#define PKF_GOTLEN       2
#define PKF_GOTDATA      3
#define PKF_GOTRADIOCRC  4
#ifdef MOTERADIO_META_FLAGS
#define PKF_GOTTRAILER   5
#define PKF_GOTSERIALCRC 6
#endif
#define PKF_ERROR        8

/* Stream flags */

#ifdef MOTERADIO_BYTE_STUFFING
#define STF_STUFFED      1  /* Next byte should be treated as non-control */
#endif

/* Special characters */

#define CHAR_BEGIN      0x7E
#ifdef MOTERADIO_BYTE_STUFFING
#define CHAR_STUFF      0xAA
#endif

/* Intra-packet flags */

#ifdef MOTERADIO_META_FLAGS
#define IPF_POSSIBLE_COLLISION 0x01
#endif

/* The packet that is actually received from the radio, with metadata */

struct radio_packet {

    /* metadata: */
    
    struct radio_packet *next;
    struct radio_packet *prev;
    struct net_device *dev;

    int magic;
    
    unsigned int rcount;

    /* These fields are actually recieved over the radio: */

    unsigned int datalen;
    
    u8 data[RADIO_MAX_DATA_LEN];
    u8 radio_crc[RADIO_RADIO_CRC_LEN];
#if (RADIO_SOFT_TRAILER_LEN + RADIO_HARD_TRAILER_LEN) > 0
    u8 trailer[RADIO_SOFT_TRAILER_LEN + RADIO_HARD_TRAILER_LEN];
#endif
#if (RADIO_SERIAL_CRC_LEN > 0)
    u8 serial_crc[RADIO_SERIAL_CRC_LEN];
#endif

    /* These fields are added and/or processed: */
    
#ifndef MOTERADIO_NO_TIMESTAMP
  uint64_t timestamp;
#endif
  unsigned long flags;
};

/* Internal status of the network interface device */

struct moteradio_priv_nic {
  struct net_device       *dev;
  struct net_device_stats  stats;
  struct radio_packet     *rx_queue;
  struct sk_buff          *skb;
};

/* Internal status of the serial terminal device */

struct moteradio_priv_tty {
  struct tty_struct    *dev;

  dev_t line;
  pid_t pid;
  
  int status;
};

/* Internal status of the moteradio driver */

struct moteradio_priv {
    struct moteradio_priv_nic nic;
    struct moteradio_priv_tty tty;
    
    spinlock_t lock;

    int magic;

    unsigned long flags;         /* for current packet */
    unsigned long stream_flags;  /* persistent across packets */
    unsigned long irq_flags;     /* for storing state during locks */
    
#ifdef MOTERADIO_KERNEL_2_4
    void* container;
#endif

};

/* Misc constants */

#ifndef MOTERADIO_NO_TIMESTAMP
#define USECS_PER_SEC 1000000
#endif

/* Workarounds for 2.4 backport */

#ifdef MOTERADIO_KERNEL_2_4

#define netdev_priv(dev) (struct moteradio_priv*)(dev->priv)
#define tty_devnum(tty) tty->device
 
typedef struct moteradio_ctrl {
  struct moteradio_priv priv;
  struct net_device dev;
} moteradio_ctrl_t;

#endif // kernel 2.4

#endif // include guard
